#!/usr/bin/env python
"""
Top-level build script for the universal_lib package.

This program requires either Python version 2.6+ or 3.2+, and is importable as a
python module.
"""

from __future__ import print_function, absolute_import, division, unicode_literals

import sys
import os
import subprocess
import getopt
import shlex
import shutil
import glob

app_label = "universal_lib"
body_moniker = ''

#------------------------------------------------------------------------------
# Emulates 'rm -f' functionality:
def rm_f(fpath):
    """
    Emulates 'rm -f' functionality.
    """

    if os.path.isfile(fpath) or os.path.islink(fpath):
        os.remove(fpath)
#------------------------------------------------------------------------------


#------------------------------------------------------------------------------
def extend_composite_pnf(anchor_path, pnf_fpath, sdwe_fpath, pnf_pfx,
                  F2003_capable):  # All strings, except F2003_capable (boolean)
    """Augment a composite pathname file (list of filepaths)."""

    root_src_path = ''  # Initialize
    with open(pnf_fpath, "at") as pnf_f:   # Open for appending
        with open(sdwe_fpath, "rt") as input_f:
            input_f_line_list = input_f.readlines()  # Read a subdirectory list
        for line in input_f_line_list:
            trimmed_line = line.strip()
            if trimmed_line:     # If not completely "whitespace"
                if not root_src_path:   # Set root source code path
                    root_src_path = os.path.abspath( \
                        os.path.join(anchor_path, trimmed_line))
                else:      # A subdirectory to query
                    query_sub_pnf = True   # Assume it will be queried

                      # Inhibit F2003 subdir processing if not supported:
                    if os.path.join("Fortran", "2003") in trimmed_line:
                        query_sub_pnf = F2003_capable

                    if query_sub_pnf:   # Add to composite pathname list
                        sub_pnf_fpath = os.path.join(root_src_path,
                                                  trimmed_line, "etc", pnf_pfx)
                        if os.path.isfile(sub_pnf_fpath):
                            with open(sub_pnf_fpath, "rt") as sub_pnf_f:
                                  # Read a list of source code files:
                                sub_pnf_f_line_list = sub_pnf_f.readlines()
                            for pnf_line in sub_pnf_f_line_list:
                                trimmed_pnf_line = pnf_line.strip()
                                if trimmed_pnf_line:  # If not all "whitespace"
                                      # Append root source code path:
                                    line_to_append = os.path.join(root_src_path,
                                                                  pnf_line)
                                    pnf_f.write(line_to_append)
#------------------------------------------------------------------------------


#------------------------------------------------------------------------------
def main(action):
    """Build driver."""

    #=== Define some useful parameters:

      # Determine fully-qualified filesystem location of this script:
    anchor_path = os.path.abspath(os.path.dirname(sys.argv[0]))
    os.chdir(anchor_path)   # Start from a "known" location

      # Frequently referenced parameters:

    local_infra_path = os.path.abspath(os.path.join(anchor_path, ".."))
    local_iscripts_path = os.path.join(local_infra_path, "misc_scripts")
    work_path = os.path.join(anchor_path, "work")
    makedir_path = work_path    # Where source code is compiled 
    cbenv_path = os.path.join(anchor_path, "configure_build_env")
    cbenv_work_path = os.path.join(cbenv_path, "work_cbenv")

    build_envf_pfx = "build_env"
    bec_path = os.path.join(anchor_path, build_envf_pfx+"_config")
    pnf_pfx = "path_names"
    univ_kind_defs_fn = "univ_kind_defs.f90"
    if body_moniker:
        app_label_wo_body = app_label.replace('_'+body_moniker, '')
    else:
        app_label_wo_body = app_label
    app_kinds_fn = app_label_wo_body+"_kinds.f90"
    src_dirs_w_etc_fpath_pfx = os.path.join(anchor_path, "src_dirs_with_etc")
    make_exe = "make"   # Name of default system 'make' program to use

    makefile_fpath = os.path.join(work_path, "Makefile")
    crmf_include_fpath = os.path.join(makedir_path,
                                      build_envf_pfx+".mk.include")
    src_dirs_w_etc_fpath = src_dirs_w_etc_fpath_pfx+".cfg"
    crmf_fpath = os.path.join(anchor_path, "create_mf.py")   # Path to script
    univ_kind_defs_fpath = os.path.join(local_infra_path, "src",
                                        univ_kind_defs_fn)
    app_kinds_fpath = os.path.join(local_infra_path, "src", app_kinds_fn)

      # Further argument processing:

    arg_check_action = ["clean", "build", "query_build"]
    arg_check_clean = ["ALL"]
    arg_check_build = ["serial", "DM_only", "SM_only", "hybrid", "debug"]
    arg_check_unary = ["query_build"]
    if action[0] in arg_check_action:   # Is the specified action valid?
        this_action = action[0]
        len_args_action_list = len(action)
        if this_action in arg_check_unary:   # Should not have extraneous args
            if len_args_action_list > 1:
                usage()   # Disallow extraneous arguments
                sys.exit(1)
        elif this_action == "clean":   # May have 0 or 1 extra argument
            clean_ALL = False   # Default
            for arg_item in action[1:]:
                if arg_item in arg_check_clean:  # Is the specified token valid?
                    clean_ALL = True
                else:                      # The specified token is invalid
                    usage()
                    sys.exit(1)
        elif this_action == "build":   # May have 1, 2, or 3 extra arguments
            build_debug = False              # Default build settings
            mf_template_type = "normal"      #
            build_parallel = False           #
            parallel_method = ''             #

            if len_args_action_list < 2:
                usage()   # Require at least 1 extra argument
                sys.exit(1)
            else:
                if action[1] in arg_check_build:
                    usage()   # Must not be a valid optional arg
                    sys.exit(1)
                else:
                    libexe_fpath = action[1]

            if len_args_action_list > 2:
                for arg_item in action[2:]:
                    if arg_item in arg_check_build:  # Is the given token valid?
                        if arg_item == "debug":
                            build_debug = True
                            mf_template_type = "debug"
                        elif arg_item == "serial":
                            build_parallel = False
                        else:
                            build_parallel = True
                            parallel_method = arg_item
                    else:                      # The specified token is invalid
                        usage()
                        sys.exit(1)
    else:                         # The specified action is invalid
        usage()
        sys.exit(1)

    #=== Perform specified action(s):

    if this_action == "clean":
        if os.path.isdir(makedir_path):  # If "make" work directory exists
            if os.path.isfile(makefile_fpath):  # If Makefile exists
                beg_cwd = os.getcwd()  # Record the current working directory
                os.chdir(makedir_path)   # Enter the build "work" directory
                subprocess.check_call([make_exe, "-f", makefile_fpath, "clean"])
                os.chdir(beg_cwd)   # Return to the original directory
            shutil.rmtree(makedir_path)   # Delete "make" work directory

        if clean_ALL:
            rm_f(univ_kind_defs_fpath)    # Remove old source file
            rm_f(app_kinds_fpath)    # Remove old source file
            if os.path.isdir(cbenv_work_path):  # If cbenv work directory exists
                shutil.rmtree(cbenv_work_path)     # Remove it

    elif this_action == "build":
          # If needed, create the "make" work directory:
        if not os.path.isdir(makedir_path):
            os.mkdir(makedir_path)

        # Configure the build environment:

        if not os.path.isfile(crmf_include_fpath):
            build_env_cfg_fpath = os.path.join(anchor_path,
                                  build_envf_pfx+"_config."+mf_template_type)
            supp_defs_fpath = os.path.join(anchor_path,
                                           "supplementary_defs.sh.include")

            rm_f(univ_kind_defs_fpath)   # Remove old "kind_defs" file
            rm_f(app_kinds_fpath)   # Remove old "kinds" file

              # Write supplemental definitions file:
            text_to_write = \
"supp_C_defs=;\n"+ \
"supp_Fortran_defs=;\n"+ \
"supp_C_includes=\"-I"+makedir_path+"\";\n"+ \
"supp_Fortran_includes=\"-I"+makedir_path+"\";\n"+ \
"supp_Fortran_modpaths=\"-I"+makedir_path+"\";\n"+ \
"supp_C_linker_flags=;\n"+ \
"supp_C_linker_libs=;\n"+ \
"supp_Fortran_linker_flags=;\n"+ \
"supp_Fortran_linker_libs=;\n"
            with open(supp_defs_fpath, "wt") as text_f:
                text_f.write(text_to_write)

              # Create build environment information file from templates:
            beg_cwd = os.getcwd()  # Record the current working directory
            os.chdir(bec_path)   # Enter the build_env_config directory
            script_fpath = os.path.join(local_iscripts_path,
                                        "process_template_file.py")
            subprocess.check_call([script_fpath, "z-bec."+mf_template_type])
            os.chdir(beg_cwd)   # Return to previous working directory

              # Execute the "build_env_configure" scripts to determine the
              #  build environment:
            script_fpath = os.path.join(cbenv_path, "src",
                                        "build_env_configure.py")
            subprocess.check_call([script_fpath, anchor_path,
                                   build_env_cfg_fpath, supp_defs_fpath,
                                   app_label_wo_body])

              # Move new "kind_defs" source code file to its proper location:
            fpath_to_move = os.path.join(anchor_path, univ_kind_defs_fn)
            shutil.move(fpath_to_move, univ_kind_defs_fpath)

              # Move new "kinds" source code file to its proper location:
            fpath_to_move = os.path.join(anchor_path, app_kinds_fn)
            shutil.move(fpath_to_move, app_kinds_fpath)

              # Move other needed files to their proper locations:
            fpath_globs_to_move = [ \
                      os.path.join(anchor_path, build_envf_pfx+"_app*.include"),
                      os.path.join(anchor_path, build_envf_pfx+"_app*.h")]
            for fpath_glob in fpath_globs_to_move:
                for fpath in glob.glob(fpath_glob):
                    shutil.copy(fpath, makedir_path)
                    rm_f(fpath)
            shutil.copy(supp_defs_fpath, makedir_path)
            rm_f(supp_defs_fpath)

        # Read build environment information files, parse their contents, and
        #  save the results in a dictionary for later reference:

        abd = {}

        with open(os.path.join(makedir_path,
                         build_envf_pfx+"_appind.sh.include"), "rt") as input_f:
            input_f_raw = input_f.read()
        for pair in shlex.split(input_f_raw, True, True):
            split_pair = (pair.replace(';', ' ')).split('=', 1)
            abd[split_pair[0]] = split_pair[1] 

        with open(os.path.join(makedir_path,
                         build_envf_pfx+"_appdep.sh.include"), "rt") as input_f:
            input_f_raw = input_f.read()
        for pair in shlex.split(input_f_raw, True, True):
            split_pair = (pair.replace(';', ' ')).split('=', 1)
            abd[split_pair[0]] = split_pair[1]

        F2003_capable = abd["some_F2003_capability"].strip() == "yes"

        # Create composite pathname file (list); code that is independent of
        #  external libraries:

        pnf_fpath = os.path.join(makedir_path, pnf_pfx)
        if not os.path.isfile(pnf_fpath):
            create_new_pnf = True
            extend_composite_pnf(anchor_path, pnf_fpath, src_dirs_w_etc_fpath,
                                 pnf_pfx, F2003_capable)
        else:
            create_new_pnf = False

        # Augment composite pathname file (list); code that is dependent on
        #  external libraries that were deemed usable:

        if create_new_pnf:
            for aux_lib in abd["usable_aux_libs"].split():
                sdwe_aux_fpath = src_dirs_w_etc_fpath_pfx+'-'+aux_lib+".cfg"
                extend_composite_pnf(anchor_path, pnf_fpath, sdwe_aux_fpath,
                                     pnf_pfx, F2003_capable)

        # Create build environment configuration file for 'create_mf.py':
        text_to_write = \
"# make-syntax build environment configuration file for create_mf.py\n"+ \
"\n"+ \
"#~ auto-generated\n"+ \
"\n"+ \
"FC = "+abd["Fortran_compiler"]+"\n"+ \
"FFLAGS = "+abd["Fortran_compiler_flags"]+abd["app_Fortran_compiler_flags"]+"\n"+ \
"FPPDEFS = "+abd["Fortran_preproc_defines"]+abd["app_Fortran_preproc_defines"]+"\n"+ \
"FPPFLAGS = "+abd["Fortran_preproc_miscflags"]+abd["app_Fortran_preproc_miscflags"]+abd["Fortran_preproc_includes"]+abd["app_Fortran_preproc_includes"]+"\n"+ \
"FMODPATHS = "+abd["app_Fortran_modpaths"]+"\n"+ \
"\n"+ \
"CC = "+abd["C_compiler"]+"\n"+ \
"CFLAGS = "+abd["C_compiler_flags"]+abd["app_C_compiler_flags"]+"\n"+ \
"CPPDEFS = "+abd["C_preproc_defines"]+abd["app_C_preproc_defines"]+"\n"+ \
"CPPFLAGS = "+abd["C_preproc_miscflags"]+abd["app_C_preproc_miscflags"]+abd["C_preproc_includes"]+abd["app_C_preproc_includes"]+"\n"+ \
"\n"+ \
"LD = "+abd["Fortran_compiler"]+"\n"+ \
"LDFLAGS = "+abd["Fortran_linker_flags"]+abd["Fortran_linker_libs"]+abd["C_linker_flags"]+abd["C_linker_libs"]+abd["app_Fortran_linker_flags"]+abd["app_Fortran_linker_libs"]+abd["app_C_linker_flags"]+abd["app_C_linker_libs"]+"\n"+ \
"\n"
        with open(crmf_include_fpath, "wt") as text_f:
            text_f.write(text_to_write)

        # Generate a Makefile, then execute its instructions:

        pnf_fpath = os.path.join(makedir_path, pnf_pfx)
        subprocess.check_call([crmf_fpath, "--mffpath", makefile_fpath,
                               "--finaltgt", libexe_fpath, "--mfpreamble",
                               crmf_include_fpath, pnf_fpath])

        beg_cwd = os.getcwd()  # Record the current working directory
        os.chdir(makedir_path)   # Enter the build "work" directory
        subprocess.check_call([make_exe, "-f", makefile_fpath])
        os.chdir(beg_cwd)   # Return to the original directory

    if this_action == "query_build":
        if os.path.isfile(crmf_include_fpath):
            appind_sh_inc_fpath = os.path.join(makedir_path,
                                            build_envf_pfx+"_appind.sh.include")
            appind_mk_inc_fpath = os.path.join(makedir_path,
                                            build_envf_pfx+"_appind.mk.include")
            appind_c_h_fpath = os.path.join(makedir_path,
                                            build_envf_pfx+"_appind.h")
            extra_C_inc_fpath = os.path.join(anchor_path, "..", "..", "src",
                                             'C', "include")

              # Read build environment information files, parse their contents,
              #  and save the results in a dictionary for later reference:
            abd = {}
            with open(appind_sh_inc_fpath, "rt") as input_f:
                input_f_raw = input_f.read()
            for pair in shlex.split(input_f_raw, True, True):
                split_pair = (pair.replace(';', ' ')).split('=', 1)
                abd[split_pair[0]] = split_pair[1]

              # Construct and "deliver" query reply:
            text_to_print = \
"BUILT_LIBS_PATH=\""+makedir_path+"\";\n"+ \
"BUILT_MODS_PATH=\""+makedir_path+"\";\n"+ \
"\n"+ \
"APPIND_SH_INC=\""+appind_sh_inc_fpath+"\";\n"+ \
"APPIND_MK_INC=\""+appind_mk_inc_fpath+"\";\n"+ \
"APPIND_C_H=\""+appind_c_h_fpath+"\";\n"+ \
"\n"+ \
"C_INCLUDES=\"-I"+extra_C_inc_fpath+" "+abd["C_preproc_includes"]+"\";\n"+ \
"FORTRAN_INCLUDES=\""+abd["Fortran_preproc_includes"]+"\";\n"+ \
"\n"+ \
"C_LINKER=\""+abd["C_compiler"]+"\";\n"+ \
"FORTRAN_LINKER=\""+abd["Fortran_compiler"]+"\";\n"+ \
"\n"+ \
"C_LIBS_TO_LINK=\""+abd["C_linker_libs"]+"\";\n"+ \
"C_LINKER_FLAGS=\""+abd["C_linker_flags"]+"\";\n"+ \
"FORTRAN_LIBS_TO_LINK=\""+abd["Fortran_linker_libs"]+"\";\n"+ \
"FORTRAN_LINKER_FLAGS=\""+abd["Fortran_linker_flags"]+"\";\n"+ \
"\n"+ \
"FAILED_LIB_CHECKS=\""+abd["failed_aux_libs"]+"\";\n"+ \
"USABLE_LIB_CHECKS=\""+abd["usable_aux_libs"]+"\";\n"
            print(text_to_print)
        else:
            print("build_script.py ERROR: "+crmf_include_fpath+ \
                  " must exist for query_build mode to work.")
            sys.exit(1)
#------------------------------------------------------------------------------


#------------------------------------------------------------------------------
def usage():
    usage_msg = """
usage: build_script.py [-h]  action

Top-level build script for this package.  This program requires either Python
version 2.6+ or 3.2+, and is importable as a python module.

mandatory argument:
  action -- One of the following:
      clean  :  Clean up local non-configure build process debris
      clean ALL  :  Clean up all known build process debris, including
                    that of configure and other linked packages
      build LIBEXE_FILE [debug]  :  Build package
       "         "      serial [debug]  :  (DEFAULT) disable parallel features
       "         "      DM_only [debug]  :  distributed-memory parallel (MPI)
       "         "      SM_only [debug]  :  shared-memory parallel (OpenMP)
                                             (**no code uses this, yet**)
       "         "      hybrid [debug]  :  hybrid parallel (MPI + OpenMP)
                                             (**no code uses this, yet**)
      query_build  :  Usage/linking info for the built package

optional arguments:
  -h, --help   show this help message and exit
"""
    print(usage_msg)
#------------------------------------------------------------------------------


if __name__ == "__main__":
    # Process arguments:
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'h', ["help"])
    except getopt.GetoptError as err:
        print(err)
        usage()
        sys.exit(1)
      # Default option values:
    # NONE
      # Act according to the options given:
    for o, a in opts:
        if o in ("-h", "--help"):
            usage()
            sys.exit(1)
        else:
            assert False, "unhandled option"
      # Act according to the arguments given:
    if (len(args) > 4 or len(args) < 1):
        print("An incorrect number of arguments were provided.")
        usage()
        sys.exit(1)
    action = args

    main(action)
