!```````````````````````````````````````````````````````````````````
module univ_phys_constants_mod

!+ Module containing "universal" physical constants

use univ_kind_defs_mod, only: f8, f4
use univ_math_constants_mod, only: ONE_f8, fourpi_f8, fourpi_f4, ONE_f4,  &
     twopi_f8

implicit none

!--- "Higher" precision definitions (see next section for "normal" precision):

! Fundamental constants:

real(f8), parameter ::  &
     stfnbltz_f8 = 5.670400e-8_f8,  & ! Stefan-Boltzmann [W/(K^4.m^2)] {1}
     grav_const_f8 = 6.6742e-11_f8,  & ! Gravitation constant [N.m^2/kg^2] {1}
     c_f8 = 2.99792458e8_f8,  & ! Speed of light (EM rad) [m/s] {1}
     avogadnum_f8 = 6.0221415e23_f8,  & ! Avogadro's number [mol^-1] {1}
     boltzmann_f8 = 1.3806505e-23_f8,  & ! Boltzmann's const [J/K] {1}
     planck_f8 = 6.6260693e-34_f8,  & ! Planck constant [J.s] {1}
     R_univ_gas_f8 = 8.314472_f8,  & ! "Universal" gas const [J/(mol.K)] {1}
     amu_f8 = 1.66053886e-27_f8,  & ! Atomic mass constant (unit) [kg] {1} 
     e_charge_f8 = 1.602176487e-19_f8   ! Charge of an electron [C] {1}

! Derived constants:

real(f8), parameter ::  &
     magnetic_const_f8 =  &             ! Magnetic constant [N/A^2] {1} (fn.1)
          1.e-7_f8*fourpi_f8,  &
     electric_const_f8 =  &             ! Electric constant [F/m] (fn.2)
          ONE_f8/(c_f8*c_f8*magnetic_const_f8),  &
     coulomb_const_f8 =  &              ! Coulomb's (force) constant [m/F?]
          ONE_f8/(fourpi_f8*electric_const_f8),  &
     h_bar_f8 = planck_f8/twopi_f8  ! Reduced Planck (Dirac) constant [J.s]

!*** REFERENCES:
!
!{1} 2002 CODATA (Committee on Data for Science and Technology) value
!    http://physics.nist.gov/constants
!    Peter J. Mohr and Barry N. Taylor, CODATA Recommended Values of the 
!      Fundamental Physical Constants: 2002, published in Rev. Mod. Phys.
!      vol. 77(1) 1-107 (2005).

! Footnotes (fn.):
!
! 1) Also known as "permeability of free space" or "vacuum permeability".
! 2) Also known as "permittivity of free space" or "vacuum permittivity".

!--- "Normal" precision definitions:
!
!    Note: explicitly calculate (with *_f4 constants) any values that were
!    calculated in the "higher" precision section -- for numerical consistency

! Fundamental constants:

real(f4), parameter ::  &
     stfnbltz_f4 = real(stfnbltz_f8, kind=f4),  &
     grav_const_f4 = real(grav_const_f8, kind=f4),  &
     c_f4 = real(c_f8, kind=f4),  &
     avogadnum_f4 = real(avogadnum_f8, kind=f4),  &
     boltzmann_f4 = real(boltzmann_f8, kind=f4),  &
     planck_f4 = real(planck_f8, kind=f4),  &
     R_univ_gas_f4 = real(R_univ_gas_f8, kind=f4),  &
     amu_f4 = real(amu_f8, kind=f4),  &
     e_charge_f4 = real(e_charge_f8, kind=f4)

! Derived constants:

real(f4), parameter :: &
     magnetic_const_f4 =  &             ! Magnetic constant [N/A^2] {1} (fn.1)
          1.e-7_f4*fourpi_f4,  &
     electric_const_f4 =  &             ! Electric constant [F/m] (fn.2)
          ONE_f4/(c_f4*c_f4*magnetic_const_f4),  &
     coulomb_const_f4 =  &              ! Coulomb's (force) constant [m/F?]
          ONE_f4/(fourpi_f4*electric_const_f4),  &
     h_bar_f4 = real(h_bar_f8, kind=f4)  ! Reduced Planck (Dirac) constant [J.s]

end module univ_phys_constants_mod
