!```````````````````````````````````````````````````````````````````
module univ_GrADS_format_mod

!+ Module that contains subroutines, etc. for writing/reading classic
!   GrADS-format files

use univ_kind_defs_mod, only: f4

implicit none

private

integer, public, parameter :: max_GrADSc_vname_len = 15,  &
                              max_GrADSc_vdesc_len = 140

integer, public, parameter :: max_file_handle_length = 8
real(f4), public :: undef = 1.e34_f4

integer :: ngf

character(len=80) :: gtdef_str

! {ngf}:
integer, allocatable, dimension(:) :: current_rec, io_unit
character(len=max_file_handle_length), allocatable, dimension(:) :: gf_handle
logical, allocatable, dimension(:) :: file_is_open

type in_ptr_array
   integer, pointer, dimension(:) :: ptr => NULL()
end type in_ptr_array

type cn_ptr_array
   character(len=max_GrADSc_vname_len), pointer, dimension(:) :: ptr => NULL()
end type cn_ptr_array

! {ngf}:
type(in_ptr_array), allocatable, dimension(:) :: rdfvar_nz
type(cn_ptr_array), allocatable, dimension(:) :: rdfvar_name
integer, allocatable, dimension(:) :: rdf_nv, rdf_nt, rdf_nt_zrec, rdf_n2,  &
     rdf_n3

public :: init_GrADS_format, exit_GrADS_format, open_GrADS_dfile,  &
     close_GrADS_dfile, read_GrADS_dfile, write_GrADS_dfile,  &
     read_GrADS_hfile, write_GrADS_hfile


contains


!```````````````````````````````````````````````````````````````````
subroutine init_GrADS_format(nfiles, file_handle)

!+ Initializes GrADS module storage, tracking variables

!===================================================================
!----- IN ----------------------------------------------------------
! 'nfiles' = Number of GrADS data files to be open (and/or read/written)
!             simultaneously
! 'file_handle' = File "handle" (nickname) used to specify which file to
!                  perform operations on

use univ_std_fileIO_mod, only: find_file_IOunit
use univ_IO_mem_status_mod, only: illegal_alloc_stop, for_allocation
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

integer, intent(IN) :: nfiles
character(len=*), dimension(nfiles), intent(INOUT) :: file_handle

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'init_GrADS_format'

integer :: iq, iu_last, e

!=== Executable statements:

ngf = nfiles    ! Max number of simultaneously open files

! Allocate storage:
allocate(current_rec(ngf), io_unit(ngf), gf_handle(ngf), rdfvar_name(ngf),  &
     file_is_open(ngf), rdfvar_nz(ngf), rdf_nv(ngf), rdf_nt(ngf),  &
     rdf_nt_zrec(ngf), rdf_n2(ngf), rdf_n3(ngf), stat=e)
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name, 'ngf')

! Fill tracking arrays with initial values:

rdf_nv = 0; rdf_nt = 0; rdf_n2 = 0; rdf_n3 = 0

iu_last = 19
do iq=1,ngf
   nullify(rdfvar_nz(iq)%ptr)
   nullify(rdfvar_name(iq)%ptr) 

   current_rec(iq) = 1
   io_unit(iq) = find_file_IOunit(start_IOunit=iu_last+1)
   iu_last = io_unit(iq)
   gf_handle(iq) = file_handle(iq)(1:max_file_handle_length)
   file_is_open(iq) = .FALSE.
end do

end subroutine init_GrADS_format


!```````````````````````````````````````````````````````````````````
subroutine exit_GrADS_format

!+ Deallocates GrADS module storage, tracking variables

use univ_IO_mem_status_mod, only: illegal_alloc_stop, for_deallocation
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'exit_GrADS_format'

integer :: iq, e

!=== Executable statements:

if (associated(rdfvar_nz(1)%ptr) .and. associated(rdfvar_name(1)%ptr)) then
   do iq=1,ngf
      deallocate(rdfvar_nz(iq)%ptr,rdfvar_name(iq)%ptr, stat=e)
      if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, 'A')
      nullify(rdfvar_nz(iq)%ptr)
      nullify(rdfvar_name(iq)%ptr)
   end do
end if

deallocate(current_rec, io_unit, gf_handle, file_is_open, rdfvar_nz, rdf_nv,  &
     rdf_nt, rdfvar_name, rdf_nt_zrec, rdf_n2, rdf_n3, stat=e)
if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, 'B')

end subroutine exit_GrADS_format


!```````````````````````````````````````````````````````````````````
subroutine find_IOunit(handle, IOunit_idx)

!+ Finds array index of I/O unit corresponding to the given file handle

!===================================================================
!----- IN ----------------------------------------------------------
! 'handle' = "handle" (nickname) of file in question
!----- OUT ---------------------------------------------------------
! 'IOunit_idx' = GrADS module internal array index corresponding to input
!                 file handle

use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_exception_mod, only: exception, FATAL

implicit none

!=== Argument declarations:

character(len=max_file_handle_length), intent(IN) :: handle 

integer, intent(OUT) :: IOunit_idx

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'find_IOunit'

integer :: iq

!=== Executable statements:

! Find I/O unit slot corresponding to given file handle:

iq = 1
do while (.NOT. gf_handle(iq) == handle)
   if (iq /= ngf) then
      iq = iq+1
   else
      call exception(procedure_name, FATAL,  &
           'Cannot identify the given file handle ( '//trim(handle)//' )')
   end if
end do

IOunit_idx = iq

end subroutine find_IOunit


!```````````````````````````````````````````````````````````````````
subroutine open_GrADS_dfile(dfn, n2, n3, action, handle)

!+ Opens a GrADS data file (for writing or reading)

!===================================================================
!----- IN ----------------------------------------------------------
! 'n2' = Number of gridpoints in X-direction
! 'n3' = Number of gridpoints in Y-direction
! 'dfn' = Name of GrADS datafile to open/close
! 'action' = String describing action to perform:
!             'R' = open file for reading
!             'W' = open file for writing
! 'handle' = "Handle" (nickname) of file to operate on

use univ_kind_defs_mod, only: f4
use univ_std_fileIO_mod, only: determine_IO_reclen
use univ_exception_mod, only: exception, WARN
use univ_IO_mem_status_mod, only: illegal_alloc_stop, for_allocation,  &
     for_deallocation
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

integer, intent(IN) :: n2, n3
character(len=*), intent(IN) :: dfn, action, handle

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'open_GrADS_dfile'

integer :: this_IOunit, IO_reclen, e
logical :: is_already_open
real(f4), allocatable, dimension(:,:) :: reclen_test

!=== Executable statements:

call find_IOunit(handle, this_IOunit)   ! Identify corresponding I/O unit

inquire(unit=io_unit(this_IOunit), opened=is_already_open)
if (is_already_open) then

   call exception(procedure_name, WARN, 'File handle ( '//  &
        handle(1:max_file_handle_length)//' ) is already open.')

else

   allocate(reclen_test(n2,n3), stat=e)
   if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name, 'A')
   IO_reclen = determine_IO_reclen(reclen_test)
   deallocate(reclen_test, stat=e)
   if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, 'A')

   if (action(1:1) == 'R') then    ! Open datafile for *reading*

      open(io_unit(this_IOunit), file=dfn(1:len_trim(dfn)), status='old',  &
           form='unformatted', access='direct', recl=IO_reclen)
      
   else if (action(1:1) == 'W') then    ! Open datafile for *writing*
        
      open(io_unit(this_IOunit), file=dfn(1:len_trim(dfn)), status='replace',  &
           form='unformatted', access='direct', recl=IO_reclen)

   end if

   file_is_open(this_IOunit) = .TRUE.
   current_rec(this_IOunit) = 1   ! Reset current record # for this file

end if

end subroutine open_GrADS_dfile


!```````````````````````````````````````````````````````````````````
subroutine close_GrADS_dfile(handle)

!+ Closes a GrADS data file

!===================================================================
!----- IN ----------------------------------------------------------
! 'handle' = "Handle" (nickname) of file to operate on

use univ_exception_mod, only: exception, WARN
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

character(len=*), intent(IN) :: handle

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'close_GrADS_dfile'

integer :: this_IOunit
logical :: open

!=== Executable statements:

call find_IOunit(handle, this_IOunit)   ! Identify corresponding I/O unit

inquire(unit=io_unit(this_IOunit), opened=open)
if (.NOT.open) then
   call exception(procedure_name, WARN, 'File handle ( '//  &
        handle(1:max_file_handle_length)//' ) is already closed.')
else   
   close(io_unit(this_IOunit))
end if
file_is_open(this_IOunit) = .FALSE.

end subroutine close_GrADS_dfile


!```````````````````````````````````````````````````````````````````
subroutine read_GrADS_dfile(n2, i2_beg, i2_end, n3, i3_beg, i3_end, n1,  &
     i1_beg, i1_end, tgt_it, handle, data, tgt_vname)

!+ Reads a variable from a GrADS data file

!===================================================================
!----- IN ----------------------------------------------------------
! 'n2' = Dimensions of 'data' array that will hold the output of this routine
! 'n3' =     "
! 'n1' =     "
! 'i2_beg' = Datafile direction_2 ("x") bounds for requested data section
! 'i2_end' =      "
! 'i3_beg' = Datafile direction_3 ("y") bounds for requested data section
! 'i3_end' =      "
! 'i1_beg' = Datafile direction_1 ("z") bounds for requested data section
! 'i1_end' =      "
! 'tgt_it' = Target snapshot ID #
! 'handle' = "Handle" (nickname) of file to operate on
! 'tgt_vname' = Name of variable to extract (must = var_name in *.ctl file)
!----- OUT -------------------------------------------------------
! 'data' = Array of values to be read from the data file

use univ_kind_defs_mod, only: f4
use univ_exception_mod, only: exception, WARN, FATAL
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_MESSAGE_LEN
use univ_IO_mem_status_mod, only: illegal_alloc_stop, for_allocation,  &
     for_deallocation

implicit none

!=== Argument declarations:

integer, intent(IN) :: n1, n2, n3, tgt_it, i2_beg, i2_end, i3_beg, i3_end,  &
     i1_beg, i1_end
character(len=*), intent(IN) :: handle, tgt_vname 

real(f4), dimension(n2,n3,n1), intent(INOUT) :: data

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'read_GrADS_dfile'

integer :: this_file, rcnt, iv, nc2, tgt_iv, i1, i2, i3, nc1, e
real(f4), allocatable, dimension(:,:) :: rec_slab
character(len=MAX_MESSAGE_LEN) :: msg

!=== Executable statements:

call find_IOunit(handle, this_file)   ! Identify corresponding I/O unit

nc1 = len_trim(tgt_vname)
if (nc1 > max_GrADSc_vname_len) then
   nc1 = max_GrADSc_vname_len
   write (msg, '(a,i2,a)') 'GrADS varname field is > ',  &
        max_GrADSc_vname_len, ' chars long, and will be truncated.'
   call exception(procedure_name, WARN, msg)
end if

iv = 1
nc2 = len_trim(rdfvar_name(this_file)%ptr(iv))
do while (.NOT. rdfvar_name(this_file)%ptr(iv)(1:nc2) == tgt_vname(1:nc1))
   if (iv /= rdf_nv(this_file)) then
      iv = iv+1
      nc2 = len_trim(rdfvar_name(this_file)%ptr(iv))
   else
      call exception(procedure_name, FATAL,  &
           'Cannot identify the given var_name ( '//tgt_vname(1:nc1)//' )')
   end if
end do
tgt_iv = iv

rcnt = (tgt_it-1)*rdf_nt_zrec(this_file)+1
do iv=1,tgt_iv-1
   rcnt = rcnt+rdfvar_nz(this_file)%ptr(iv)
end do
rcnt = rcnt+(i1_beg-1)

allocate(rec_slab(rdf_n2(this_file),rdf_n3(this_file)), stat=e)
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name, 'A')

do i1=1,i1_end-i1_beg+1
   read (io_unit(this_file), rec=rcnt) rec_slab

   do i3=1,i3_end-i3_beg+1
      do i2=1,i2_end-i2_beg+1
         data(i2,i3,i1) = rec_slab(i2+(i2_beg-1),i3+(i3_beg-1)) 
      end do
   end do

   rcnt = rcnt+1
end do

deallocate(rec_slab, stat=e)
if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, 'A')

end subroutine read_GrADS_dfile


!```````````````````````````````````````````````````````````````````
subroutine write_GrADS_dfile(n2, n3, n1, data, handle)

!+ Writes a variable to a GrADS data file

!===================================================================
!----- IN ----------------------------------------------------------
! 'n2' = Number of gridpoints in X-direction
! 'n3' = Number of gridpoints in Y-direction
! 'n1' = Number of gridpoints in Z-direction
! 'handle' = "Handle" (nickname) of file to operate on
! 'data' = Array of values to be written to the data file

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: n1, n2, n3
character(len=*), intent(IN) :: handle 
real(f4), dimension(n2,n3,n1), intent(IN) :: data

!== Local declarations:

integer :: this_IOunit, ir, rcnt

!=== Executable statements:

call find_IOunit(handle, this_IOunit)   ! Identify corresponding I/O unit

rcnt = current_rec(this_IOunit)    ! Start at current file record
     
do ir=1,n1
   write (io_unit(this_IOunit), rec=rcnt) data(:,:,ir)
   rcnt = rcnt+1
end do

current_rec(this_IOunit) = rcnt    ! Set current record counter
   
end subroutine write_GrADS_dfile


!```````````````````````````````````````````````````````````````````
subroutine read_GrADS_hfile(dfn, hfn, handle, maxn1, n2, n3, nt, nv, undef,  &
     which_endian, title,  &
     gzlev, gvarn, gvlevs, gvdesc, gvunits)    ! Optional args

!+ Opens, reads, then closes a classic GrADS header file (of the style written
!   by `write_GrADS_hfile`).

!===================================================================
!----- IN ----------------------------------------------------------
! 'hfn' = Filename of GrADS header file to be read/written
! 'handle' = "Handle" (nickname) of file to operate on
!----- OUT ---------------------------------------------------------
! 'dfn' = Filename of associated GrADS datafile
! 'maxn1' = Maximum number of gridpoints in Z-direction
! 'n2' = Number of gridpoints in X-direction
! 'n3' = Number of gridpoints in Y-direction
! 'nt' = Number of time "gridpoints"
! 'nv' = Number of variables in the datafile
! 'undef' = Floating point value that will be ignored by plotting routines
! 'which_endian' = Type of machine datafile is being created on:
!                       0 = "little_endian", 1 = "big_endian"
! 'title' = Brief, informative description of data in datafile
!----- OUT (optional) ----------------------------------------------
! 'gzlev' = Heights of each vertical level
! 'gvarn' = Names of each variable in datafile  
! 'gvlevs' = Number of vertical levels each variable field has
! 'gvdesc' = Brief description of each variable
! 'gvunits' = Units of each variable quantity

use univ_kind_defs_mod, only: f4
use univ_std_fileIO_mod, only: find_file_IOunit
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_IO_mem_status_mod, only: illegal_alloc_stop, for_allocation

implicit none

!=== Argument declarations:

character(len=*), intent(IN) :: hfn, handle

integer, intent(OUT) :: maxn1, n2, n3, nt, nv, which_endian
character(len=*), intent(INOUT) :: dfn, title
real(f4), intent(OUT) :: undef

real(f4), dimension(:), optional, intent(INOUT) :: gzlev   ! {maxn1}
integer, dimension(:), optional, intent(INOUT) :: gvlevs   ! {nv}
character(len=*), optional, dimension(:), intent(INOUT) :: gvarn,  &
     gvdesc, gvunits    ! {nv}

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'read_GrADS_hfile'

integer :: this_IOunit, iv, this_file, nc1, gvlevs_tmp, e, nc2
character(len=80) :: junkc, junkc1
character(len=max_GrADSc_vname_len) :: gvarn_tmp
character(len=max_GrADSc_vdesc_len) :: gvdesc_tmp
character(len=max_GrADSc_vname_len*2+max_GrADSc_vdesc_len) :: vline

!=== Executable statements:

this_IOunit = find_file_IOunit()

open(this_IOunit, file=hfn(1:len_trim(hfn)), status='old', form='formatted')

read (this_IOunit, '(a4,1x,a)') junkc, dfn
read (this_IOunit, *) junkc, undef

read (this_IOunit, *) junkc, junkc1
if (junkc1(1:13) == 'little_endian') then
   which_endian = 0
else if (junkc1(1:10) == 'big_endian') then
   which_endian = 1
end if
   
read (this_IOunit, *) junkc, title
     
read (this_IOunit, *) junkc, n2
read (this_IOunit, *) junkc, n3

read (this_IOunit, *) junkc, maxn1, junkc1    ! zdef line scan

if (present(gzlev)) then
   if (maxn1 >= 2) then
      read (this_IOunit, '(5(f12.3))') gzlev(1:maxn1)
   else
      backspace this_IOunit
      read (this_IOunit, *) junkc, maxn1, junkc1, gzlev(maxn1)  ! zdef line read
   end if
else    ! Skip ahead to 'tdef'
   do
      read (this_IOunit, *) junkc
      if (junkc(1:4) == 'tdef') then
         backspace this_IOunit; exit
      end if
   end do
end if

read (this_IOunit, *) junkc, nt
read (this_IOunit, *) junkc, nv

call find_IOunit(handle, this_file)

rdf_n2(this_file) = n2; rdf_n3(this_file) = n3
rdf_nv(this_file) = nv; rdf_nt(this_file) = nt

if (.NOT.associated(rdfvar_nz(this_file)%ptr) .and.  &
     .NOT.associated(rdfvar_name(this_file)%ptr)) then
   allocate(rdfvar_nz(this_file)%ptr(nv), rdfvar_name(this_file)%ptr(nv),  &
        stat=e)
   if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name, 'A')
end if

rdf_nt_zrec(this_file) = 0
do iv=1,nv
   read (this_IOunit, '(a)') vline
   nc1 = index(vline, '['); nc2 = index(vline, ']')

   read (vline(1:nc1-1), *) gvarn_tmp, gvlevs_tmp
   rdfvar_nz(this_file)%ptr(iv) = max(1, gvlevs_tmp)
   rdfvar_name(this_file)%ptr(iv) = gvarn_tmp

   read (vline(nc2+2:), '(a)') gvdesc_tmp   ! Var desc (skip preceding space)

   if (present(gvarn) .and. present(gvlevs) .and. present(gvdesc) .and.  &
        present(gvunits)) then
      gvunits(iv) = vline(nc1+1:nc2-1)
      gvarn(iv) = gvarn_tmp
      gvlevs(iv) = gvlevs_tmp
      gvdesc(iv) = trim(gvdesc_tmp)
   end if
   rdf_nt_zrec(this_file) = rdf_nt_zrec(this_file)+max(1, gvlevs_tmp)
end do

close(this_IOunit)   ! Close header file

end subroutine read_GrADS_hfile


!```````````````````````````````````````````````````````````````````
subroutine write_GrADS_hfile(dfn, hfn, n2, n3, nt, undef, which_endian,  &
     title, gzlev, fbeg_ET, tinc, start_UT, gvarn, gvlevs, gvdesc, gvunits,  &
     var_in_file, diurnal_period)

!+ Opens, writes, then closes a GrADS header file

!===================================================================
!----- IN ----------------------------------------------------------
! 'dfn' = Filename of associated GrADS datafile
! 'hfn' = Filename of GrADS header file to be read/written
! 'n2' = Number of gridpoints in X-direction
! 'n3' = Number of gridpoints in Y-direction
! 'nt' = Number of time "gridpoints"
! 'undef' = Floating point value that will be ignored by plotting routines
! 'which_endian' = Type of machine datafile is being created on:
!                       0 = "little_endian", 1 = "big_endian"
! 'title' = Brief, informative description of data in datafile
! 'maxn1' = Maximum number of gridpoints in Z-direction
! 'gzlev' = Heights of each vertical level
! 'fbeg_ET' = ideally, model elapsed time [s] of first data in GrADS datafile
! 'tinc' = Time increment between times in datafile [s] 
! 'start_UT' = Mars UT (time at 0deg longitude) of first data in datafile
! 'gvarn' = Names of each variable in datafile  
! 'gvlevs' = Number of vertical levels each variable field has
! 'gvdesc' = Brief description of each variable
! 'gvunits' = Units of each variable quantity
! 'var_in_file' = Logical array indicating whether a given variable
!                  belongs in this file
! 'diurnal period' = "Diurnal period" of planetary body [s]

use univ_kind_defs_mod, only: f4, f8
use univ_std_fileIO_mod, only: find_file_IOunit
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_MESSAGE_LEN
use univ_exception_mod, only: exception, WARN

implicit none

!=== Argument declarations:

character(len=*), intent(IN) :: hfn, dfn, title
real(f4), intent(IN) :: undef, start_UT
real(f8), intent(IN) :: diurnal_period, fbeg_ET, tinc
integer, intent(IN) :: n2, n3, nt, which_endian
logical, dimension(:), intent(INOUT) :: var_in_file   ! {maxnv}
real(f4), dimension(:), intent(INOUT) :: gzlev     ! {maxn1}
integer, dimension(:), intent(INOUT) :: gvlevs   ! {maxnv}
character(len=*), dimension(:), intent(INOUT) :: gvarn,  &
     gvdesc, gvunits    ! {maxnv}

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'write_GrADS_hfile'

integer :: this_IOunit, iv, nc1, nc2, nc3, maxn1, maxnv
integer, dimension(1) :: tmp
character(len=MAX_MESSAGE_LEN) :: msg

!=== Executable statements:

this_IOunit = find_file_IOunit()

open(this_IOunit, file=hfn(1:len_trim(hfn)), status='replace', form='formatted')

write (this_IOunit, '(a,1x,a)') 'dset', dfn(1:len_trim(dfn))
write (this_IOunit, '(a,1x,e12.3)') 'undef', undef

if (which_endian == 0) then
   write (this_IOunit, '(a,1x,a)') 'options', 'little_endian'
else
   write (this_IOunit, '(a,1x,a)') 'options', 'big_endian'
end if
   
write (this_IOunit, '(a,1x,a)') 'title', title(1:len_trim(title))
     
write (this_IOunit, '(a,1x,i4,1x,a,1x,2f15.7)') 'xdef', n2, 'linear', 1._f4,  &
     1._f4
write (this_IOunit, '(a,1x,i4,1x,a,1x,2f15.7)') 'ydef', n3, 'linear', 1._f4,  &
     1._f4

tmp = maxval(gvlevs, var_in_file)  ! Find max # of levels of any variable
maxn1 = max(1, tmp(1))             !
if (maxn1 >= 2) then
   write (this_IOunit, '(a,1x,i4,1x,a,1x)') 'zdef', maxn1, 'levels'
   write (this_IOunit, '(5(f12.3))') gzlev(1:maxn1)
else
   write (this_IOunit, '(a,1x,i4,1x,a,1x,2(f12.3))') 'zdef', maxn1, 'linear',  &
        gzlev(maxn1), 1._f4
end if

! Create tdef string:
call det_GrADS_timedate(fbeg_ET, tinc, start_UT, diurnal_period, nt)

write (this_IOunit, '(a)') trim(gtdef_str)

maxnv = size(var_in_file)

tmp = count(var_in_file(1:maxnv))
write (this_IOunit, '(a,1x,i4)') 'vars', tmp(1)

do iv=1,maxnv
   if (var_in_file(iv)) then
      nc1 = len_trim(gvarn(iv))
      if (nc1 > max_GrADSc_vname_len) then
         nc1 = max_GrADSc_vname_len
         write (msg, '(a,i2,a)') 'GrADS varname field is > ',  &
              max_GrADSc_vname_len, ' chars long, and will be truncated.'
         call exception(procedure_name, WARN, msg)
      end if
      nc2 = index(gvunits(iv),';')-1; nc3 = index(gvdesc(iv),';')-1
      if (nc2+nc3+3 > max_GrADSc_vdesc_len) then
         nc3 = max(1, max_GrADSc_vdesc_len-(nc2+3))
         write (msg, '(a,i3,a)') 'GrADS var description/units field is > ',  &
              max_GrADSc_vdesc_len, ' chars long, and will be truncated.'
         call exception('find_IOunit', WARN, msg)
      end if
      write (this_IOunit, '(a,1x,i4,1x,i2,1x,a1,a,a1,1x,a)')  &
           gvarn(iv)(1:nc1), gvlevs(iv), 99, '[', gvunits(iv)(1:nc2), ']',  &
           gvdesc(iv)(1:nc3)
   end if
end do
     
write (this_IOunit, '(a)') 'endvars'

close(this_IOunit)   ! Close header file

end subroutine write_GrADS_hfile


!```````````````````````````````````````````````````````````````````
subroutine det_GrADS_timedate(fcstsec, tinc_sec, start_ut, diurnal_period, nt)

!+ Creates GrADS time descriptor strings

!===================================================================
!----- IN ----------------------------------------------------------
! 'fcstsec' = Ideally, time (elapsed since sim start) [s] of first datachunk
!              in GrADS datafile
! 'tinc_sec' = Time [s] between each pair of datachunks in GrADS datafile
! 'start_UT' = UT (time at 0deg longitude) of first data in GrADS
!               datafile
! 'diurnal period' = "Diurnal period" of planetary body [s]

use univ_kind_defs_mod, only: f4, f8

implicit none

!=== Argument declarations:

integer, intent(IN) :: nt
real(f4), intent(IN) :: start_ut
real(f8), intent(IN) :: diurnal_period, fcstsec, tinc_sec

!== Local declarations:

integer, parameter :: i_hours_per_dp = 24,  &
                      i_minutes_per_hour = 60,  &
                      i_minutes_per_dp = i_hours_per_dp*i_minutes_per_hour

real(f8), parameter :: hours_per_dp = i_hours_per_dp,  &
                       minutes_per_hour =  i_minutes_per_hour,  &
                       minutes_per_dp = i_minutes_per_dp

real(f8) :: t_dp, t_hour, t_min
real(f4) :: UT
integer :: ihh, imin, i
character(len=2) :: t_unit_str
character(len=4) :: t_inc_str

!=== Executable statements:

t_dp = diurnal_period   ! body-day [s]
t_hour = t_dp/hours_per_dp     ! body-hour [s]
t_min = t_dp/minutes_per_dp   ! body-minute [s]

i = max(1, nint(tinc_sec/t_min)); t_unit_str = 'mn'
if (i > 99) then
   i = max(1, nint(tinc_sec/t_hour))
   t_unit_str = 'hr'
   if (i >= i_hours_per_dp) then
      i = max(1, nint(tinc_sec/t_dp))
      t_unit_str = 'dy'
   end if
end if
write (t_inc_str, '(i2.2,a2)') i, t_unit_str    ! Approximate time interval

UT = real(fcstsec/t_hour, kind=f4)+start_ut   ! Fractional body UT [body-hours]
ihh = int(UT)
imin = int((UT-real(ihh, kind=f4))*real(minutes_per_hour, kind=f4))
do while (imin >= i_minutes_per_hour) 
   imin = imin-i_minutes_per_hour
   ihh = ihh+1
end do
do while (ihh >= i_hours_per_dp) 
   ihh = ihh-i_hours_per_dp
end do

write (gtdef_str, '(a,i5,a,i2.2,a1,i2.2,2a)') 'TDEF ', nt, ' linear ',  &
     ihh, ':', imin, 'Z04jul1997 ', t_inc_str

end subroutine det_GrADS_timedate


end module univ_GrADS_format_mod
