!```````````````````````````````````````````````````````````````````
module univ_OpenDX_format_mod

!+ Module that contains subroutines, etc. for writing/reading OpenDX format
!   files

use univ_kind_defs_mod, only: f4

implicit none

private

! Data file object type (in Cartesian space; encompasses everything needed to
!  write data and header files):

type :: OpenDX_C_dfobj

   ! Data dimensions (3 spatial dimensions + time):
   integer :: nx, ny, nz, nt

   ! Number of items (data values) per data object (per timestep):
   integer :: nspatial_items

   ! Byte offsets for data files:
   integer, pointer, dimension(:) :: datf_Boffset => NULL()    ! {nt}

   ! Header file object ID#s:
   integer :: pos_objID, con_objID
   integer, pointer, dimension(:) :: dat_objID => NULL(),  &
        fld_objID => NULL(), ser_objID => NULL()  ! {nt}

   ! Timeseries "position" tags:
   real(f4), pointer, dimension(:) :: dat_objtag => NULL()    ! {nt}

   ! Filenames of the data and header files:
   character, pointer, dimension(:) :: dat_fn => NULL(), hdr_fn => NULL() ! {?}
   integer :: hdr_fn_len

   ! Corresponding positions object ID#:
   integer :: pof_ID
   
   ! Descriptive info about the corresponding data field:
   character, pointer, dimension(:) :: field_name => NULL(),  &
        field_desc => NULL(), field_units => NULL()  ! {?}

   ! Miscellaneous variables:
   character(len=3) :: endian
   logical :: data_is_binary, data_is_vector
   integer :: IO_unit, shape

end type OpenDX_C_dfobj

! Position file object type:

type :: OpenDX_posfobj

   ! Position set ID tag:
   integer :: set_ID

   ! Position file output filename:
   character, pointer, dimension(:) :: pos_fn => NULL()

   ! Position dimensions (3 spatial dimensions):
   integer :: nx, ny, nz

   ! Number of items (data values) per position object:
   integer :: nspatial_items

   ! Byte offset for position file data:
   integer :: posf_Boffset

   ! Miscellaneous variables:
   character(len=3) :: endian
   logical :: data_is_binary
   integer :: IO_unit

end type OpenDX_posfobj


! Global variables:

integer :: noutputf
type(OpenDX_C_dfobj), allocatable, dimension(:) :: odfobj   ! {noutputf}

integer :: nposf,  &  ! Number of separate position sets (1 per file)
     curr_pofID = 0
type(OpenDX_posfobj), allocatable, dimension(:) :: pofobj     ! {nposf}

real(f4), allocatable, dimension(:,:) :: pos

real(f4), allocatable, dimension(:,:) :: vect


public :: OpenDX_init_mod, OpenDX_end_mod, OpenDX_init_pofobj,  &
     OpenDX_chk_pofobj, OpenDX_set_pofobj, OpenDX_C_Wpos,  &
     OpenDX_C_init_dfobj, OpenDX_Csclr_Wdat, OpenDX_end_dfobj,  &
     OpenDX_Csclr_Whdr, OpenDX_Cvect_Whdr, OpenDX_Cvect_Wdat, OpenDX_get_endian

contains


!```````````````````````````````````````````````````````````````````
subroutine OpenDX_init_mod(nof, npf)

!+ Initialize/set/allocate global module memory, etc.

use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_IO_mem_status_mod, only: for_allocation, illegal_alloc_stop

implicit none

!=== Argument declarations:

integer, intent(IN) :: nof, npf

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'OpenDX_init_mod'

integer :: i, aerr

!=== Executable statements:

! Set some dimensions:

noutputf = nof
nposf = npf

! Allocate position and datafile object arrays:

allocate(odfobj(nof), pofobj(npf), stat=aerr)
if (aerr /= 0) call illegal_alloc_stop(for_allocation, procedure_name, '1')

! Nullify all pointers in derived-type vars that were just allocated:

do i=1,nof
   nullify(odfobj(i)%datf_Boffset)
   nullify(odfobj(i)%dat_objID)
   nullify(odfobj(i)%fld_objID)
   nullify(odfobj(i)%dat_objtag)
   nullify(odfobj(i)%dat_fn)
   nullify(odfobj(i)%hdr_fn)
   nullify(odfobj(i)%field_name)
   nullify(odfobj(i)%field_desc)
   nullify(odfobj(i)%field_units)
end do

do i=1,npf
   nullify(pofobj(i)%pos_fn)
end do

end subroutine OpenDX_init_mod


!```````````````````````````````````````````````````````````````````
subroutine OpenDX_end_mod

!+ Deallocate global module memory, etc.

use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_IO_mem_status_mod, only: for_deallocation, illegal_alloc_stop

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'OpenDX_end_mod'

integer :: aerr

!=== Executable statements:

! Deallocate position and datafile object arrays:

deallocate(odfobj, pofobj, stat=aerr)
if (aerr /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, '1')

if (allocated(vect)) then
   deallocate(vect, stat=aerr)
   if (aerr /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, '2')
end if

end subroutine OpenDX_end_mod


!```````````````````````````````````````````````````````````````````
subroutine OpenDX_init_pofobj

!+ (Re)initialize parameters of (all) position file object(s)

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'OpenDX_init_pofobj'

integer :: ID

!=== Executable statements:

do ID=1,nposf
   pofobj(ID)%set_ID = 0
   pofobj(ID)%IO_unit = 0
   if (associated(pofobj(ID)%pos_fn)) then
      deallocate(pofobj(ID)%pos_fn)
   end if
end do

curr_pofID = 0     ! Init counter

end subroutine OpenDX_init_pofobj


!```````````````````````````````````````````````````````````````````
subroutine OpenDX_chk_pofobj(vtype_ID, pof_not_set, cpof_ID)

!+ Check whether a given position file object has been set

implicit none

!=== Argument declarations:

integer, intent(IN) :: vtype_ID

logical, intent(OUT) :: pof_not_set
integer, intent(OUT) :: cpof_ID

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'OpenDX_chk_pofobj'

integer :: ID

!=== Executable statements:

pof_not_set = .TRUE.
do ID=1,nposf
   if (pofobj(ID)%set_ID == vtype_ID) then
      pof_not_set = .FALSE.
      cpof_ID = ID
      return   ! Nothing left to do
   end if
end do

curr_pofID = curr_pofID+1
cpof_ID = curr_pofID

end subroutine OpenDX_chk_pofobj


!```````````````````````````````````````````````````````````````````
subroutine OpenDX_set_pofobj(vtype_ID, mx, my, mz, pfn, iu)

!+ Set position file object parameters

use univ_kind_defs_mod, only: f4
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_IO_mem_status_mod, only: for_allocation, illegal_alloc_stop,  &
     for_deallocation
use univ_std_fileIO_mod, only: determine_IO_reclen

implicit none

!=== Argument declarations:

integer, intent(IN) :: mx, my, mz, vtype_ID, iu
character(len=*), intent(IN) :: pfn

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'OpenDX_set_pofobj'

integer :: c1, ID, ic, e, IO_reclen
real(f4), allocatable, dimension(:,:) :: reclen_test

!=== Executable statements:

ID = curr_POFID

! Set IO unit ID:
pofobj(ID)%IO_unit = iu

! Set parameters:
pofobj(ID)%set_ID = vtype_ID
pofobj(ID)%nx = mx
pofobj(ID)%ny = my
pofobj(ID)%nz = mz
pofobj(ID)%nspatial_items = mx*my*mz

! Miscellaneous settings:
pofobj(ID)%data_is_binary = .TRUE.

! MUST be done before opening file 'iu':
call OpenDX_get_endian(pofobj(ID)%endian, iu)

pofobj(ID)%posf_Boffset = 0

allocate(reclen_test(pofobj(ID)%nspatial_items,3), stat=e)
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name, '0')
IO_reclen = determine_IO_reclen(reclen_test)  ! 'recl' per position object
deallocate(reclen_test, stat=e)
if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, '0')

! Associated filenames:
c1 = len_trim(pfn)
allocate(pofobj(ID)%pos_fn(c1), stat=e)
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name, '1')
do ic=1,c1
   pofobj(ID)%pos_fn(ic) = pfn(ic:ic) 
enddo

open(pofobj(ID)%IO_unit, file=pfn(1:c1), status='unknown',  &
     form='unformatted', access='direct', recl=IO_reclen)  ! Open position file

end subroutine OpenDX_set_pofobj


!```````````````````````````````````````````````````````````````````
subroutine OpenDX_C_Wpos(cpof_ID, mx, my, mz, pos_x, pos_y, pos_z)

!+ Writes Cartesian (i.e., x,y,z) position elements to a file

use univ_kind_defs_mod, only: f4
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_IO_mem_status_mod, only: for_allocation, for_deallocation,  &
     illegal_alloc_stop

implicit none

!=== Argument declarations:

integer, intent(IN) :: mx, my, mz, cpof_ID
real(f4), dimension(mx,my,mz), intent(IN) :: pos_x, pos_y, pos_z

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'OpenDX_C_Wpos'

integer :: count, i, j, k, iu, aerr

!=== Executable statements:

allocate(pos(3,mx*my*mz), stat=aerr)   ! Allocate memory
if (aerr /= 0) call illegal_alloc_stop(for_allocation, procedure_name, '1')

! Fill 'pos' array (NOTE: input data is rearranged into row-major order): 

count = 0
do i=1,mx
   do j=1,my
      do k=1,mz
         count = count+1
         pos(1,count) = pos_x(i,j,k)
         pos(2,count) = pos_y(i,j,k)
         pos(3,count) = pos_z(i,j,k)
      end do
   end do
end do

iu = pofobj(cpof_ID)%IO_unit

write (iu, rec=1) pos    ! Write positions

close(iu)   ! Close positions file

deallocate(pos, stat=aerr)   ! Deallocate work array
if (aerr /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, '1')

end subroutine OpenDX_C_Wpos


!```````````````````````````````````````````````````````````````````
subroutine OpenDX_C_init_dfobj(ID, mx, my, mz, mt, dfn, hfn, fld_name,  &
     fld_desc, fld_units, cpof_ID, time_tags, iu, ishape)

!+ Initialize/set datafile object (Cartesian) parameters

use univ_kind_defs_mod, only: f4, f4_B
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_IO_mem_status_mod, only: for_allocation, illegal_alloc_stop,  &
     for_deallocation
use univ_std_fileIO_mod, only: determine_IO_reclen

implicit none

!=== Argument declarations:

integer, intent(IN) :: mx, my, mz, mt, ID, cpof_ID, iu, ishape
character(len=*), intent(IN) :: dfn, hfn, fld_name, fld_desc, fld_units
real(f4), dimension(mt), intent(IN) :: time_tags

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'OpenDX_C_init_dfobj'

integer :: it, ic, c1, c2, c3, Bperdt, Bcnt, e, IO_reclen
real(f4), allocatable, dimension(:,:) :: reclen_test

!=== Executable statements:

! Set IO unit ID:
odfobj(ID)%IO_unit = iu

! Set dimensions:
odfobj(ID)%nx = mx
odfobj(ID)%ny = my
odfobj(ID)%nz = mz
odfobj(ID)%nt = mt

odfobj(ID)%data_is_vector = (ishape > 0)
odfobj(ID)%shape = ishape
odfobj(ID)%nspatial_items = mx*my*mz

! Miscellaneous settings:
odfobj(ID)%data_is_binary = .TRUE.

! MUST be done before opening file 'iu':
call OpenDX_get_endian(odfobj(ID)%endian, iu)

odfobj(ID)%pof_ID = cpof_ID

allocate(odfobj(ID)%datf_Boffset(mt), stat=e)
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name, '1')

allocate(reclen_test(odfobj(ID)%nspatial_items,max(ishape, 1)), stat=e)
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name, '0')
IO_reclen = determine_IO_reclen(reclen_test)  ! 'recl' per position object
deallocate(reclen_test, stat=e)
if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, '0')

Bperdt = f4_B*odfobj(ID)%nspatial_items*max(ishape, 1)
Bcnt = 0
do it=1,mt
   odfobj(ID)%datf_Boffset(it) = Bcnt
   Bcnt = Bcnt+Bperdt
end do

! Header file object ID#s:
odfobj(ID)%pos_objID = 1
odfobj(ID)%con_objID = 2
allocate(odfobj(ID)%dat_objID(mt), odfobj(ID)%fld_objID(mt),  &
     odfobj(ID)%ser_objID(mt), odfobj(ID)%dat_objtag(mt), stat=e)
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name, '2')
do it=1,mt
   odfobj(ID)%dat_objID(it) = it+2 
   odfobj(ID)%fld_objID(it) = it+mt+2
   odfobj(ID)%ser_objID(it) = it-1
   odfobj(ID)%dat_objtag(it) = time_tags(it)
end do

! Associated filenames:
c1 = len_trim(dfn)
c2 = len_trim(hfn)
allocate(odfobj(ID)%dat_fn(c1), odfobj(ID)%hdr_fn(c2), stat=e)
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name, '3')
do ic=1,c1
   odfobj(ID)%dat_fn(ic) = dfn(ic:ic) 
end do
do ic=1,c2
   odfobj(ID)%hdr_fn(ic) = hfn(ic:ic) 
end do
odfobj(ID)%hdr_fn_len = c2

open(odfobj(ID)%IO_unit, file=dfn(1:c1), status='unknown',  &
     form='unformatted', access='direct', recl=IO_reclen)    ! Open data file

! Associated field info:
c1 = len_trim(fld_name)
c2 = len_trim(fld_desc)
c3 = len_trim(fld_units)
allocate(odfobj(ID)%field_name(c1), odfobj(ID)%field_desc(c2),  &
     odfobj(ID)%field_units(c3), stat=e)
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name, '4')
do ic=1,c1
   odfobj(ID)%field_name(ic) = fld_name(ic:ic) 
end do
do ic=1,c2
   odfobj(ID)%field_desc(ic) = fld_desc(ic:ic) 
end do
do ic=1,c3
   odfobj(ID)%field_units(ic) = fld_units(ic:ic) 
end do

end subroutine OpenDX_C_init_dfobj


!```````````````````````````````````````````````````````````````````
subroutine OpenDX_Csclr_Wdat(dfobj_ID, mx, my, mz, data, recn)

!+ Writes Cartesian (i.e., x,y,z) scalar data elements to a file

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: mx, my, mz, dfobj_ID, recn
real(f4), dimension(mx,my,mz), intent(IN) :: data

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'OpenDX_Csclr_Wdat'

integer :: iu, i, j, k

!=== Executable statements:

iu = odfobj(dfobj_ID)%IO_unit
write (iu, rec=recn) (((data(i,j,k), k=1,mz), j=1,my), i=1,mx)

end subroutine OpenDX_Csclr_Wdat


!```````````````````````````````````````````````````````````````````
subroutine OpenDX_end_dfobj

!+ Close data file(s); deallocate datafile object memory, etc.

use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_IO_mem_status_mod, only: for_deallocation, illegal_alloc_stop

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'OpenDX_end_dfobj'

integer :: ID, aerr

!=== Executable statements:

do ID=1,noutputf

   close(odfobj(ID)%IO_unit)   ! Close output datafile

   deallocate(odfobj(ID)%datf_Boffset, stat=aerr)
   if (aerr /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, '1')
   deallocate(odfobj(ID)%dat_objID, odfobj(ID)%fld_objID,  &
        odfobj(ID)%ser_objID, odfobj(ID)%dat_objtag, stat=aerr)
   if (aerr /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, '2')
   deallocate(odfobj(ID)%dat_fn, odfobj(ID)%hdr_fn, stat=aerr)
   if (aerr /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, '3')
   deallocate(odfobj(ID)%field_name, odfobj(ID)%field_desc,  &
        odfobj(ID)%field_units, stat=aerr)
   if (aerr /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, '4')

end do

end subroutine OpenDX_end_dfobj


!```````````````````````````````````````````````````````````````````
subroutine OpenDX_Csclr_Whdr(dfobj_ID, iu)

!+ Writes Cartesian (i.e., x,y,z) scalar data descriptor ('*.dx') file

use univ_exception_mod, only: exception, FATAL
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_FILE_PATH_LEN

implicit none

!=== Argument declarations:

integer, intent(IN) :: dfobj_ID, iu

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'OpenDX_Csclr_Whdr'

integer :: it, ic, datID, posID
character(len=MAX_FILE_PATH_LEN) :: ctmp

!=== Executable statements:

datID = dfobj_ID
posID = odfobj(datID)%pof_ID

if (odfobj(datID)%hdr_fn_len > MAX_FILE_PATH_LEN) call exception(  &
     procedure_name, FATAL, 'OpenDX filepath name is too long.')
ctmp = ';'
do ic=1,odfobj(datID)%hdr_fn_len
   ctmp(ic:ic) = odfobj(datID)%hdr_fn(ic)
end do
open(iu, file=ctmp, form='formatted', status='unknown')

! Object 1: positions

write (iu, *) 'object ', odfobj(datID)%pos_objID,  &
     ' class array type float rank 1 shape 3 items ',  &
     pofobj(posID)%nspatial_items, pofobj(posID)%endian, ' binary'
write (iu, *) 'data file ', pofobj(posID)%pos_fn, ', ',  &
     pofobj(posID)%posf_Boffset
write (iu, *)

! Object 2: connections

write (iu, *) 'object ', odfobj(datID)%con_objID,  &
     ' class gridconnections counts ', odfobj(datID)%nx,odfobj(datID)%ny,  &
     odfobj(datID)%nz
write (iu, *)

do it=1,odfobj(datID)%nt

   ! Object ?: data array

   write (iu, *) 'object ', odfobj(datID)%dat_objID(it),  &
        ' class array type float rank 0 items ',  &
        odfobj(datID)%nspatial_items, odfobj(datID)%endian, ' binary'
   write (iu, *) 'data file ', odfobj(datID)%dat_fn, ', ',  &
        odfobj(datID)%datf_Boffset(it)
   write (iu, *) 'attribute "dep" string "positions"'
   write (iu, *)

end do

do it=1,odfobj(datID)%nt

   ! Object ?: data field

   write (iu, *) 'object ', odfobj(datID)%fld_objID(it), ' class field'
   write (iu, *) 'component "positions" value ', odfobj(datID)%pos_objID
   write (iu, *) 'component "connections" value ', odfobj(datID)%con_objID
   write (iu, *) 'component "data" value ', odfobj(datID)%dat_objID(it)
   write (iu, *)

end do

write (iu, *) 'object "', odfobj(datID)%field_name, '" class series'
do it=1,odfobj(datID)%nt
   write (iu, *) 'member ', it-1, ' value ', odfobj(datID)%fld_objID(it),  &
        ' position ', odfobj(datID)%dat_objtag(it)
enddo
write (iu, *)

write (iu, *) 'end'

close(iu)

end subroutine OpenDX_Csclr_Whdr


!```````````````````````````````````````````````````````````````````
subroutine OpenDX_Cvect_Whdr(dfobj_ID, iu)

!+ Writes Cartesian (i.e., x,y,z) vector data descriptor ('*.dx') file

use univ_exception_mod, only: exception, FATAL
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_FILE_PATH_LEN

implicit none

!=== Argument declarations:

integer, intent(IN) :: dfobj_ID, iu

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'OpenDX_Cvect_Whdr'

integer :: it, ic, datID, posID
character(len=MAX_FILE_PATH_LEN) :: ctmp

!=== Executable statements:

datID = dfobj_ID
posID = odfobj(datID)%pof_ID

if (odfobj(datID)%hdr_fn_len > MAX_FILE_PATH_LEN) call exception(  &
     procedure_name, FATAL, 'OpenDX filepath name is too long.')
ctmp = ';'
do ic=1,odfobj(datID)%hdr_fn_len
   ctmp(ic:ic) = odfobj(datID)%hdr_fn(ic)
end do
open(iu, file=ctmp, form='formatted', status='unknown')

! Object 1: positions

write (iu, *) 'object ', odfobj(datID)%pos_objID,  &
     ' class array type float rank 1 shape 3 items ',  &
     pofobj(posID)%nspatial_items, pofobj(posID)%endian, ' binary'
write (iu, *) 'data file ', pofobj(posID)%pos_fn, ', ',  &
     pofobj(posID)%posf_Boffset
write (iu, *)

! Object 2: connections

write (iu, *) 'object ', odfobj(datID)%con_objID,  &
     ' class gridconnections counts ', odfobj(datID)%nx, odfobj(datID)%ny,  &
     odfobj(datID)%nz
write (iu, *)

do it=1,odfobj(datID)%nt

   ! Object ?: data array

   write (iu, *) 'object ', odfobj(datID)%dat_objID(it),  &
        ' class array type float rank 1 shape ', odfobj(datID)%shape,  &
        ' items ',  &
        odfobj(datID)%nspatial_items, odfobj(datID)%endian, ' binary'
   write (iu, *) 'data file ', odfobj(datID)%dat_fn, ', ',  &
        odfobj(datID)%datf_Boffset(it)
   write (iu, *) 'attribute "dep" string "positions"'
   write (iu, *)

end do

do it=1,odfobj(datID)%nt

   ! Object ?: data field

   write (iu, *) 'object ', odfobj(datID)%fld_objID(it), ' class field'
   write (iu, *) 'component "positions" value ', odfobj(datID)%pos_objID
   write (iu, *) 'component "connections" value ', odfobj(datID)%con_objID
   write (iu, *) 'component "data" value ', odfobj(datID)%dat_objID(it)
   write (iu, *)

end do

write (iu, *) 'object "', odfobj(datID)%field_name, '" class series'
do it=1,odfobj(datID)%nt
   write (iu, *) 'member ', it-1, ' value ', odfobj(datID)%fld_objID(it),  &
        ' position ', odfobj(datID)%dat_objtag(it)
end do
write (iu, *)

write (iu, *) 'end'

close(iu)

end subroutine OpenDX_Cvect_Whdr


!```````````````````````````````````````````````````````````````````
subroutine OpenDX_Cvect_Wdat(dfobj_ID, mx, my, mz, ndim, raw_data, recn)

!+ Writes Cartesian (i.e., x,y,z) position elements to a file

use univ_kind_defs_mod, only: f4
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_IO_mem_status_mod, only: for_allocation, for_deallocation,  &
     illegal_alloc_stop

implicit none

!=== Argument declarations:

integer, intent(IN) :: mx, my, mz, dfobj_ID, ndim, recn
real(f4), dimension(mx,my,mz,ndim), intent(IN) :: raw_data

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'OpenDX_Cvect_Wdat'

integer :: count, i, j, k, vd1, vd2, iq, iu, aerr

!=== Executable statements:

vd1 = ndim
vd2 = mx*my*mz
if (.NOT. allocated(vect)) then
   allocate(vect(vd1,vd2), stat=aerr)   ! Allocate memory
   if (aerr /= 0) call illegal_alloc_stop(for_allocation, procedure_name, '1')
else if (vd2 /= size(vect,2) .or. vd1 /= size(vect,1)) then
   if (allocated(vect)) then
      deallocate(vect, stat=aerr)
      if (aerr /= 0) call illegal_alloc_stop(for_deallocation,  &
           procedure_name, '1')
   end if
   allocate(vect(vd1,vd2), stat=aerr)   ! Allocate memory
   if (aerr /= 0) call illegal_alloc_stop(for_allocation, procedure_name, '2')
end if

! Fill 'vect' array (NOTE: input data is rearranged into row-major order): 

count = 0
do i=1,mx
   do j=1,my
      do k=1,mz
         count = count+1
         do iq=1,ndim
            vect(iq,count) = raw_data(i,j,k,iq)
         end do
      end do
   end do
end do

iu = odfobj(dfobj_ID)%IO_unit
write (iu, rec=recn) vect

end subroutine OpenDX_Cvect_Wdat


!```````````````````````````````````````````````````````````````````
subroutine OpenDX_get_endian(which_endian, iu)

!+ Determines which "endian" (byte-order) the current machine is

use univ_kind_defs_mod, only: i4, i2

implicit none

!=== Argument declarations:

integer, intent(IN) :: iu

character(len=3), intent(INOUT) :: which_endian

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'OpenDX_get_endian'

integer(i4) :: i_1
integer(i2) :: i_2

!=== Executable statements:

open(iu, status='scratch', form='unformatted')

i_1 = 7     ! Anything with an asymmetric bit pattern

write (iu) i_1    !
rewind(iu)        ! Write out 4 byte integer; then read back in 
read (iu) i_2     !  the first 2 bytes
close(iu)         !
   
if (i_2 == i_1) then     ! Test for equality
   which_endian = 'lsb'
else
   which_endian = 'msb'
end if

end subroutine OpenDX_get_endian


end module univ_OpenDX_format_mod
