!```````````````````````````````````````````````````````````````````
module univ_Earth_date_time_mod

!+ Contains routines related to dealing with Earth calendars/dates/times

!~ 12.01.2011 (last modified; TIM)

use univ_kind_defs_mod, only: f8

implicit none

private

integer, public, parameter :: canon_s_per_min_i = 60,  &
                              canon_min_per_hour_i = 60,  &
                              canon_hours_per_day_i = 24
integer, public, parameter ::  &
     canon_min_per_day_i = canon_min_per_hour_i*canon_hours_per_day_i,  &
     canon_s_per_day_i = canon_s_per_min_i*canon_min_per_day_i

real(f8), public, parameter ::  &
     canon_s_per_min_f = real(canon_s_per_min_i, kind=f8),  &
     canon_min_per_hour_f = real(canon_min_per_hour_i, kind=f8),  &
     canon_hours_per_day_f = real(canon_hours_per_day_i, kind=f8),  &
     canon_min_per_day_f = real(canon_min_per_day_i, kind=f8),  &
     canon_s_per_day_f = real(canon_s_per_day_i, kind=f8)

integer, public, parameter :: months_per_year = 12

integer, parameter, dimension(months_per_year) :: Gregorian_cal_mdays_norm =  &
     (/ 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 /)
integer, parameter, dimension(months_per_year) :: Gregorian_cal_mdays_leap =  &
     (/ 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 /)

public :: G_date_to_JD, JD_to_G_date_time, G_year_is_a_leap_year,  &
     UTC_date_from_dnt, TAI_minus_UTC_time, decimal_days_to_components,  &
     G_date_time_to_spec_fmt, spec_fmt_to_G_date_time, UTC_to_TAI, TAI_to_TT

contains


!```````````````````````````````````````````````````````````````````
function G_year_is_a_leap_year(year)  result(answer)

!+ Public method that indicates whether a given Gregorian (Earth) calendar year
!   is a "leap" year (i.e., February has 29 days).
!
!  Returns .TRUE. if given year is a "leap" year.

!~ 10.01.2011 (last modified; TIM)

!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

integer, intent(IN) :: year

logical :: answer         ! Procedure result

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'G_year_is_a_leap_year'

!=== Executable statements:

answer = .FALSE.
if (mod(year, 4) == 0) then
   answer = .TRUE.
   if (mod(year, 100) == 0 .and. mod(year, 400) /= 0) answer = .FALSE.
end if
      
end function G_year_is_a_leap_year


!```````````````````````````````````````````````````````````````````
subroutine UTC_date_from_dnt(year, month, day_of_month, hour, minute,  &
     f_seconds, input_dnt_value)

!+ Construct a UTC date from the output of the 'date_and_time' intrinsic
!   function. Optionally, (for testing, etc.) one can provide an input
!   'dnt_value' array (instead of obtaining 'date_and_time' information within
!   this routine).

!~ 11.01.2011 (last modified; TIM)

use univ_kind_defs_mod, only: f8
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

integer, dimension(8), optional, intent(IN) :: input_dnt_value

integer, intent(OUT) :: year, month, day_of_month, hour, minute
real(f8), intent(OUT) ::  f_seconds

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'UTC_date_from_dnt'

integer, dimension(8) :: dnt_value
integer, dimension(months_per_year) :: n_days_per_month
integer :: work_minutes, work_day, work_year, work_month

!=== Executable statements:

if (present(input_dnt_value)) then
   dnt_value(:) = input_dnt_value(:)
else
   call date_and_time(values=dnt_value)
end if

! Will not change by more than one month, so given year "leap" info is OK to
!  use:
work_year = dnt_value(1)
if (G_year_is_a_leap_year(work_year)) then
   n_days_per_month(:) = Gregorian_cal_mdays_leap(:)
else
   n_days_per_month(:) = Gregorian_cal_mdays_norm(:)
end if

work_month = dnt_value(2)

! (Whole minutes in current date/time) + (UTC difference in minutes);
!  UTC adjustment will only alter things by +/- 12 hours, so
!  -720 <= work_minutes < 2160:
work_minutes = dnt_value(5)*canon_min_per_hour_i+dnt_value(6)-dnt_value(4)

work_day = dnt_value(3)

! Adjust date and minutes accordingly:
if (work_minutes < 0 .or. work_minutes >= canon_min_per_day_i) then
   work_day = work_day+sign(1, work_minutes)

   if (work_day == 0) then
      work_month = work_month-1
      if (work_month == 0) then
         work_month = 12; work_year = work_year-1
      end if
      work_day = n_days_per_month(work_month)
   else if (work_day > n_days_per_month(work_month)) then
      work_day = work_day-n_days_per_month(work_month)
      work_month = work_month+1
      if (work_month == 13) then
         work_month = 1; work_year = work_year+1
      end if
   end if

end if
work_minutes = mod(work_minutes+canon_min_per_day_i, canon_min_per_day_i)

year = work_year
month = work_month
day_of_month = work_day
hour = work_minutes/canon_min_per_hour_i
minute = mod(work_minutes, canon_min_per_hour_i)
f_seconds = real(dnt_value(7), kind=f8)+real(dnt_value(8), kind=f8)/1000._f8

end subroutine UTC_date_from_dnt


!```````````````````````````````````````````````````````````````````
subroutine G_date_to_JD(IY, IM, ID, DJM0, DJM, ierr)

!+ == Derived from the original FORTRAN 77 SOFA (Standards Of
!   Fundamental Astronomy Board of the International Astronomical Union)
!   software -- see the copyright info at the end of this procedure)
!
!  == Based on: routine `iau_CAL2JD`, revision 2001 September 16, 
!               SOFA release 2010-12-01
!
!  == Modifications (versus original SOFA code): 
!     1) "Free format" source code ("fixed format" originally)
!     2) Conforms to Fortran 95 standard (FORTRAN 77 originally)
!     3) Procedure name change (as required by SOFA license)
!     4) Introduced dependencies on an external module for KIND constants
!     5) Changed status argument name from 'J' to 'ierr'
!     6) MTAB is now filled from a constant ("parameter") array 'MTAB0', instead
!        of an archaic DATA statement
!
!  Gregorian Calendar to Julian Date.
!
!  Given:   **NOTE**: TT date should be used
!     IY,IM,ID    i     year, month, day in Gregorian calendar (Note 1)
!
!  Returned:
!     DJM0        f8     MJD zero-point: always 2400000.5
!     DJM         f8     Modified Julian Date for 0 hrs
!     ierr        i     status:
!                           0 = OK
!                          -1 = bad year   (Note 3: JD not computed)
!                          -2 = bad month  (JD not computed)
!                          -3 = bad day    (JD computed)
!
!  Notes:
!
!  1) The algorithm used is valid from -4800 March 1, but this
!     implementation rejects dates before -4799 January 1.
!
!  2) The Julian Date is returned in two pieces, in the usual SOFA
!     manner, which is designed to preserve time resolution.  The
!     Julian Date is available as a single number by adding DJM0 and
!     DJM.
!
!  3) In early eras the conversion is from the "Proleptic Gregorian
!     Calendar";  no account is taken of the date(s) of adoption of
!     the Gregorian Calendar, nor is the AD/BC numbering convention
!     observed.
!
!  Reference:
!
!     Explanatory Supplement to the Astronomical Almanac,
!     P. Kenneth Seidelmann (ed), University Science Books (1992),
!     Section 12.92 (p604).

!~ 07.01.2011 (last modified; TIM)

use univ_kind_defs_mod, only: f8
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

integer, intent(IN) :: IY, IM, ID

real(f8), intent(OUT) :: DJM0, DJM
integer, intent(OUT) :: ierr

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'G_date_to_JD'

! Earliest year allowed (4800BC):
integer, parameter :: IYMIN = -4799

integer :: MY, IYPMY
integer, dimension(12) :: MTAB

!=== Executable statements:

ierr = 0       ! Preset status
MTAB(:) = Gregorian_cal_mdays_norm(:)

IF (IY < IYMIN) THEN     ! Validate year
   ierr = -1
ELSE
   IF (IM >= 1 .AND. IM <= 12) THEN    ! Validate month

      ! Allow for leap year.
      IF (MOD(IY, 4) == 0) THEN
         MTAB(2) = 29
      ELSE
         MTAB(2) = 28
      END IF
      IF (MOD(IY, 100) == 0 .AND. MOD(IY, 400) /= 0) MTAB(2) = 28
      
      IF (ID < 1 .OR. ID > MTAB(IM)) ierr = -3   ! Validate day

      ! Result:
      MY = ( IM - 14 ) / 12
      IYPMY = IY + MY
      DJM0 = 2400000.5_f8
      DJM = real( ( 1461 * ( IYPMY + 4800 ) ) / 4  &
           + (  367 * ( IM-2 - 12*MY ) ) / 12  &
           - (    3 * ( ( IYPMY + 4900 ) / 100 ) ) / 4  &
           + ID - 2432076, kind=f8)

   ELSE            ! Bad month
      ierr = -2
   END IF
END IF

!  Excerpted from the original SOFA Software License:
!
!  3. You (the user) may copy and distribute SOFA source code to others,
!     and use and adapt its code and algorithms in your own software,
!     on a world-wide, royalty-free basis.  That portion of your
!     distribution that does not consist of intact and unchanged copies
!     of SOFA source code files is a "derived work" that must comply
!     with the following requirements:
!
!     a) Your work shall be marked or carry a statement that it
!        (i) uses routines and computations derived by you from
!        software provided by SOFA under license to you; and
!        (ii) does not itself constitute software provided by and/or
!        endorsed by SOFA.
!
!     b) The source code of your derived work must contain descriptions
!        of how the derived work is based upon, contains and/or differs
!        from the original SOFA software.
!
!     c) The name(s) of all routine(s) in your derived work shall not
!        include the prefix "iau".
!
!     d) The origin of the SOFA components of your derived work must
!        not be misrepresented;  you must not claim that you wrote the
!        original software, nor file a patent application for SOFA
!        software or algorithms embedded in the SOFA software.
!
!     e) These requirements must be reproduced intact in any source
!        distribution and shall apply to anyone to whom you have
!        granted a further right to modify the source code of your
!        derived work.
!
!     Note that, as originally distributed, the SOFA software is
!     intended to be a definitive implementation of the IAU standards,
!     and consequently third-party modifications are discouraged.  All
!     variations, no matter how minor, must be explicitly marked as
!     such, as explained above.

end subroutine G_date_to_JD


!```````````````````````````````````````````````````````````````````
subroutine JD_to_G_date_time(DJ1, DJ2, IY, IM, ID, FD, ierr)

!+ == Derived from the original FORTRAN 77 SOFA (Standards Of
!   Fundamental Astronomy Board of the International Astronomical Union)
!   software -- see the copyright info at the end of this procedure)
!
!  == Based on: routine `iau_JD2CAL`, revision 2007 October 16, 
!               SOFA release 2010-12-01
!
!  == Modifications (versus original SOFA code): 
!     1) "Free format" source code ("fixed format" originally)
!     2) Conforms to Fortran 95 standard (FORTRAN 77 originally)
!     3) Procedure name change (as required by SOFA license)
!     4) Introduced dependencies on external modules for KIND and math
!        constants
!     5) Changed status argument name from 'J' to 'ierr'
!
!  Julian Date to Gregorian year, month, day, and fraction of a day.
!
!  Given:
!     DJ1,DJ2     f8     Julian Date (Notes 1, 2)
!
!  Returned:   **NOTE**: implicitly TT date/time
!     IY          i     year
!     IM          i     month
!     ID          i     day
!     FD          f8     fraction of day
!     ierr        i     status:
!                           0 = OK
!                          -1 = unacceptable date (Note 3)
!
!  Notes:
!
!  1) The earliest valid date is -68569.5 (-4900 March 1).  The
!     largest value accepted is 10^9.
!
!  2) The Julian Date is apportioned in any convenient way between
!     the arguments DJ1 and DJ2.  For example, JD=2450123.7 could
!     be expressed in any of these ways, among others:
!
!             DJ1            DJ2
!
!         2450123.7D0        0D0        (JD method)
!          2451545D0      -1421.3D0     (J2000 method)
!         2400000.5D0     50123.2D0     (MJD method)
!         2450123.5D0       0.2D0       (date & time method)
!
!  3) In early eras the conversion is from the "Proleptic Gregorian
!     Calendar";  no account is taken of the date(s) of adoption of
!     the Gregorian Calendar, nor is the AD/BC numbering convention
!     observed.
!
!  Reference:
!
!     Explanatory Supplement to the Astronomical Almanac,
!     P. Kenneth Seidelmann (ed), University Science Books (1992),
!     Section 12.92 (p604).

!~ 07.01.2011 (last modified; TIM)

use univ_kind_defs_mod, only: f8
use univ_math_constants_mod, only: ONE_f8, ZERO_f8
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

real(f8), intent(IN) :: DJ1, DJ2

integer, intent(OUT) :: IY, IM, ID, ierr
real(f8), intent(OUT) :: FD

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'JD_to_G_date_time'

! Minimum and maximum allowed JD:
real(f8), parameter :: DJMIN = -68569.5_f8, DJMAX = 1.e9_f8

integer :: JD, L, N, I, K
real(f8) :: DJ, D1, D2, F1, F2, F, D

!=== Executable statements:

DJ = DJ1 + DJ2
IF (DJ < DJMIN .OR. DJ > DJMAX ) THEN   ! Check if date is acceptable
   ierr = -1
ELSE
   ierr = 0

   ! Copy the date, big then small, and re-align to midnight.
   IF (DJ1 >= DJ2) THEN
      D1 = DJ1
      D2 = DJ2
   ELSE
      D1 = DJ2
      D2 = DJ1
   END IF
   D2 = D2 - 0.5_f8

   ! Separate day and fraction.
   F1 = MOD(D1, ONE_f8)
   F2 = MOD(D2, ONE_f8)
   F = MOD(F1+F2, ONE_f8)
   IF (F < ZERO_f8) F = F+ONE_f8
   D = ANINT(D1-F1) + ANINT(D2-F2) + ANINT(F1+F2-F)
   JD = NINT(D) + 1

   ! Express day in Gregorian calendar.
   L = JD + 68569
   N = ( 4*L ) / 146097
   L = L - ( 146097*N + 3 ) / 4
   I = ( 4000 * (L+1) ) / 1461001
   L = L - ( 1461*I ) / 4 + 31
   K = ( 80*L ) / 2447
   ID = L - ( 2447*K ) / 80
   L = K / 11
   IM = K + 2 - 12*L
   IY = 100 * ( N-49 ) + I + L

   FD = F
END IF

!  Excerpted from the original SOFA Software License:
!
!  3. You (the user) may copy and distribute SOFA source code to others,
!     and use and adapt its code and algorithms in your own software,
!     on a world-wide, royalty-free basis.  That portion of your
!     distribution that does not consist of intact and unchanged copies
!     of SOFA source code files is a "derived work" that must comply
!     with the following requirements:
!
!     a) Your work shall be marked or carry a statement that it
!        (i) uses routines and computations derived by you from
!        software provided by SOFA under license to you; and
!        (ii) does not itself constitute software provided by and/or
!        endorsed by SOFA.
!
!     b) The source code of your derived work must contain descriptions
!        of how the derived work is based upon, contains and/or differs
!        from the original SOFA software.
!
!     c) The name(s) of all routine(s) in your derived work shall not
!        include the prefix "iau".
!
!     d) The origin of the SOFA components of your derived work must
!        not be misrepresented;  you must not claim that you wrote the
!        original software, nor file a patent application for SOFA
!        software or algorithms embedded in the SOFA software.
!
!     e) These requirements must be reproduced intact in any source
!        distribution and shall apply to anyone to whom you have
!        granted a further right to modify the source code of your
!        derived work.
!
!     Note that, as originally distributed, the SOFA software is
!     intended to be a definitive implementation of the IAU standards,
!     and consequently third-party modifications are discouraged.  All
!     variations, no matter how minor, must be explicitly marked as
!     such, as explained above.

end subroutine JD_to_G_date_time


!```````````````````````````````````````````````````````````````````
subroutine TAI_minus_UTC_time(IY, IM, ID, FD, DELTAT, ierr)

!+ == Derived from the original FORTRAN 77 SOFA (Standards Of
!   Fundamental Astronomy Board of the International Astronomical Union)
!   software -- see the copyright info at the end of this procedure)
!
!  == Based on: routine `iau_DAT`, revision 2009 November 12, 
!               SOFA release 2010-12-01
!
!  == Modifications (versus original SOFA code): 
!     1) "Free format" source code ("fixed format" originally)
!     2) Conforms to Fortran 95 standard (FORTRAN 77 originally)
!     3) Procedure name change (as required by SOFA license)
!     4) Introduced dependencies on external modules for KIND and math
!        constants
!     5) Changed status argument name from 'J' to 'ierr'
!     6) Replaced GOTO statements with more modern control structures
!
!  For a given UTC date, calculate delta(AT) = TAI-UTC.
!
!     :------------------------------------------:
!     :                                          :
!     :                 IMPORTANT                :
!     :                                          :
!     :  A new version of this routine must be   :
!     :  produced whenever a new leap second is  :
!     :  announced.  There are five items to     :
!     :  change on each such occasion:           :
!     :                                          :
!     :  1) The parameter NDAT must be           :
!     :     increased by 1.                      :
!     :                                          :
!     :  2) A new line must be added to the set  :
!     :     of DATA statements that initialize   :
!     :     the arrays IDATE and DATS.           :
!     :                                          :
!     :  3) The parameter IYV must be set to     :
!     :     the current year.                    :
!     :                                          :
!     :  4) The "Latest leap second" comment     :
!     :     below must be set to the new leap    :
!     :     second date.                         :
!     :                                          :
!     :  5) The "This revision" comment, later,  :
!     :     must be set to the current date.     :
!     :                                          :
!     :  Change (3) must also be carried out     :
!     :  whenever the routine is re-issued,      :
!     :  even if no leap seconds have been       :
!     :  added.                                  :
!     :                                          :
!     :  Latest leap second:  2008 December 31   :
!     :                                          :
!     :__________________________________________:
!
!  Given:
!     IY       i     UTC:  year (Notes 1 and 2)
!     IM       i           month (Note 2)
!     ID       i           day (Notes 2 and 3)
!     FD       d           fraction of day (Note 4)
!
!  Returned:
!     DELTAT   d     TAI minus UTC, seconds
!     ierr     i     status (Note 5):
!                       1 = dubious year (Note 1)
!                       0 = OK
!                      -1 = bad year
!                      -2 = bad month
!                      -3 = bad day (Note 3)
!                      -4 = bad fraction (Note 4)
!
!  Notes:
!
!  1) UTC began at 1960 January 1.0 (JD 2436934.5) and it is improper
!     to call the routine with an earlier date.  If this is attempted,
!     zero is returned together with a warning status.
!
!     Because leap seconds cannot, in principle, be predicted in
!     advance, a reliable check for dates beyond the valid range is
!     impossible.  To guard against gross errors, a year five or more
!     after the release year of the present routine (see parameter IYV)
!     is considered dubious.  In this case a warning status is returned
!     but the result is computed in the normal way.
!
!     For both too-early and too-late years, the warning status is J=+1.
!     This is distinct from the error status J=-1, which signifies a
!     year so early that JD could not be computed.
!
!  2) If the specified date is for a day which ends with a leap second,
!     the UTC-TAI value returned is for the period leading up to the
!     leap second.  If the date is for a day which begins as a leap
!     second ends, the UTC-TAI returned is for the period following the
!     leap second.
!
!  3) The day number must be in the normal calendar range, for example
!     1 through 30 for April.  The "almanac" convention of allowing
!     such dates as January 0 and December 32 is not supported in this
!     routine, in order to avoid confusion near leap seconds.
!
!  4) The fraction of day is used only for dates before the introduction
!     of leap seconds, the first of which occurred at the end of 1971.
!     It is tested for validity (zero to less than 1 is the valid range)
!     even if not used;  if invalid, zero is used and status J=-4 is
!     returned.  For many applications, setting FD to zero is
!     acceptable;  the resulting error is always less than 3 ms (and
!     occurs only pre-1972).
!
!  5) The status value returned in the case where there are multiple
!     errors refers to the first error detected.  For example, if the
!     month and day are 13 and 32 respectively, J=-2 (bad month) will be
!     returned.
!
!  6) In cases where a valid result is not available, zero is returned.
!
!  References:
!
!  1) For dates from 1961 January 1 onwards, the expressions from the
!     file ftp://maia.usno.navy.mil/ser7/tai-utc.dat are used.
!
!  2) The 5ms timestep at 1961 January 1 is taken from 2.58.1 (p87) of
!     the 1992 Explanatory Supplement.
!
!  Called:
!     G_date_to_JD   Gregorian calendar to Julian Day number

!~ 10.01.2011 (last modified; TIM)

use univ_kind_defs_mod, only: f8
use univ_math_constants_mod, only: ZERO_f8, ONE_f8
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

integer, intent(IN) :: IY, IM, ID
real(f8), intent(IN) :: FD

real(f8), intent(OUT) :: DELTAT
integer, intent(OUT) :: ierr

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'TAI_minus_UTC_time'

! Release year for this version of iau_DAT:
integer, parameter :: IYV = 2009

! Number of Delta(AT) changes (increase by 1 for each new leap second):
integer, parameter :: NDAT = 39

! Number of Delta(AT) expressions before leap seconds were introduced:
integer, parameter :: NERA1 = 14

! Dates (year, month) on which new Delta(AT) came into force:
integer, dimension(2,NDAT) :: IDATE

! New Delta(AT) which came into force on the given dates:
real(f8), dimension(NDAT) :: DATS

! Reference dates (MJD) and drift rates (s/day), pre leap seconds:
real(f8), dimension(2,NERA1) :: DRIFT

logical :: MORE
integer :: JS, I, M, IS
real(f8) :: DA, DJM0, DJM

!=== Executable statements:

! Dates and Delta(AT)s:
IDATE(1:2,1) = (/ 1960, 1 /); DATS(1) = 1.4178180_f8
IDATE(1:2,2) = (/ 1961, 1 /); DATS(2) = 1.4228180_f8
IDATE(1:2,3) = (/ 1961, 8 /); DATS(3) = 1.3728180_f8
IDATE(1:2,4) = (/ 1962, 1 /); DATS(4) = 1.8458580_f8
IDATE(1:2,5) = (/ 1963, 11 /); DATS(5) = 1.9458580_f8
IDATE(1:2,6) = (/ 1964, 1 /); DATS(6) = 3.2401300_f8
IDATE(1:2,7) = (/ 1964, 4 /); DATS(7) = 3.3401300_f8
IDATE(1:2,8) = (/ 1964, 9 /); DATS(8) = 3.4401300_f8
IDATE(1:2,9) = (/ 1965, 1 /); DATS(9) = 3.5401300_f8
IDATE(1:2,10) = (/ 1965, 3 /); DATS(10) = 3.6401300_f8
IDATE(1:2,11) = (/ 1965, 7 /); DATS(11) = 3.7401300_f8
IDATE(1:2,12) = (/ 1965, 9 /); DATS(12) = 3.8401300_f8
IDATE(1:2,13) = (/ 1966, 1 /); DATS(13) = 4.3131700_f8
IDATE(1:2,14) = (/ 1968, 2 /); DATS(14) = 4.2131700_f8
IDATE(1:2,15) = (/ 1972, 1 /); DATS(15) = 10._f8
IDATE(1:2,16) = (/ 1972, 7 /); DATS(16) = 11._f8
IDATE(1:2,17) = (/ 1973, 1 /); DATS(17) = 12._f8
IDATE(1:2,18) = (/ 1974, 1 /); DATS(18) = 13._f8
IDATE(1:2,19) = (/ 1975, 1 /); DATS(19) = 14._f8
IDATE(1:2,20) = (/ 1976, 1 /); DATS(20) = 15._f8
IDATE(1:2,21) = (/ 1977, 1 /); DATS(21) = 16._f8
IDATE(1:2,22) = (/ 1978, 1 /); DATS(22) = 17._f8
IDATE(1:2,23) = (/ 1979, 1 /); DATS(23) = 18._f8
IDATE(1:2,24) = (/ 1980, 1 /); DATS(24) = 19._f8
IDATE(1:2,25) = (/ 1981, 7 /); DATS(25) = 20._f8
IDATE(1:2,26) = (/ 1982, 7 /); DATS(26) = 21._f8
IDATE(1:2,27) = (/ 1983, 7 /); DATS(27) = 22._f8
IDATE(1:2,28) = (/ 1985, 7 /); DATS(28) = 23._f8
IDATE(1:2,29) = (/ 1988, 1 /); DATS(29) = 24._f8
IDATE(1:2,30) = (/ 1990, 1 /); DATS(30) = 25._f8
IDATE(1:2,31) = (/ 1991, 1 /); DATS(31) = 26._f8
IDATE(1:2,32) = (/ 1992, 7 /); DATS(32) = 27._f8
IDATE(1:2,33) = (/ 1993, 7 /); DATS(33) = 28._f8
IDATE(1:2,34) = (/ 1994, 7 /); DATS(34) = 29._f8
IDATE(1:2,35) = (/ 1996, 1 /); DATS(35) = 30._f8
IDATE(1:2,36) = (/ 1997, 7 /); DATS(36) = 31._f8
IDATE(1:2,37) = (/ 1999, 1 /); DATS(37) = 32._f8
IDATE(1:2,38) = (/ 2006, 1 /); DATS(38) = 33._f8
IDATE(1:2,39) = (/ 2009, 1 /); DATS(39) = 34._f8

! Reference dates and drift rates:
DRIFT(1:2,1) = (/ 37300._f8, 0.001296_f8 /)
DRIFT(1:2,2) = (/ 37300._f8, 0.001296_f8 /)
DRIFT(1:2,3) = (/ 37300._f8, 0.001296_f8 /)
DRIFT(1:2,4) = (/ 37665._f8, 0.0011232_f8 /)
DRIFT(1:2,5) = (/ 37665._f8, 0.0011232_f8 /)
DRIFT(1:2,6) = (/ 38761._f8, 0.001296_f8 /)
DRIFT(1:2,7) = (/ 38761._f8, 0.001296_f8 /)
DRIFT(1:2,8) = (/ 38761._f8, 0.001296_f8 /)
DRIFT(1:2,9) = (/ 38761._f8, 0.001296_f8 /)
DRIFT(1:2,10) = (/ 38761._f8, 0.001296_f8 /)
DRIFT(1:2,11) = (/ 38761._f8, 0.001296_f8 /)
DRIFT(1:2,12) = (/ 38761._f8, 0.001296_f8 /)
DRIFT(1:2,13) = (/ 39126._f8, 0.002592_f8 /)
DRIFT(1:2,14) = (/ 39126._f8, 0.002592_f8 /)

! Initialize the result to zero and the status to OK.
DA = ZERO_f8
JS = 0

! If invalid fraction of a day, set error status and give up.
IF (FD < ZERO_f8 .OR. FD >= ONE_f8) THEN
   JS = -4
   DELTAT = DA; ierr = JS; return
END IF

! Convert the date into an MJD.
call G_date_to_JD(IY, IM, ID, DJM0, DJM, JS)

! If invalid year, month, or day, give up.
if (JS < 0) then
   DELTAT = DA; ierr = JS; return
end if

! If pre-UTC year, set warning status and give up.
IF (IY < IDATE(1,1)) THEN
   JS = 1
   DELTAT = DA; ierr = JS; return
END IF

! If suspiciously late year, set warning status but proceed.
IF (IY > IYV+5) JS = 1

! Combine year and month.
M = 12*IY+IM

!  Find the most recent table entry.
IS = 0
MORE = .TRUE.
do I=NDAT,1,-1
   IF (MORE) THEN
      IS = I
      MORE = M < ( 12*IDATE(1,I) + IDATE(2,I) )
   END IF
end do

! Get the Delta(AT).
DA = DATS(IS)

! If pre-1972, adjust for drift.
IF (IS <= NERA1) DA = DA + ( DJM + FD - DRIFT(1,IS) ) * DRIFT(2,IS)

! Return the Delta(AT) value and the status.
DELTAT = DA
ierr = JS

!  Excerpted from the original SOFA Software License:
!
!  3. You (the user) may copy and distribute SOFA source code to others,
!     and use and adapt its code and algorithms in your own software,
!     on a world-wide, royalty-free basis.  That portion of your
!     distribution that does not consist of intact and unchanged copies
!     of SOFA source code files is a "derived work" that must comply
!     with the following requirements:
!
!     a) Your work shall be marked or carry a statement that it
!        (i) uses routines and computations derived by you from
!        software provided by SOFA under license to you; and
!        (ii) does not itself constitute software provided by and/or
!        endorsed by SOFA.
!
!     b) The source code of your derived work must contain descriptions
!        of how the derived work is based upon, contains and/or differs
!        from the original SOFA software.
!
!     c) The name(s) of all routine(s) in your derived work shall not
!        include the prefix "iau".
!
!     d) The origin of the SOFA components of your derived work must
!        not be misrepresented;  you must not claim that you wrote the
!        original software, nor file a patent application for SOFA
!        software or algorithms embedded in the SOFA software.
!
!     e) These requirements must be reproduced intact in any source
!        distribution and shall apply to anyone to whom you have
!        granted a further right to modify the source code of your
!        derived work.
!
!     Note that, as originally distributed, the SOFA software is
!     intended to be a definitive implementation of the IAU standards,
!     and consequently third-party modifications are discouraged.  All
!     variations, no matter how minor, must be explicitly marked as
!     such, as explained above.

end subroutine TAI_minus_UTC_time


!```````````````````````````````````````````````````````````````````
subroutine UTC_to_TAI(UTC1, UTC2, TAI1, TAI2, ierr)

!+ == Derived from the original FORTRAN 77 SOFA (Standards Of
!   Fundamental Astronomy Board of the International Astronomical Union)
!   software -- see the copyright info at the end of this procedure)
!
!  == Based on: routine `iau_UTCTAI`, revision 2010 September 10, 
!               SOFA release 2010-12-01
!
!  == Modifications (versus original SOFA code): 
!     1) "Free format" source code ("fixed format" originally)
!     2) Conforms to Fortran 95 standard (FORTRAN 77 originally)
!     3) Procedure name change (as required by SOFA license)
!     4) Introduced dependencies on an external module for KIND constants
!     5) Changed status argument name from 'J' to 'ierr'
!
!  Time scale transformation:  Coordinated Universal Time, UTC, to
!  International Atomic Time, TAI.
!
!  Given:
!     UTC1,UTC2    d      UTC as a 2-part quasi Julian Date (Notes 1-4)
!
!  Returned:
!     TAI1,TAI2    d      TAI as a 2-part Julian Date (Note 5)
!     ierr         i      status: +1 = dubious year (Note 3)
!                                  0 = OK
!                                 -1 = unacceptable date
!
!  Notes:
!
!  1) UTC1+UTC2 is quasi Julian Date (see Note 2), apportioned in any
!     convenient way between the two arguments, for example where UTC1
!     is the Julian Day Number and UTC2 is the fraction of a day.
!
!  2) JD cannot unambiguously represent UTC during a leap second unless
!     special measures are taken.  The convention in the present routine
!     is that the JD day represents UTC days whether the length is
!     86399, 86400 or 86401 SI seconds.
!
!  3) The warning status "dubious year" flags UTCs that predate the
!     introduction of the time scale and that are too far in the future
!     to be trusted.  See iau_DAT for further details.
!
!  4) The routine iau_DTF2D converts from calendar date and time of day
!     into 2-part Julian Date, and in the case of UTC implements the
!     leap-second-ambiguity convention described above.
!
!  5) The returned TAI1,TAI2 are such that their sum is the TAI Julian
!     Date.
!
!  Called:
!     JD_to_G_date_time   JD to Gregorian calendar
!     TAI_minus_UTC_time      delta(AT) = TAI-UTC
!     G_date_to_JD   Gregorian calendar to JD
!
!  References:
!
!     McCarthy, D. D., Petit, G. (eds.), IERS Conventions (2003),
!     IERS Technical Note No. 32, BKG (2004)
!
!     Explanatory Supplement to the Astronomical Almanac,
!     P. Kenneth Seidelmann (ed), University Science Books (1992)

!~ 12.01.2011 (last modified; TIM)

use univ_kind_defs_mod, only: f8
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

real(f8), intent(IN) :: UTC1, UTC2

real(f8), intent(OUT) :: TAI1, TAI2
integer, intent(OUT) :: ierr

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'UTC_to_TAI'

! Days to seconds:
real(f8), parameter :: D2S = canon_s_per_day_f

logical :: BIG1
integer :: IY, IM, ID, JS, IYT, IMT, IDT
real(f8) :: U1, U2, FD, DATS, FDT, DATST, DDAT, Z1, Z2, A2

!=== Executable statements:

! Put the two parts of the UTC into big-first order.
BIG1 = UTC1 >= UTC2
IF (BIG1) THEN
   U1 = UTC1
   U2 = UTC2
ELSE
   U1 = UTC2
   U2 = UTC1
END IF

! Get TAI-UTC now.
call JD_to_G_date_time(U1, U2, IY, IM, ID, FD, JS)
if (JS /= 0) then
   JS = -1; ierr = JS; return
end if
call TAI_minus_UTC_time(IY, IM, ID, FD, DATS, JS)
if (JS < 0) then
   JS = -1; ierr = JS; return
end if

! Get TAI-UTC tomorrow.
call JD_to_G_date_time(U1+1.5_f8, U2-FD, IYT, IMT, IDT, FDT, JS)
if (JS /= 0) then
   JS = -1; ierr = JS; return
end if
call TAI_minus_UTC_time(IYT, IMT, IDT, FDT, DATST, JS)
if (JS < 0) then
   JS = -1; ierr = JS; return
end if

! If today ends in a leap second, scale the fraction into SI days.
DDAT = DATST - DATS
IF (ABS(DDAT) > 0.5_f8) FD = FD + FD*DDAT/D2S

! Today's calendar date to 2-part JD.
call G_date_to_JD(IY, IM, ID, Z1, Z2, JS)
if (JS /= 0)  then
   JS = -1; ierr = JS; return
end if

! Assemble the TAI result, preserving the UTC split and order.
A2 = Z1 - U1
A2 = ( A2 + Z2 ) + ( FD + DATS/D2S )
IF (BIG1) THEN
   TAI1 = U1
   TAI2 = A2
ELSE
   TAI1 = A2
   TAI2 = U1
END IF

! Status.
IF (JS /= 0) JS = -1
ierr = JS

!  Excerpted from the original SOFA Software License:
!
!  3. You (the user) may copy and distribute SOFA source code to others,
!     and use and adapt its code and algorithms in your own software,
!     on a world-wide, royalty-free basis.  That portion of your
!     distribution that does not consist of intact and unchanged copies
!     of SOFA source code files is a "derived work" that must comply
!     with the following requirements:
!
!     a) Your work shall be marked or carry a statement that it
!        (i) uses routines and computations derived by you from
!        software provided by SOFA under license to you; and
!        (ii) does not itself constitute software provided by and/or
!        endorsed by SOFA.
!
!     b) The source code of your derived work must contain descriptions
!        of how the derived work is based upon, contains and/or differs
!        from the original SOFA software.
!
!     c) The name(s) of all routine(s) in your derived work shall not
!        include the prefix "iau".
!
!     d) The origin of the SOFA components of your derived work must
!        not be misrepresented;  you must not claim that you wrote the
!        original software, nor file a patent application for SOFA
!        software or algorithms embedded in the SOFA software.
!
!     e) These requirements must be reproduced intact in any source
!        distribution and shall apply to anyone to whom you have
!        granted a further right to modify the source code of your
!        derived work.
!
!     Note that, as originally distributed, the SOFA software is
!     intended to be a definitive implementation of the IAU standards,
!     and consequently third-party modifications are discouraged.  All
!     variations, no matter how minor, must be explicitly marked as
!     such, as explained above.

end subroutine UTC_to_TAI


!```````````````````````````````````````````````````````````````````
subroutine TAI_to_TT(TAI1, TAI2, TT1, TT2, ierr)

!+ == Derived from the original FORTRAN 77 SOFA (Standards Of
!   Fundamental Astronomy Board of the International Astronomical Union)
!   software -- see the copyright info at the end of this procedure)
!
!  == Based on: routine `iau_TAITT`, revision 2010 April 16, 
!               SOFA release 2010-12-01
!
!  == Modifications (versus original SOFA code): 
!     1) "Free format" source code ("fixed format" originally)
!     2) Conforms to Fortran 95 standard (FORTRAN 77 originally)
!     3) Procedure name change (as required by SOFA license)
!     4) Introduced dependencies on an external module for KIND constants
!     5) Changed status argument name from 'J' to 'ierr'
!
!  Time scale transformation:  International Atomic Time, TAI, to
!  Terrestrial Time, TT.
!
!  Given:
!     TAI1,TAI2    d      TAI as a 2-part Julian Date
!
!  Returned:
!     TT1,TT2      d      TT as a 2-part Julian Date
!     ierr         i      status:  0 = OK
!
!  Note:
!
!     TAI1+TAI2 is Julian Date, apportioned in any convenient way
!     between the two arguments, for example where TAI1 is the Julian
!     Day Number and TAI2 is the fraction of a day.  The returned
!     TT1,TT2 follow suit.
!
!  References:
!
!     McCarthy, D. D., Petit, G. (eds.), IERS Conventions (2003),
!     IERS Technical Note No. 32, BKG (2004)
!
!     Explanatory Supplement to the Astronomical Almanac,
!     P. Kenneth Seidelmann (ed), University Science Books (1992)

!~ 12.01.2011 (last modified; TIM)

use univ_kind_defs_mod, only: f8
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

real(f8), intent(IN) :: TAI1, TAI2

real(f8), intent(OUT) :: TT1, TT2
integer, intent(OUT) :: ierr

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'TAI_to_TT'

!  TT minus TAI (days).
real(f8), parameter :: DTAT = 32.184_f8/canon_s_per_day_f

!=== Executable statements:

! Result, safeguarding precision.
IF (TAI1 > TAI2) THEN
   TT1 = TAI1
   TT2 = TAI2 + DTAT
ELSE
   TT1 = TAI1 + DTAT
   TT2 = TAI2
END IF

! Status (always OK).
ierr = 0

!  Excerpted from the original SOFA Software License:
!
!  3. You (the user) may copy and distribute SOFA source code to others,
!     and use and adapt its code and algorithms in your own software,
!     on a world-wide, royalty-free basis.  That portion of your
!     distribution that does not consist of intact and unchanged copies
!     of SOFA source code files is a "derived work" that must comply
!     with the following requirements:
!
!     a) Your work shall be marked or carry a statement that it
!        (i) uses routines and computations derived by you from
!        software provided by SOFA under license to you; and
!        (ii) does not itself constitute software provided by and/or
!        endorsed by SOFA.
!
!     b) The source code of your derived work must contain descriptions
!        of how the derived work is based upon, contains and/or differs
!        from the original SOFA software.
!
!     c) The name(s) of all routine(s) in your derived work shall not
!        include the prefix "iau".
!
!     d) The origin of the SOFA components of your derived work must
!        not be misrepresented;  you must not claim that you wrote the
!        original software, nor file a patent application for SOFA
!        software or algorithms embedded in the SOFA software.
!
!     e) These requirements must be reproduced intact in any source
!        distribution and shall apply to anyone to whom you have
!        granted a further right to modify the source code of your
!        derived work.
!
!     Note that, as originally distributed, the SOFA software is
!     intended to be a definitive implementation of the IAU standards,
!     and consequently third-party modifications are discouraged.  All
!     variations, no matter how minor, must be explicitly marked as
!     such, as explained above.

end subroutine TAI_to_TT


!```````````````````````````````````````````````````````````````````
subroutine decimal_days_to_components(NDP, DAYS, SIGN, IHMSF)

!+ == Derived from the original FORTRAN 77 SOFA (Standards Of
!   Fundamental Astronomy Board of the International Astronomical Union)
!   software -- see the copyright info at the end of this procedure)
!
!  == Based on: routine `iau_D2TF`, revision 2005 August 26, 
!               SOFA release 2010-12-01
!
!  == Modifications (versus original SOFA code): 
!     1) "Free format" source code ("fixed format" originally)
!     2) Conforms to Fortran 95 standard (FORTRAN 77 originally)
!     3) Procedure name change (as required by SOFA license)
!     4) Introduced dependencies on external modules for KIND and math
!        constants
!     5) Changed status argument name from 'J' to 'ierr'
!     6) Replaced GOTO statements with more modern control structures
!
!  Decompose days to hours, minutes, seconds, fraction.
!
!  Given:
!     NDP       i        resolution (Note 1)
!     DAYS      d        interval in days
!
!  Returned:
!     SIGN      c        '+' or '-'
!     IHMSF     i(4)     hours, minutes, seconds, fraction
!
!  Notes:
!
!  1) NDP is interpreted as follows:
!
!     NDP         resolution
!      :      ...0000 00 00
!     -7         1000 00 00
!     -6          100 00 00
!     -5           10 00 00
!     -4            1 00 00
!     -3            0 10 00
!     -2            0 01 00
!     -1            0 00 10
!      0            0 00 01
!      1            0 00 00.1
!      2            0 00 00.01
!      3            0 00 00.001
!      :            0 00 00.000...
!
!  2) The largest positive useful value for NDP is determined by the
!     size of DAYS, the format of DOUBLE PRECISION floating-point
!     numbers on the target platform, and the risk of overflowing
!     IHMSF(4).  On a typical platform, for DAYS up to 1D0, the
!     available floating-point precision might correspond to NDP=12.
!     However, the practical limit is typically NDP=9, set by the
!     capacity of a 32-bit IHMSF(4).
!
!  3) The absolute value of DAYS may exceed 1D0.  In cases where it
!     does not, it is up to the caller to test for and handle the
!     case where DAYS is very nearly 1D0 and rounds up to 24 hours,
!     by testing for IHMSF(1)=24 and setting IHMSF(1-4) to zero.

!~ 08.01.2011 (last modified; TIM)

use univ_kind_defs_mod, only: f8
use univ_math_constants_mod, only: ZERO_f8
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

integer, intent(IN) :: NDP
real(f8), intent(IN) :: DAYS

character(len=1), intent(OUT) :: SIGN
integer, dimension(4), intent(OUT) :: IHMSF

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'decimal_days_to_components'

!  Days to seconds:
real(f8), parameter :: D2S = canon_s_per_day_f

integer :: NRS, N
real(f8) :: RS, RM, RH, A, AH, AM, AS, AF

!=== Executable statements:

! Handle sign.
IF (DAYS >= ZERO_f8) THEN
   SIGN = '+'
ELSE
   SIGN = '-'
END IF

! Interval in seconds.
A = D2S * ABS(DAYS)

! Pre-round if resolution coarser than 1 second (then pretend NDP=1).
IF (NDP < 0) THEN
   NRS = 1
   do N=1,-NDP
      IF (N == 2 .OR. N == 4) THEN
         NRS = NRS * 6
      ELSE
         NRS = NRS * 10
      END IF
   end do
   RS = real(NRS, kind=f8)
   A = RS * ANINT(A/RS)
END IF

! Express the unit of each field in resolution units.
NRS = 1
do N=1,NDP
   NRS = NRS * 10
end do
RS = real(NRS, kind=f8)
RM = RS * 60._f8
RH = RM * 60._f8

! Round the interval and express in resolution units.
A = ANINT(RS*A)

! Break into fields.
AH = AINT(A/RH)
A = A - AH*RH
AM = AINT(A/RM)
A = A - AM*RM
AS = AINT(A/RS)
AF = A - AS*RS

! Return results.
IHMSF(1) = NINT(AH)
IHMSF(2) = NINT(AM)
IHMSF(3) = NINT(AS)
IHMSF(4) = NINT(AF)

!  Excerpted from the original SOFA Software License:
!
!  3. You (the user) may copy and distribute SOFA source code to others,
!     and use and adapt its code and algorithms in your own software,
!     on a world-wide, royalty-free basis.  That portion of your
!     distribution that does not consist of intact and unchanged copies
!     of SOFA source code files is a "derived work" that must comply
!     with the following requirements:
!
!     a) Your work shall be marked or carry a statement that it
!        (i) uses routines and computations derived by you from
!        software provided by SOFA under license to you; and
!        (ii) does not itself constitute software provided by and/or
!        endorsed by SOFA.
!
!     b) The source code of your derived work must contain descriptions
!        of how the derived work is based upon, contains and/or differs
!        from the original SOFA software.
!
!     c) The name(s) of all routine(s) in your derived work shall not
!        include the prefix "iau".
!
!     d) The origin of the SOFA components of your derived work must
!        not be misrepresented;  you must not claim that you wrote the
!        original software, nor file a patent application for SOFA
!        software or algorithms embedded in the SOFA software.
!
!     e) These requirements must be reproduced intact in any source
!        distribution and shall apply to anyone to whom you have
!        granted a further right to modify the source code of your
!        derived work.
!
!     Note that, as originally distributed, the SOFA software is
!     intended to be a definitive implementation of the IAU standards,
!     and consequently third-party modifications are discouraged.  All
!     variations, no matter how minor, must be explicitly marked as
!     such, as explained above.

end subroutine decimal_days_to_components


!```````````````````````````````````````````````````````````````````
subroutine G_date_time_to_spec_fmt(SCALE, IY, IM, ID, IHR, IMN, SEC, D1, D2,  &
     ierr)

!+ == Derived from the original FORTRAN 77 SOFA (Standards Of
!   Fundamental Astronomy Board of the International Astronomical Union)
!   software -- see the copyright info at the end of this procedure)
!
!  == Based on: routine `iau_DTF2D`, revision 2010 May 25, 
!               SOFA release 2010-12-01
!
!  == Modifications (versus original SOFA code): 
!     1) "Free format" source code ("fixed format" originally)
!     2) Conforms to Fortran 95 standard (FORTRAN 77 originally)
!     3) Procedure name change (as required by SOFA license)
!     4) Introduced dependencies on external modules for KIND and math
!        constants
!     5) Changed status argument name from 'J' to 'ierr'
!     6) Replaced GOTO statements with more modern control structures
!
!  Encode date and time fields into 2-part Julian Date (or in the case
!  of UTC a quasi-JD form that includes special provision for leap
!  seconds).
!
!  Given:
!     SCALE      c(len=*)  time scale ID (Note 1)
!     IY,IM,ID   i      year, month, day in Gregorian calendar (Note 2)
!     IHR,IMN    i      hour, minute
!     SEC        d      seconds
!
!  Returned:
!     DJ1,DJ2    d      2-part Julian Date (Notes 3,4)
!     ierr       i      status: +3 = both of next two
!                               +2 = time is after end of day (Note 5)
!                               +1 = dubious year (Note 6)
!                                0 = OK
!                               -1 = bad year
!                               -2 = bad month
!                               -3 = bad day
!                               -4 = bad hour
!                               -5 = bad minute
!                               -6 = bad second (<0)
!
!  Notes:
!
!  1) SCALE identifies the time scale.  Only the value 'UTC' (in upper
!     case) is significant, and enables handling of leap seconds (see
!     Note 4).
!
!  2) For calendar conventions and limitations, see iau_CAL2JD.
!
!  3) The sum of the results, D1+D2, is Julian Date, where normally D1
!     is the Julian Day Number and D2 is the fraction of a day.  In the
!     case of UTC, where the use of JD is problematical, special
!     conventions apply:  see the next note.
!
!  4) JD cannot unambiguously represent UTC during a leap second unless
!     special measures are taken.  The SOFA internal convention is that
!     the quasi-JD day represents UTC days whether the length is 86399,
!     86400 or 86401 SI seconds.
!
!  5) The warning status "time is after end of day" usually means that
!     the SEC argument is greater than 60D0.  However, in a day ending
!     in a leap second the limit changes to 61D0 (or 59D0 in the case of
!     a negative leap second).
!
!  6) The warning status "dubious year" flags UTCs that predate the
!     introduction of the time scale and that are too far in the future
!     to be trusted.  See iau_DAT for further details.
!
!  7) Only in the case of continuous and regular time scales (TAI, TT,
!     TCG, TCB and TDB) is the result DJ1+DJ2 a Julian Date, strictly
!     speaking.  In the other cases (UT1 and UTC) the result must be
!     used with circumspection;  in particular the difference between
!     two such results cannot be interpreted as a precise time
!     interval.
!
!  Called:
!     G_date_to_JD   Gregorian calendar to JD
!     TAI_minus_UTC_time      delta(AT) = TAI-UTC
!     JD_to_G_date_time   JD to Gregorian calendar

!~ 08.01.2011 (last modified; TIM)

use univ_kind_defs_mod, only: f8
use univ_math_constants_mod, only: ONE_f8, ZERO_f8
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

character(len=*), intent(IN) :: SCALE
integer, intent(IN) :: IY, IM, ID, IHR, IMN
real(f8), intent(IN) :: SEC

real(f8), intent(OUT) :: D1, D2
integer, intent(OUT) :: ierr

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'G_date_time_to_spec_fmt'

real(f8), parameter :: D2S = canon_s_per_day_f     !  Days to seconds

integer :: JS, IY2, IM2, ID2
real(f8) :: DJ, W, DAY, SECLIM, DAT1, DAT2, DDT, TIME

!=== Executable statements:

call G_date_to_JD(IY, IM, ID, DJ, W, JS)   ! Today's Julian Day Number

if (JS /= 0) then
   ierr = JS; return
end if

DJ = DJ + W

! Day length and final minute length in seconds (provisional).
DAY = D2S
SECLIM = 60._f8

! Deal with the UTC leap second case.
IF (SCALE == 'UTC') THEN

   ! TAI-UTC today.
   call TAI_minus_UTC_time(IY, IM, ID, ZERO_f8, DAT1, JS)
   if (JS < 0) then
      ierr = JS; return
   end if

   ! TAI-UTC tomorrow.
   call JD_to_G_date_time(DJ, ONE_f8, IY2, IM2, ID2, W, JS)
   if (JS /= 0) then
      ierr = JS; return
   end if
   call TAI_minus_UTC_time(IY2, IM2, ID2, ZERO_f8, DAT2, JS)
   if (JS < 0) then
      ierr = JS; return
   end if

   ! The change in TAI-UTC (seconds).
   DDT = DAT2 - DAT1

   ! If leap second day, correct the day and final minute lengths.
   IF (ABS(DDT) > 0.5_f8) THEN
      DAY = DAY + DDT
      IF (IHR == 23 .AND. IMN == 59) SECLIM = SECLIM + DDT
   END IF
END IF

! Validate the time.
IF (IHR >= 0 .AND. IHR <= 23) THEN
   IF (IMN >= 0 .AND. IMN <= 59) THEN
      IF (SEC >= ZERO_f8) THEN
         IF (SEC >= SECLIM) THEN
            JS = JS + 2
         END IF
      ELSE
         JS = -6
      END IF
   ELSE
      JS = -5
   END IF
ELSE
   JS = -4
END IF
if (JS < 0) then
   ierr = JS; return
end if

! The time in days.
TIME = (60._f8*real(60*IHR+IMN, kind=f8)+SEC) / DAY

! Return the date and time.
D1 = DJ
D2 = TIME

! Return the status.
ierr = JS

!  Excerpted from the original SOFA Software License:
!
!  3. You (the user) may copy and distribute SOFA source code to others,
!     and use and adapt its code and algorithms in your own software,
!     on a world-wide, royalty-free basis.  That portion of your
!     distribution that does not consist of intact and unchanged copies
!     of SOFA source code files is a "derived work" that must comply
!     with the following requirements:
!
!     a) Your work shall be marked or carry a statement that it
!        (i) uses routines and computations derived by you from
!        software provided by SOFA under license to you; and
!        (ii) does not itself constitute software provided by and/or
!        endorsed by SOFA.
!
!     b) The source code of your derived work must contain descriptions
!        of how the derived work is based upon, contains and/or differs
!        from the original SOFA software.
!
!     c) The name(s) of all routine(s) in your derived work shall not
!        include the prefix "iau".
!
!     d) The origin of the SOFA components of your derived work must
!        not be misrepresented;  you must not claim that you wrote the
!        original software, nor file a patent application for SOFA
!        software or algorithms embedded in the SOFA software.
!
!     e) These requirements must be reproduced intact in any source
!        distribution and shall apply to anyone to whom you have
!        granted a further right to modify the source code of your
!        derived work.
!
!     Note that, as originally distributed, the SOFA software is
!     intended to be a definitive implementation of the IAU standards,
!     and consequently third-party modifications are discouraged.  All
!     variations, no matter how minor, must be explicitly marked as
!     such, as explained above.

end subroutine G_date_time_to_spec_fmt


!```````````````````````````````````````````````````````````````````
subroutine spec_fmt_to_G_date_time(SCALE, NDP, D1, D2, IY, IM, ID, IHMSF, ierr)

!+ == Derived from the original FORTRAN 77 SOFA (Standards Of
!   Fundamental Astronomy Board of the International Astronomical Union)
!   software -- see the copyright info at the end of this procedure)
!
!  == Based on: routine `iau_D2DTF`, revision 2010 July 26, 
!               SOFA release 2010-12-01
!
!  == Modifications (versus original SOFA code): 
!     1) "Free format" source code ("fixed format" originally)
!     2) Conforms to Fortran 95 standard (FORTRAN 77 originally)
!     3) Procedure name change (as required by SOFA license)
!     4) Introduced dependencies on external modules for KIND and math
!        constants
!     5) Changed status argument name from 'J' to 'ierr'
!     6) Replaced GOTO statements with more modern control structures
!
!  Format for output a 2-part Julian Date (or in the case of UTC a
!  quasi-JD form that includes special provision for leap seconds).
!
!  Given:
!     SCALE      c(len=*)  time scale ID (Note 1)
!     NDP        i      resolution (Note 2)
!     D1,D2      d      time as a 2-part Julian Date (Notes 3,4)
!
!  Returned:
!     IY,IM,ID   i      year, month, day in Gregorian calendar (Note 5)
!     IHMSF      i(4)   hours, minutes, seconds, fraction (Note 1)
!     ierr       i      status: +1 = dubious year (Note 5)
!                                0 = OK
!                               -1 = unacceptable date (Note 6)
!
!  Notes:
!
!  1) SCALE identifies the time scale.  Only the value 'UTC' (in upper
!     case) is significant, and enables handling of leap seconds (see
!     Note 4).
!
!  2) NDP is the number of decimal places in the seconds field, and can
!     have negative as well as positive values, such as:
!
!     NDP         resolution
!     -4            1 00 00
!     -3            0 10 00
!     -2            0 01 00
!     -1            0 00 10
!      0            0 00 01
!      1            0 00 00.1
!      2            0 00 00.01
!      3            0 00 00.001
!
!     The limits are platform dependent, but a safe range is -5 to +9.
!
!  3) D1+D2 is Julian Date, apportioned in any convenient way between
!     the two arguments, for example where D1 is the Julian Day Number
!     and D2 is the fraction of a day.  In the case of UTC, where the
!     use of JD is problematical, special conventions apply:  see the
!     next note.
!
!  4) JD cannot unambiguously represent UTC during a leap second unless
!     special measures are taken.  The SOFA internal convention is that
!     the quasi-JD day represents UTC days whether the length is 86399,
!     86400 or 86401 SI seconds.
!
!  5) The warning status "dubious year" flags UTCs that predate the
!     introduction of the time scale and that are too far in the future
!     to be trusted.  See iau_DAT for further details.
!
!  6) For calendar conventions and limitations, see iau_CAL2JD.
!
!  Called:
!     JD_to_G_date_time   JD to Gregorian calendar
!     decimal_days_to_components     decompose days to hms
!     TAI_minus_UTC_time      delta(AT) = TAI-UTC

!~ 08.01.2011 (last modified; TIM)

use univ_kind_defs_mod, only: f8
use univ_math_constants_mod, only: ZERO_f8
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

character(len=*), intent(IN) :: SCALE
integer, intent(IN) :: NDP
real(f8), intent(IN) :: D1, D2

integer, intent(OUT) :: IY, IM, ID, ierr
integer, dimension(4), intent(OUT) :: IHMSF

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'spec_fmt_to_G_date_time'

! Days to seconds:
real(f8), parameter :: D2S = canon_s_per_day_f

logical :: LEAP
character(len=1) :: S
integer :: IY1, IM1, ID1, JS, IY2, IM2, ID2, I
integer, dimension(4) :: IHMSF1
real(f8) :: A1, B1, FD, DAT1, W, DAT2, DDT

!=== Executable statements:

! The two-part JD.
A1 = D1
B1 = D2

! Provisional calendar date.
call JD_to_G_date_time(A1, B1, IY1, IM1, ID1, FD, JS)
if (JS /= 0) then
   IF (JS < 0) JS = -1
   ierr = JS; return
end if

! Is this a leap second day?
LEAP = .FALSE.
IF (SCALE == 'UTC') THEN

   ! TAI-UTC today.
   call TAI_minus_UTC_time(IY1, IM1, ID1, FD, DAT1, JS)
   if (JS < 0) then
      IF (JS < 0) JS = -1
      ierr = JS; return
   end if

   ! TAI-UTC tomorrow (at noon, to avoid rounding effects).
   call JD_to_G_date_time(A1+1.5_f8, B1-FD, IY2, IM2, ID2, W, JS)
   call TAI_minus_UTC_time(IY2, IM2, ID2, ZERO_f8, DAT2, JS)
   if (JS < 0) then
      IF (JS < 0) JS = -1
      ierr = JS; return
   end if

   ! The change in TAI-UTC (seconds).
   DDT = DAT2 - DAT1

   ! If leap second day, scale the fraction of a day into SI.
   LEAP = ABS(DDT) > 0.5_f8
   IF (LEAP) FD = FD + FD*DDT/D2S
END IF

! Provisional time of day.
call decimal_days_to_components(NDP, FD, S, IHMSF1)

! Is this a leap second day?
IF (.NOT.LEAP) THEN

   ! No.  Has the time rounded up to 24h?
   IF (IHMSF1(1) > 23) THEN
      
      ! Yes.  We will need tomorrow's calendar date.
      call JD_to_G_date_time(A1+1.5_f8, B1-FD, IY2, IM2, ID2, W, JS)
      
      ! Use 0h tomorrow.
      IY1 = IY2
      IM1 = IM2
      ID1 = ID2
      do I=1,4
         IHMSF1(I) = 0
      end do
   END IF
ELSE

! This is a leap second day.  Has the time reached or passed 24h?
   IF (IHMSF1(1) > 23) THEN

      ! Yes.  Use 23 59 60...
      IHMSF1(1) = 23
      IHMSF1(2) = 59
      IHMSF1(3) = 60
   END IF
END IF

! Results.
IY = IY1
IM = IM1
ID = ID1
do I=1,4
   IHMSF(I) = IHMSF1(I)
end do

! Status.
IF (JS < 0) JS = -1
ierr = JS

!  Excerpted from the original SOFA Software License:
!
!  3. You (the user) may copy and distribute SOFA source code to others,
!     and use and adapt its code and algorithms in your own software,
!     on a world-wide, royalty-free basis.  That portion of your
!     distribution that does not consist of intact and unchanged copies
!     of SOFA source code files is a "derived work" that must comply
!     with the following requirements:
!
!     a) Your work shall be marked or carry a statement that it
!        (i) uses routines and computations derived by you from
!        software provided by SOFA under license to you; and
!        (ii) does not itself constitute software provided by and/or
!        endorsed by SOFA.
!
!     b) The source code of your derived work must contain descriptions
!        of how the derived work is based upon, contains and/or differs
!        from the original SOFA software.
!
!     c) The name(s) of all routine(s) in your derived work shall not
!        include the prefix "iau".
!
!     d) The origin of the SOFA components of your derived work must
!        not be misrepresented;  you must not claim that you wrote the
!        original software, nor file a patent application for SOFA
!        software or algorithms embedded in the SOFA software.
!
!     e) These requirements must be reproduced intact in any source
!        distribution and shall apply to anyone to whom you have
!        granted a further right to modify the source code of your
!        derived work.
!
!     Note that, as originally distributed, the SOFA software is
!     intended to be a definitive implementation of the IAU standards,
!     and consequently third-party modifications are discouraged.  All
!     variations, no matter how minor, must be explicitly marked as
!     such, as explained above.

end subroutine spec_fmt_to_G_date_time


end module univ_Earth_date_time_mod
