!```````````````````````````````````````````````````````````````````
module univ_exception_mod

!+ Contains routines that facilitate the handling/reporting of program
!   "exceptions" (i.e., errors, warnings)

implicit none

integer, parameter :: WARN = 1, FATAL = 2 

interface exception
   module procedure exception_stdout
   module procedure exception_auxstdout
end interface exception


contains


!```````````````````````````````````````````````````````````````````
subroutine exception_stdout(calling_proc_name, flag, message)

!+ Routine that facilitates the handling/reporting of program "exceptions";
!   writes to stdout only.

use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_MESSAGE_LEN

implicit none

!=== Argument declarations:

character(len=*), intent(IN) :: calling_proc_name
integer, intent(IN) :: flag
character(len=*), optional, intent(IN) :: message

!== Local declarations:

character(len=MAX_MESSAGE_LEN) :: output_buffer, msg
integer :: ic1, ic2, aux_IOunit
logical :: program_should_abort
logical :: use_std, use_aux

!=== Executable statements:

! Limit input character strings to "reasonable" lengths:
ic1 = min(MAX_PROC_NAME_LEN, len_trim(calling_proc_name))
if (present(message)) then
   ic2 = min(MAX_MESSAGE_LEN, len_trim(message))
   msg = ' ::: '//message(1:ic2); ic2 = ic2+5
else
   ic2 = 1; msg = ' '
end if

! Populate output buffer; set abort toggle:
select case (flag)
case (WARN)
   write (output_buffer, '(4a)') '=== In procedure ',  &
        calling_proc_name(1:ic1), ', *WARNING*', msg(1:ic2)
   program_should_abort = .FALSE.
case (FATAL)
   write (output_buffer, '(4a)') '=== In procedure ',  &
        calling_proc_name(1:ic1), ', *FATAL ERROR*', msg(1:ic2)
   program_should_abort = .TRUE.
end select

! Write contents of output buffer to stdout:
write (*, '(/,a,/)') trim(output_buffer)

! Abort program execution, if directed to do so:
if (program_should_abort) stop

end subroutine exception_stdout


!```````````````````````````````````````````````````````````````````
subroutine exception_auxstdout(calling_proc_name, flag, IOunit_flag, message)

!+ Routine that facilitates the handling/reporting of program "exceptions";
!   writes to stdout and/or an "auxiliary" file unit.
!
!    IOunit_flag = 0  :  Write to stdout only
!    IOunit_flag < 0  :  Write only to file unit abs(IOunit_flag)
!    IOunit_flag > 0  :  Write to *both* stdout and file unit IOunit_flag

use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_MESSAGE_LEN

implicit none

!=== Argument declarations:

character(len=*), intent(IN) :: calling_proc_name
integer, intent(IN) :: flag
integer, intent(IN) :: IOunit_flag
character(len=*), optional, intent(IN) :: message

!== Local declarations:

character(len=MAX_MESSAGE_LEN) :: output_buffer, msg
integer :: ic1, ic2, aux_IOunit
logical :: program_should_abort
logical :: use_std, use_aux

!=== Executable statements:

! Determine some control parameters:
use_std = ( IOunit_flag >= 0 )
use_aux = ( IOunit_flag /= 0 )
if (use_aux) aux_IOunit = abs(IOunit_flag)

! Limit input character strings to "reasonable" lengths:
ic1 = min(MAX_PROC_NAME_LEN, len_trim(calling_proc_name))
if (present(message)) then
   ic2 = min(MAX_MESSAGE_LEN, len_trim(message))
   msg = ' ::: '//message(1:ic2); ic2 = ic2+5
else
   ic2 = 1; msg = ' '
end if

! Populate output buffer; set abort toggle:
select case (flag)
case (WARN)
   write (output_buffer, '(4a)') '=== In procedure ',  &
        calling_proc_name(1:ic1), ', *WARNING*', msg(1:ic2)
   program_should_abort = .FALSE.
case (FATAL)
   write (output_buffer, '(4a)') '=== In procedure ',  &
        calling_proc_name(1:ic1), ', *FATAL ERROR*', msg(1:ic2)
   program_should_abort = .TRUE.
end select

! Write contents of output buffer to stdout and/or an "auxiliary" file unit:
if (use_std) write (*, '(/,a,/)') trim(output_buffer)
if (use_aux) write (aux_IOunit, '(/,a,/)') trim(output_buffer)

! Abort program execution, if directed to do so:
if (program_should_abort) stop

end subroutine exception_auxstdout


end module univ_exception_mod
