!```````````````````````````````````````````````````````````````````
module univ_filelist_mod

!+ Contains routines related to obtaining lists of filenames

!~ 02.04.2010 (last modified; TIM)

implicit none

private

public :: filelist_count, filelist

contains


!```````````````````````````````````````````````````````````````````
subroutine filelist_count(file_prefix, n_files)

!+ Counts how many files with a given prefix exist

!~ 24.01.2012 (last modified; TIM)

use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_FILE_PATH_LEN

implicit none

!=== Argument declarations:

character(len=*), intent(IN) :: file_prefix

integer, intent(OUT) :: n_files

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'filelist_count'

character(len=MAX_FILE_PATH_LEN) :: f_pfx_tmp
integer :: pfx_len, err, n_f_tmp

! Explicit interface for C routine (to help the compiler catch errors):
interface
   subroutine univ_filematch_count_f(e_arg1, e_arg2, e_arg3, e_arg4)
      integer, intent(INOUT) :: e_arg1, e_arg3, e_arg4
      character(len=*), intent(INOUT) :: e_arg2
   end subroutine univ_filematch_count_f
end interface

!=== Executable statements:

pfx_len = min(len_trim(file_prefix), MAX_FILE_PATH_LEN)
write (*,'(/,4a)') '{', trim(procedure_name), '}: Checking prefix: ',  &
     file_prefix(1:pfx_len)

! Search for and count files that match prefix:

! `univ_filematch_count_f` is implemented as a C routine:
f_pfx_tmp = file_prefix   ! local copy of intent(IN) argument
call univ_filematch_count_f(pfx_len, f_pfx_tmp, n_f_tmp, err)
n_files = n_f_tmp   ! copy to intent(OUT) argument

! File search was UNsuccessful for some reason; specify that zero matching
!  files were found: 
if (err /= 0) n_files = 0

end subroutine filelist_count


!```````````````````````````````````````````````````````````````````
subroutine filelist(f_names, file_prefix, n_files)

!+ Returns the names of all files with the given prefix 

!~ 2013-08-14 (last modified; TIM)

use univ_maxlen_string_mod, only:  MAX_FILE_NAME_LEN, MAX_FILE_PATH_LEN
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_string_mod, only: IO_fmt_for_string
use univ_std_fileIO_mod, only: find_file_IOunit

implicit none

!=== Argument declarations:

integer, intent(IN) :: n_files
character(len=*), intent(IN) :: file_prefix

character(len=*), dimension(n_files), intent(OUT) :: f_names

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'filelist'

character(len=MAX_FILE_PATH_LEN) :: file_name, f_pfx_tmp
character(len=MAX_FILE_NAME_LEN) :: tmpname, tmp_str
integer :: cnt, nf, in_unit, pfx_len, tmp_len, err
logical :: file_exists

! Explicit interfaces for C routines (to help the compiler catch errors):
interface
   subroutine univ_filematch_f(e_arg1, e_arg2, e_arg3, e_arg4, e_arg5)
      integer, intent(INOUT) :: e_arg1, e_arg3, e_arg5
      character(len=*), intent(INOUT) :: e_arg2, e_arg4
   end subroutine univ_filematch_f
end interface
interface
   subroutine univ_remove_f(e_arg1, e_arg2, e_arg3)
      integer, intent(INOUT) :: e_arg1, e_arg3
      character(len=*), intent(INOUT) :: e_arg2
   end subroutine univ_remove_f
end interface

!=== Executable statements:

pfx_len = min(len_trim(file_prefix), MAX_FILE_PATH_LEN)
      
! Create a unique filename (to be used for list of match files):

tmpname = './tmpufc_fl_000.list'; tmp_len = len_trim(tmpname)

cnt = 0
do
   cnt = cnt+1
   write (tmpname(13:15), '(i3.3)') cnt  ! Change temporary filename
   inquire(file=tmpname(1:tmp_len), exist=file_exists)
   if (.NOT. file_exists) exit
end do

! Search for and write out a list of files that match pattern:

! `univ_filematch_f` is implemented as a C routine:
f_pfx_tmp = file_prefix   ! local copy of intent(IN) argument
call univ_filematch_f(pfx_len, f_pfx_tmp, tmp_len, tmpname, err)

! Open the temporary list file and read through the entries:

if (err == 0) then  ! File search was successful

   in_unit = find_file_IOunit()
   open (in_unit, file=tmpname(1:tmp_len), status='old')
   tmp_str = IO_fmt_for_string(MAX_FILE_PATH_LEN, .TRUE.)
   do nf=1,n_files
      read (in_unit, trim(tmp_str)) file_name
      f_names(nf) = trim(file_name)
   end do
     
   close (in_unit)

   ! Remove the temporary file:
   call univ_remove_f(tmp_len, tmpname, err)

end if

end subroutine filelist


end module univ_filelist_mod
