!```````````````````````````````````````````````````````````````````
module univ_integer_time_mod

!+ Contains routines related to date and time calculations

!~ 28.12.2010 (last modified; TIM)

use univ_kind_defs_mod, only: i8, i4, f8

implicit none

private

logical :: module_time_is_initialized = .FALSE.

!==== Public methods provided by this module:
!
! {(Time1 <operator> Time2) = Time_answer}:
!      (+), (-)
! {(Time1 <operator> Time2) = boolean_answer}:
!      (>), (>=), (==), (/=), (<), (<=)
! {(Time * i_factor) = Time_answer}, {(i_factor * Time) = Time_answer}:
!      (*)
! {(Time / i_factor) = Time_answer}, {(Time1 / Time2) = i_answer}:
!      (/)
! {(Time1 // Time2) = fp_answer}:
!      (//)
!
! time_eqv_fp_seconds  :  {Time --> fp_answer}
! fp_seconds_eqv_time  :  {fp_seconds --> Time}
! get_ticks_per_second  :  extract 'ticks_per_second' from Time
! get_time_component_values  :  extract 'seconds' and 'ticks' from Time
! time_eqv_brel_days  :  {Time --> (i_days, fp_day_fraction)}
!====

! Time can be accurately stored using a type(time_obj) object, comprised of
!  three 8-byte integer components (counters):
!
!    'seconds'  :  SI seconds; full use of the 8-byte integer representation is
!                   currently permitted
!    'ticks'  :  A "tick" is the smallest increment of time tracked by this
!                 module, and equals (1.)/'ticks_per_second' SI seconds; note
!                 that due to the use of 8-byte floating point variables for
!                 some operations, 'ticks' must only contain a value that
!                 a 4-byte integer can represent -- someday this restriction
!                 may be able to be lifted by using 16-byte floating point ops.
!    'ticks_per_second'  :  Number of ticks per second. Must be 10^n,
!                            where n is an integer.  Can currently only have a
!                            "4-byte-integer-type" value (thus 0 <= n <= 9).
!
! === Rules to enforce (by design):
! * 'ticks' should never be less than 0, since a negative tick count can always
!   be represented by decrementing the 'seconds' counter and calculating an
!   appropriate positive tick count.
! * OUTSIDE this module, 'seconds' should never be negative
! * 'seconds' can be negative WITHIN this module for calculation expediency

integer(i8), parameter :: default_tps = 1000000000   ! 1 tick = 1 nanosecond

type time_obj
   private
   integer(i8) :: seconds     ! SI unit of time
   integer(i8) :: ticks   ! See 'ticks_per_second'
   integer(i8) :: ticks_per_second
end type time_obj
public :: time_obj

! Maximum integer values of the 'seconds' and 'ticks' components of a
!  type(time_obj) object:
integer(i8), parameter :: seconds_huge = huge(0_i8),  &
                          ticks_huge = huge(0_i8)

integer(i4), parameter :: i4_huge = huge(0_i4)

! Operator overloading interfaces:

interface operator (+)
   module procedure time_plus
end interface
interface operator (-)
   module procedure time_minus
end interface
interface operator (*)
   module procedure time_scalar_mult_i4, time_scalar_mult_i8,  &
        time_scalar_mult_i4_transposed, time_scalar_mult_i8_transposed
end interface
interface operator (/)
   module procedure time_scalar_divide_i4, time_scalar_divide_i8,  &
        time_divide_imethod
end interface
interface operator (//)
   module procedure time_divide_fmethod
end interface
interface operator (>)
   module procedure time_gt
end interface
interface operator (>=)
   module procedure time_ge
end interface
interface operator (<)
   module procedure time_lt
end interface
interface operator (<=)
   module procedure time_le
end interface
interface operator (==)
   module procedure time_eq
end interface
interface operator (/=)
   module procedure time_ne
end interface
public :: operator(+), operator(-), operator(*), operator(/), operator(>),  &
     operator(>=), operator(==), operator(/=), operator(<), operator(<=),  &
     operator(//)

public :: time_eqv_fp_seconds, fp_seconds_eqv_time, get_ticks_per_second,  &
     get_time_component_values, time_eqv_brel_days, set_time, time_modulo

contains


!```````````````````````````````````````````````````````````````````
subroutine set_time_private(seconds, ticks, Time_out, ticks_per_second,  &
     allow_negative_time)

!+ Returns a time interval corresponding to the given number of seconds, ticks,
!   and ticks_per_second.  Values of 'seconds' and/or 'ticks' may be negative,
!   but the resulting time interval must conform to the following:
!
!     allow_negative_time = .FALSE. :
!        0 <= Time_out%seconds <= seconds_huge
!        0 <= Time_out%ticks < ticks_per_second
!
!     allow_negative_time = .TRUE. :
!        -(seconds_huge) < Time_out%seconds <= seconds_huge
!        0 <= Time_out%ticks < ticks_per_second
!
!  NOTE: 
!    * This algorithm tries to avoid unnecessary integer overflow conditions.

!~ 22.12.2010 (last modified; TIM)

use univ_kind_defs_mod, only: i8
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_exception_mod, only: exception, FATAL

implicit none

!=== Argument declarations:

integer(i8), intent(IN) :: seconds, ticks, ticks_per_second
logical, optional, intent(IN) :: allow_negative_time

type(time_obj), intent(OUT) :: Time_out

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'set_time_private'

logical :: negative_time_allowed
integer(i8) :: seconds_new, ticks_new, huge_inc, seconds_inc

!=== Executable statements:

! Process any optional arguments:
negative_time_allowed = .FALSE.
if (present(allow_negative_time)) negative_time_allowed = allow_negative_time

seconds_inc = ticks/ticks_per_second-  &
     dim(0_i8, sign(1_i8, mod(ticks, ticks_per_second)))
ticks_new = modulo(ticks, ticks_per_second)

huge_inc = seconds_huge-abs(seconds)  ! +/- huge clearance
if (sign(1_i8, seconds)*seconds_inc <= huge_inc) then
   seconds_new = seconds+seconds_inc
else    ! Unavoidable integer overflow condition
   call exception(procedure_name, FATAL, 'seconds overflow condition')
end if

if (ticks_new < 0_i8) then    ! Procedure sanity check
   call exception(procedure_name, FATAL,  &
        'resultant ticks < 0; this should not happen; fix me')
end if

if (.NOT. negative_time_allowed .and. seconds_new < 0_i8) then
   call exception(procedure_name, FATAL,  &
        'resultant seconds < 0, but allow_negative_time = .FALSE.')
else
   Time_out%seconds = seconds_new
   Time_out%ticks = ticks_new
   Time_out%ticks_per_second = ticks_per_second
end if

end subroutine set_time_private


!```````````````````````````````````````````````````````````````````
subroutine increment_time_private(Time_in, seconds, ticks, Time_out)

!+ Increments a time interval by seconds and ticks

!~ 22.12.2010 (last modified; TIM)

use univ_kind_defs_mod, only: i8
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_exception_mod, only: exception, FATAL

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time_in
integer(i8), intent(IN) :: seconds, ticks

type(time_obj), intent(OUT) :: Time_out

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'increment_time_private'

integer(i8) :: tgt_inc, loop_inc, t_units, tmp_seconds, tmp_ticks
logical :: int_overflow

!=== Executable statements:

Time_out%seconds = 0_i8; Time_out%ticks = Time_in%ticks

if (ticks <= ticks_huge-Time_out%ticks) then
   tmp_seconds = Time_out%seconds; tmp_ticks = Time_out%ticks
   call set_time_private(tmp_seconds, tmp_ticks+ticks, Time_out,  &
        Time_in%ticks_per_second, allow_negative_time=.TRUE.)
else
   tgt_inc = ticks
   do
      if (tgt_inc == 0_i8) exit
      loop_inc = sign(1_i8, tgt_inc)*min(tgt_inc, ticks_huge-Time_out%ticks)
      tgt_inc = tgt_inc-loop_inc
      t_units = Time_out%ticks+loop_inc
      tmp_seconds = Time_out%seconds
      call set_time_private(tmp_seconds, t_units, Time_out,  &
           Time_in%ticks_per_second, allow_negative_time=.TRUE.)
   end do
end if

int_overflow = .FALSE.

! Check for partial expression overflow (seconds+Time_in%seconds):
if (sign(1_i8, seconds) == sign(1_i8, Time_in%seconds)) then
   int_overflow = ( abs(seconds) >  &
        seconds_huge-sign(1_i8, Time_in%seconds)*Time_in%seconds )
end if

! Check for full expression overflow:
if (.NOT. int_overflow) then
   int_overflow = ( abs(seconds+Time_in%seconds) >  &
        seconds_huge-sign(1_i8, Time_out%seconds)*Time_out%seconds )
else   ! Last chance for a overflow reprieve
   if (sign(1_i8, seconds) /= sign(1_i8, Time_out%seconds)) then
      int_overflow = ( abs(seconds+Time_out%seconds) >  &
           seconds_huge-sign(1_i8, Time_in%seconds)*Time_in%seconds )
   end if
end if
  
if (.NOT. int_overflow) then
   t_units = (seconds+Time_out%seconds)+Time_in%seconds
   tmp_ticks = Time_out%ticks
   call set_time_private(t_units, tmp_ticks, Time_out, Time_in%ticks_per_second)
else
   call exception(procedure_name, FATAL, 'seconds overflow condition')
end if

end subroutine increment_time_private


!```````````````````````````````````````````````````````````````````
function set_time(seconds, ticks, ticks_per_second, body_rel_days,  &
     body_rel_hours, body_rel_minutes, body_rel_daylen_fp_rep,  &
     body_rel_daylen)  result(Time_new)

!+  Public method; Given some number of seconds, ticks, and ticks_per_second
!    (and optionally body-relevant days/hours/minutes/day_length) returns a
!    type(time_obj) object that most closely approximates the input.
!
!   NOTES:
!      * The resulting time must be positive (possible even if some of the
!        inputs are negative).
!      * If a 'body_rel_*' optional input field is given, EITHER the
!        'body_rel_daylen' [type(time_obj)] OR 'body_rel_daylen_fp_len'
!        (8-byte floating point value [s]).

!~ 19.01.2011 (last modified; TIM)

use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_MESSAGE_LEN
use univ_exception_mod, only: exception, FATAL

implicit none

!=== Argument declarations:

integer(i4), intent(IN) :: seconds, ticks, ticks_per_second
integer(i4), intent(IN), optional :: body_rel_days, body_rel_hours,  &
     body_rel_minutes
real(f8), intent(IN), optional :: body_rel_daylen_fp_rep
type(time_obj), intent(IN), optional :: Body_rel_daylen

type(time_obj) :: Time_new     ! Procedure result

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'set_time'

character(len=MAX_MESSAGE_LEN) :: msg
integer(i8) :: work_brel_days, work_brel_hours, work_brel_minutes
integer(i8) :: work_ticks, work_seconds, sign_tmp, work_tps
type(time_obj) :: Time_tmp1, Time_tmp2, Work_brel_daylen
logical :: negtime_tog

!=== Executable statements:

if (.NOT. module_time_is_initialized) call module_time_init

work_seconds = int(seconds, kind=i8)
work_ticks = int(ticks, kind=i8)
work_tps = int(ticks_per_second, kind=i8)

if (present(body_rel_days) .or. present(body_rel_hours) .or.  &
     present(body_rel_minutes)) then
   if (present(Body_rel_daylen)) then
      if (Body_rel_daylen%ticks_per_second /= work_tps)  &
           call exception(procedure_name, FATAL,  &
           "'ticks_per_second' value must match that within 'Body_rel_daylen'.")
      Work_brel_daylen = Body_rel_daylen
   else if (present(body_rel_daylen_fp_rep)) then
      Work_brel_daylen = fp_seconds_eqv_time(body_rel_daylen_fp_rep,  &
           int(work_tps, kind=i4))
   else
      call exception(procedure_name, FATAL,  &
           "must provide either 'Body_rel_daylen' or 'body_rel_daylen_fp_rep'.")
   end if
end if

work_brel_days = 0_i8; if (present(body_rel_days))  &
     work_brel_days = int(body_rel_days, kind=i8)
work_brel_hours = 0_i8; if (present(body_rel_hours))  &
     work_brel_hours = int(body_rel_hours, kind=i8)
work_brel_minutes = 0_i8; if (present(body_rel_minutes))  &
     work_brel_minutes = int(body_rel_minutes, kind=i8)

! Enforce constraint that input time interval integers must contain "small"
!  (4_byte-like) values:
if (abs(max(work_seconds, work_ticks)) > int(i4_huge, kind=i8) .or.  &
     abs(max(work_brel_days, work_brel_hours,  &
     work_brel_minutes)) > int(i4_huge, kind=i8)) then
   write (msg, '(a,i10)') 'input time interval values must each be <= ', i4_huge
   call exception(procedure_name, FATAL, trim(msg))
end if
 
negtime_tog = ( work_brel_days /= 0_i8 )
call set_time_private(work_seconds, work_ticks, Time_new, work_tps, negtime_tog)

if (work_brel_days /= 0_i8) then
   Time_tmp1 = Work_brel_daylen*abs(work_brel_days)
   sign_tmp = sign(1_i8, work_brel_days)
   Time_tmp2 = Time_new
   call increment_time_private(Time_tmp2, sign_tmp*Time_tmp1%seconds,  &
        sign_tmp*Time_tmp1%ticks, Time_new)
end if

if (work_brel_hours /= 0_i8) then
   Time_tmp1 = Work_brel_daylen/24_i8*abs(work_brel_hours)
   sign_tmp = sign(1_i8, work_brel_hours)
   Time_tmp2 = Time_new
   call increment_time_private(Time_tmp2, sign_tmp*Time_tmp1%seconds,  &
         sign_tmp*Time_tmp1%ticks, Time_new)
end if

if (work_brel_minutes /= 0_i8) then
   Time_tmp1 = Work_brel_daylen/1440_i8*abs(work_brel_minutes)
   sign_tmp = sign(1_i8, work_brel_minutes)
   Time_tmp2 = Time_new
   call increment_time_private(Time_tmp2, sign_tmp*Time_tmp1%seconds,  &
        sign_tmp*Time_tmp1%ticks, Time_new)
end if

end function set_time


!```````````````````````````````````````````````````````````````````
function time_gt(Time1, Time2)  result(answer)

!+ Semi-public method [externally accessible only as the '>' operator for
!   type(time_obj) items] that returns .TRUE. if Time1 > Time2, otherwise
!   returns .FALSE.

!~ 22.12.2010 (last modified; TIM)

use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_exception_mod, only: exception, FATAL

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time1, Time2

logical :: answer      ! Procedure result

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'time_gt'

!=== Executable statements:

if (Time1%ticks_per_second /= Time2%ticks_per_second)  &
     call exception(procedure_name, FATAL,  &
     'ticks_per_second values of arguments must match')

if (Time1%seconds == Time2%seconds) then
   answer = ( Time1%ticks > Time2%ticks )
else
   answer = ( Time1%seconds > Time2%seconds )
end if

end function time_gt


!```````````````````````````````````````````````````````````````````
function time_lt(Time1, Time2)  result(answer)

!+ Semi-public method [externally accessible only as the '<' operator for
!   type(time_obj) items] that returns .TRUE. if Time1 < Time2, otherwise
!   returns .FALSE.

!~ 22.12.2010 (last modified; TIM)

use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_exception_mod, only: exception, FATAL

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time1, Time2

logical :: answer      ! Procedure result

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'time_lt'

!=== Executable statements:

if (Time1%ticks_per_second /= Time2%ticks_per_second)  &
     call exception(procedure_name, FATAL,  &
     'ticks_per_second values of arguments must match')

if (Time1%seconds == Time2%seconds) then
   answer = ( Time1%ticks < Time2%ticks )
else
   answer = ( Time1%seconds < Time2%seconds )
end if

end function time_lt


!```````````````````````````````````````````````````````````````````
function time_eq(Time1, Time2)  result(answer)

!+ Semi-public method [externally accessible only as the '==' operator for
!   type(time_obj) items] that returns .TRUE. if Time1 == Time2, otherwise
!   returns .FALSE.

!~ 22.12.2010 (last modified; TIM)

use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_exception_mod, only: exception, FATAL

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time1, Time2

logical :: answer      ! Procedure result

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'time_eq'

!=== Executable statements:

if (Time1%ticks_per_second /= Time2%ticks_per_second)  &
     call exception(procedure_name, FATAL,  &
     'ticks_per_second values of arguments must match')

answer = ( Time1%seconds == Time2%seconds .and. Time1%ticks == Time2%ticks )

end function time_eq


!```````````````````````````````````````````````````````````````````
function time_ge(Time1, Time2)  result(answer)

!+ Semi-public method [externally accessible only as the '>=' operator for
!   type(time_obj) items] that returns .TRUE. if Time1 >= Time2, otherwise
!   returns .FALSE.

!~ 22.12.2010 (last modified; TIM)

!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time1, Time2

logical :: answer      ! Procedure result

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'time_ge'

!=== Executable statements:

answer = (time_gt(Time1, Time2) .or. time_eq(Time1, Time2))

end function time_ge


!```````````````````````````````````````````````````````````````````
function time_le(Time1, Time2)  result(answer)

!+ Semi-public method [externally accessible only as the '<=' operator for
!   type(time_obj) items] that returns .TRUE. if Time1 <= Time2, otherwise
!   returns .FALSE.

!~ 22.12.2010 (last modified; TIM)

!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time1, Time2

logical :: answer      ! Procedure result

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'time_le'

!=== Executable statements:

answer = (time_lt(Time1, Time2) .or. time_eq(Time1, Time2))

end function time_le


!```````````````````````````````````````````````````````````````````
function time_ne(Time1, Time2)  result(answer)

!+ Semi-public method [externally accessible only as the '/=' operator for
!   type(time_obj) items] that returns .TRUE. if Time1 /= Time2, otherwise
!   returns .FALSE.

!~ 22.12.2010 (last modified; TIM)

!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time1, Time2

logical :: answer      ! Procedure result

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'time_ne'

!=== Executable statements:

answer = ( .NOT. time_eq(Time1, Time2) )

end function time_ne


!```````````````````````````````````````````````````````````````````
function time_plus(Time1, Time2)  result(Time_ans)

!+ Semi-public method [externally accessible only as the '+' operator for
!   type(time_obj) items] that returns the traditional sum of Time1 and Time2.

!~ 22.12.2010 (last modified; TIM)

use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_exception_mod, only: exception, FATAL

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time1, Time2

type(time_obj) :: Time_ans      ! Procedure result

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'time_plus'

!=== Executable statements:

if (.NOT. module_time_is_initialized) call module_time_init

if (Time1%ticks_per_second /= Time2%ticks_per_second)  &
     call exception(procedure_name, FATAL,  &
     'ticks_per_second values of arguments must match')

call increment_time_private(Time1, Time2%seconds, Time2%ticks, Time_ans)

end function time_plus


!```````````````````````````````````````````````````````````````````
function time_minus(Time1, Time2)  result(Time_ans)

!+ Semi-public method [externally accessible only as the '-' operator for
!   type(time_obj) items] that returns the traditional difference of
!   Time1 and Time2.
!
!  NOTE: A type(time_obj) object holds a positive time (interval), so by
!        definition, (Time1 - Time2) is the same as (Time2 - Time1).

!~ 22.12.2010 (last modified; TIM)

use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_exception_mod, only: exception, FATAL

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time1, Time2

type(time_obj) :: Time_ans      ! Procedure result

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'time_minus'

!=== Executable statements:

if (.NOT. module_time_is_initialized) call module_time_init

if (Time1%ticks_per_second /= Time2%ticks_per_second)  &
     call exception(procedure_name, FATAL,  &
     'ticks_per_second values of arguments must match')

if (Time1 > Time2) then
   call increment_time_private(Time1, -Time2%seconds, -Time2%ticks, Time_ans)
else 
   call increment_time_private(Time2, -Time1%seconds, -Time1%ticks, Time_ans)
end if

end function time_minus


!```````````````````````````````````````````````````````````````````
function time_scalar_mult_i4(Time, i4_factor)  result(Time_ans)

!+ Semi-public method [externally accessible only as the '*' operator for
!   type(time_obj) items] that returns Time multiplied by a given integer
!   factor.
!
!  NOTES: 
!       * When using operator overloading, this procedure will be invoked only
!         in the case {Time*i4_factor} (NOT for the case {i4_factor*Time})
!       * For 4-byte integer factor

!~ 23.12.2010 (last modified; TIM)

use univ_kind_defs_mod, only: i8, i4
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_MESSAGE_LEN
use univ_exception_mod, only: exception, FATAL

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time
integer(i4), intent(IN) :: i4_factor

type(time_obj) :: Time_ans      ! Procedure result

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'time_scalar_mult_i4'

character(len=MAX_MESSAGE_LEN) :: msg
type(time_obj) :: Time_tmp
integer(i8) :: i8_factor, loop_inc, tgt_inc, t_units, tmp_seconds

!=== Executable statements:

if (.NOT. module_time_is_initialized) call module_time_init

! Integer factor must be positive, since a type(time_obj) object holds a
!  positive time (interval):
if (i4_factor < 0_i4) call exception(procedure_name, FATAL,  &
     'scalar integer factor must be positive (> 0)')

i8_factor = int(i4_factor, kind=i8)

if (i8_factor <= 1_i8) then
   call set_time_private(Time%seconds*i8_factor, Time%ticks*i8_factor,  &
        Time_ans, Time%ticks_per_second)
   return    ! Nothing more to do here
end if

! Enforce constraint that input integer factor must contain "small"
!  (4_byte-like) value:
if (i8_factor > int(i4_huge, kind=i8)) then
   write (msg, *) 'scalar integer factor must be <= ', i4_huge
   call exception(procedure_name, FATAL, trim(msg))
end if

if (i8_factor <= ticks_huge/max(1_i8, Time%ticks)) then
   call set_time_private(0_i8, Time%ticks*i8_factor, Time_tmp,  &
        Time%ticks_per_second, allow_negative_time=.TRUE.)
else
   tgt_inc = i8_factor
   Time_tmp%seconds = 0_i8; Time_tmp%ticks = 0_i8
   do
      if (tgt_inc == 0_i8) exit
      loop_inc = sign(1_i8, tgt_inc)*  &
           min(tgt_inc, (ticks_huge-Time_tmp%ticks)/max(1_i8, Time%ticks))
      tgt_inc = tgt_inc-loop_inc
      t_units = Time_tmp%ticks+Time%ticks*loop_inc
      tmp_seconds = Time_tmp%seconds
      call set_time_private(tmp_seconds, t_units, Time_tmp,  &
           Time%ticks_per_second, allow_negative_time=.TRUE.)
   end do
end if

if (i8_factor <= (seconds_huge-Time_tmp%seconds)/max(1_i8, Time%seconds)) then
   t_units = Time_tmp%seconds+Time%seconds*i8_factor
   call set_time_private(t_units, Time_tmp%ticks, Time_ans,  &
        Time%ticks_per_second)
else
   call exception(procedure_name, FATAL,  &
        'scalar integer factor is too large to handle (seconds overflow)')
end if

end function time_scalar_mult_i4


!```````````````````````````````````````````````````````````````````
function time_scalar_mult_i8(Time, i8_factor)  result(Time_ans)

!+ Semi-public method [externally accessible only as the '*' operator for
!   type(time_obj) items] that returns Time multiplied by a given integer
!   factor.
!
!  NOTES: 
!      * When using operator overloading, this procedure will be invoked only
!        in the case {Time*i8_factor} (NOT for the case {i8_factor*Time})
!      * For 8-byte integer factor (must have value <= i4_huge)

!~ 23.12.2010 (last modified; TIM)

use univ_kind_defs_mod, only: i8, i4
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_MESSAGE_LEN
use univ_exception_mod, only: exception, FATAL

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time
integer(i8), intent(IN) :: i8_factor

type(time_obj) :: Time_ans      ! Procedure result

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'time_scalar_mult_i8'

character(len=MAX_MESSAGE_LEN) :: msg

!=== Executable statements:

! Enforce constraint that input integer must contain "small"
!  (4_byte-like) value:
if (i8_factor > int(i4_huge, kind=i8)) then
   write (msg, '(a,i10)') 'input integer factor must be <= ', i4_huge
   call exception(procedure_name, FATAL, trim(msg))
end if

Time_ans = time_scalar_mult_i4(Time, int(i8_factor, kind=i4))

end function time_scalar_mult_i8


!```````````````````````````````````````````````````````````````````
function time_scalar_mult_i4_transposed(i4_factor, Time)  result(Time_ans)

!+ Semi-public method [externally accessible only as the '*' operator for
!   type(time_obj) items] that returns Time multiplied by a given integer
!   factor.
!
!  NOTES: 
!       * When using operator overloading, this procedure will be invoked only
!         in the case {i4_factor*Time} (NOT for the case {Time*i4_factor})
!       * For 4-byte integer factor

!~ 23.12.2010 (last modified; TIM)

use univ_kind_defs_mod, only: i4
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time
integer(i4), intent(IN) :: i4_factor

type(time_obj) :: Time_ans      ! Procedure result

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'time_scalar_mult_i4_transposed'

!=== Executable statements:

Time_ans = time_scalar_mult_i4(Time, i4_factor)

end function time_scalar_mult_i4_transposed


!```````````````````````````````````````````````````````````````````
function time_scalar_mult_i8_transposed(i8_factor, Time)  result(Time_ans)

!+ Semi-public method [externally accessible only as the '*' operator for
!   type(time_obj) items] that returns Time multiplied by a given integer
!   factor.
!
!  NOTES: 
!       * When using operator overloading, this procedure will be invoked only
!         in the case {i8_factor*Time} (NOT for the case {Time*i8_factor})
!       * For 8-byte integer factor (must have value <= i4_huge)

!~ 23.12.2010 (last modified; TIM)

use univ_kind_defs_mod, only: i4, i8
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_MESSAGE_LEN
use univ_exception_mod, only: exception, FATAL

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time
integer(i8), intent(IN) :: i8_factor

type(time_obj) :: Time_ans      ! Procedure result

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'time_scalar_mult_i8_transposed'

character(len=MAX_MESSAGE_LEN) :: msg

!=== Executable statements:

! Enforce constraint that input integer must contain "small"
!  (4_byte-like) value:
if (i8_factor > int(i4_huge, kind=i8)) then
   write (msg, '(a,i10)') 'input integer factor must be <= ', i4_huge
   call exception(procedure_name, FATAL, trim(msg))
end if

Time_ans = time_scalar_mult_i4(Time, int(i8_factor, kind=i4))

end function time_scalar_mult_i8_transposed


!```````````````````````````````````````````````````````````````````
function time_divide_imethod(Time1, Time2)  result(i4_ans)

!+ Semi-public method [externally accessible only as the '/' operator for
!   type(time_obj) items] that returns the largest 4-byte integer, n, for which
!   Time1 >= Time2*n (equivalent to traditional integer division).

!~ 23.12.2010 (last modified; TIM)

use univ_kind_defs_mod, only: i8, i4, f8
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_exception_mod, only: exception, FATAL

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time1, Time2

integer(i4) :: i4_ans      ! Procedure result

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'time_divide_imethod'

real(f8) :: d1, d2
integer(i4) :: tgtmax, tgtmin, itgt

!=== Executable statements:

if (.NOT. module_time_is_initialized) call module_time_init

if (Time1%ticks_per_second /= Time2%ticks_per_second)  &
     call exception(procedure_name, FATAL,  &
     'ticks_per_second values of arguments must match')

if (Time2%seconds == 0_i8 .and. Time2%ticks == 0_i8) then
   call exception(procedure_name, FATAL, 'quotient error (denominator is zero)')
end if

if (Time1%seconds < 0_i8 .or. Time2%seconds < 0_i8) then
   call exception(procedure_name, FATAL,  &
        'quotient error (operands must be positive)')
end if

if (Time1 < Time2) then
   i4_ans = 0_i4
else if (Time1 == Time2) then
   i4_ans = 1_i4
else

   ! Big_time/tiny_time may yield integer overflow condition (in the
   !  'time_divide' output variable):
   if (Time1/huge(i4_ans) > Time2) then
      call exception(procedure_name, FATAL,  &
           'quotient would cause integer overflow condition')
   end if

   if (Time1%seconds == 0_i8) then     ! Compare using ticks
      ! Convert time intervals to floating point ticks:
      d1 = real(Time1%ticks, kind=f8)
      d2 = real(Time2%ticks, kind=f8)
      tgtmax = min(int(d1/d2, kind=i4)+1, i4_huge)
      tgtmin = max(1_i4, tgtmax-2)
   else   ! Compare using seconds
      ! Convert time intervals to floating point seconds:
      d1 = real(Time1%seconds, kind=f8)+  &
           Time1%ticks/real(Time1%ticks_per_second, kind=f8)
      d2 = real(Time2%seconds, kind=f8)+  &
           Time2%ticks/real(Time2%ticks_per_second, kind=f8)
      tgtmax = min(int(d1/d2, kind=i4)+1, i4_huge)
      tgtmin = max(1_i4, tgtmax-2)
   end if

   do itgt=tgtmax,tgtmin,-1
      if (time1/itgt >= time2) then
         i4_ans = itgt
         exit
      end if
   end do

end if

! Verify that 'integer_ans' is the correct answer:
if (Time2*i4_ans > Time1 .or. Time2*(i4_ans+1) <= Time1) then
   call exception(procedure_name, FATAL,  &
        'quotient error (inaccurate/imprecise calculation)')
end if

end function time_divide_imethod


!```````````````````````````````````````````````````````````````````
function time_divide_fmethod(Time1, Time2)  result(fp_ans)

!+ Semi-public method [externally accessible only as the '//' operator for
!   type(time_obj) items] that returns the 8-byte floating point quotient of
!   two time intervals.

!~ 22.12.2010 (last modified; TIM)

use univ_kind_defs_mod, only: f8
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_exception_mod, only: exception, FATAL

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time1, Time2

real(f8) :: fp_ans      ! Procedure result

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'time_divide_fmethod'

real(f8) :: d1, d2

!=== Executable statements:

if (.NOT. module_time_is_initialized) call module_time_init

if (Time1%ticks_per_second /= Time2%ticks_per_second)  &
     call exception(procedure_name, FATAL,  &
     'ticks_per_second values of arguments must match')

! Convert time intervals to floating point seconds; may be very imprecise if
!  8-byte integer values are in use:
d1 = real(Time1%seconds, kind=f8)+  &
     real(Time1%ticks, kind=f8)/real(Time1%ticks_per_second, kind=f8)
d2 = real(Time2%seconds, kind=f8)+  &
     real(Time2%ticks, kind=f8)/real(Time2%ticks_per_second, kind=f8)

fp_ans = d1/d2

end function time_divide_fmethod


!```````````````````````````````````````````````````````````````````
function time_modulo(Time1, Time2)  result(Time_ans)

!+ Public method that returns the equivalent of "mod(Time1, Time2)"

!~ 12.01.2011 (last modified; TIM)

!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time1, Time2

type(time_obj) :: Time_ans      ! Procedure result

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'time_modulo'

!=== Executable statements:

Time_ans = Time1-((Time1/Time2)*Time2)

end function time_modulo


!```````````````````````````````````````````````````````````````````
function time_eqv_fp_seconds(Time)  result(fp_ans)

!+ Public method that returns the 8-byte floating point equivalent [s] of the
!   given time interval
!
!  NOTE: May be very imprecise when 8-byte integer values are in use.

!~ 22.12.2010 (last modified; TIM)

use univ_kind_defs_mod, only: f8
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time

real(f8) :: fp_ans      ! Procedure result

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'time_eqv_fp_seconds'

!=== Executable statements:

if (.NOT. module_time_is_initialized) call module_time_init

fp_ans = real(Time%seconds, kind=f8)+  &
     real(Time%ticks, kind=f8)/real(Time%ticks_per_second, kind=f8)

end function time_eqv_fp_seconds


!```````````````````````````````````````````````````````````````````
subroutine time_eqv_brel_days(Time, brel_int_days, brel_frac_day,  &
     body_rel_daylen_fp_rep, body_rel_daylen)

!+ Public method that returns the body-relevant {integer days, 8-byte floating
!   point day fraction} equivalent of the given time interval
!
!  NOTES: 
!       * May be very imprecise when 8-byte integer values are in use.
!       * Must specify EITHER the 'body_rel_daylen' [type(time_obj)] OR
!         'body_rel_daylen_fp_len' (8-byte floating point value [s]).

!~ 24.12.2010 (last modified; TIM)

use univ_kind_defs_mod, only: f8, i4
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_exception_mod, only: exception, FATAL

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time
real(f8), intent(IN), optional :: body_rel_daylen_fp_rep
type(time_obj), intent(IN), optional :: Body_rel_daylen

integer(i4), intent(OUT) :: brel_int_days
real(f8), intent(OUT) :: brel_frac_day

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'time_eqv_brel_days'

type(time_obj) :: Work_brel_daylen, Time_tmp

!=== Executable statements:

if (.NOT. module_time_is_initialized) call module_time_init

if (present(Body_rel_daylen)) then
   if (Body_rel_daylen%ticks_per_second /= Time%ticks_per_second)  &
        call exception(procedure_name, FATAL,  &
    "'ticks_per_second' values within 'Time' and 'Body_rel_daylen' must match.")
   Work_brel_daylen = Body_rel_daylen
else if (present(body_rel_daylen_fp_rep)) then
   Work_brel_daylen = fp_seconds_eqv_time(body_rel_daylen_fp_rep,  &
        int(Time%ticks_per_second, kind=i4))
else
   call exception(procedure_name, FATAL,  &
        "must provide either 'Body_rel_daylen' or 'body_rel_daylen_fp_rep'.")
end if

brel_int_days = Time/Work_brel_daylen

Time_tmp = time_modulo(Time, Work_brel_daylen)
brel_frac_day = Time_tmp//Work_brel_daylen

end subroutine time_eqv_brel_days


!```````````````````````````````````````````````````````````````````
function fp_seconds_eqv_time(fp_seconds, ticks_per_second)  result(Time_ans)

!+ Public method that returns a type(time_obj) object as closely equivalent as
!   possible to the given ticks_per_second parameter and 8-byte floating point
!   value [s].
!
!  NOTE: Will be imprecise, particularly when the value of 'ticks_per_second'
!         is relatively small.

!~ 22.12.2010 (last modified; TIM)

use univ_kind_defs_mod, only: i8, f8, i4
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

real(f8), intent(IN) :: fp_seconds
integer(i4) :: ticks_per_second

type(time_obj) :: Time_ans      ! Procedure result

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'fp_seconds_eqv_time'

integer(i8) :: seconds, ticks
real(f8) :: real_ticks

!=== Executable statements:

if (.NOT. module_time_is_initialized) call module_time_init

seconds = int(fp_seconds, kind=i8)
real_ticks = fp_seconds-real(seconds, kind=f8)
ticks = nint(real_ticks*int(ticks_per_second, kind=i8))

call set_time_private(seconds, ticks, Time_ans, int(ticks_per_second, kind=i8))

end function fp_seconds_eqv_time


!```````````````````````````````````````````````````````````````````
function time_scalar_divide_i4(Time, i4_divisor)  result(Time_ans)

!+ Semi-public method [externally accessible only as the '/' operator for
!   type(time_obj) items] that returns the largest type(time_obj) object,
!   'Time_ans', for which i4_factor*Time_ans = Time
!
!  NOTE: For 4-byte integer divisor

!~ 23.12.2010 (last modified; TIM)

use univ_kind_defs_mod, only: i8, i4, f8
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_MESSAGE_LEN
use univ_exception_mod, only: exception, FATAL

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time
integer(i4), intent(IN) :: i4_divisor

type(time_obj) :: Time_ans      ! Procedure result

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'time_scalar_divide_i4'

character(len=MAX_MESSAGE_LEN) :: msg
integer(i8) :: seconds, ticks, rmnd_seconds, rmnd_ticks, i8_divisor
type(time_obj) :: Time_tmp
real(f8) :: ftmp
logical :: must_round

!=== Executable statements:

if (.NOT. module_time_is_initialized) call module_time_init

! Integer divisor must be positive, since a type(time_obj) object holds a
!  positive time (interval):
if (i4_divisor < 0_i4) call exception(procedure_name, FATAL,  &
     'scalar integer divisor must be positive (> 0)')

i8_divisor = int(i4_divisor, kind=i8)

! Enforce constraint that input scalar integer divisor must contain "small"
!  (4_byte-like) value:
if (i8_divisor > int(i4_huge, kind=i8)) then
   write (msg, *) 'scalar integer divisor must be <= ', i4_huge
   call exception(procedure_name, FATAL, trim(msg))
end if

seconds = Time%seconds/i8_divisor
rmnd_seconds = mod(Time%seconds, i8_divisor)
ticks = Time%ticks/i8_divisor
rmnd_ticks = mod(Time%ticks, i8_divisor)

call set_time_private(seconds, ticks, Time_ans, Time%ticks_per_second)

ftmp = Time%ticks_per_second*(rmnd_seconds/real(i8_divisor, kind=f8))+  &
     rmnd_ticks/real(i8_divisor, kind=f8)

! Round "very close" cases:

must_round = .FALSE.
if (abs(ftmp) > abs(tiny(ftmp))) then
   if (abs(ftmp-nint(ftmp)) <  &
        (10._f8**(1_i4+int(log10(ftmp), kind=i4)))*epsilon(ftmp)) then
      must_round = .TRUE.
   end if
end if

if (must_round) then
   ticks = nint(ftmp, kind=i8)
else
   ticks = int(ftmp, kind=i8)
end if

! Finalize result:

Time_tmp = Time_ans
call increment_time_private(Time_tmp, 0_i8, ticks, Time_ans)

end function time_scalar_divide_i4


!```````````````````````````````````````````````````````````````````
function time_scalar_divide_i8(Time, i8_divisor)  result(Time_ans)

!+ Semi-public method [externally accessible only as the '/' operator for
!   type(time_obj) items] that returns the largest type(time_obj) object,
!   'Time_ans', for which i4_factor*Time_ans = Time
!
!  NOTE: For 8-byte integer divisor (must have value <= i4_huge)

!~ 23.12.2010 (last modified; TIM)

use univ_kind_defs_mod, only: i8, i4
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_MESSAGE_LEN
use univ_exception_mod, only: exception, FATAL

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time
integer(i8), intent(IN) :: i8_divisor

type(time_obj) :: Time_ans      ! Procedure result

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'time_scalar_divide_i8'

character(len=MAX_MESSAGE_LEN) :: msg

!=== Executable statements:

! Enforce constraint that input integer must contain "small"
!  (4_byte-like) value:
if (i8_divisor > int(i4_huge, kind=i8)) then
   write (msg, '(a,i10)') 'input integer divisor must be <= ', i4_huge
   call exception(procedure_name, FATAL, trim(msg))
end if

Time_ans = time_scalar_divide_i4(Time, int(i8_divisor, kind=i4))

end function time_scalar_divide_i8


!```````````````````````````````````````````````````````````````````
function get_ticks_per_second(Time)  result(tps)

!+ Public method to obtain the current value of the private
!   element 'ticks_per_second' within a type(time_obj) object.
!
!  NOTE: 'ticks_per_second' can currently only have a "4-byte-integer-type"
!        value

!~ 22.12.2010 (last modified; TIM)

use univ_kind_defs_mod, only: i4
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

type(time_obj) :: Time

integer(i4) :: tps      ! Procedure result

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'get_ticks_per_second'

!=== Executable statements:

tps = int(Time%ticks_per_second, kind=i4)

end function get_ticks_per_second


!```````````````````````````````````````````````````````````````````
subroutine get_time_component_values(Time, seconds, ticks)

!+ Public method to obtain the values of the 'seconds' and 'ticks' components
!   of a type(time_obj) object.

!~ 14.12.2010 (last modified; TIM)

use univ_kind_defs_mod, only: i8
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time

integer(i8), intent(OUT) :: seconds, ticks

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'get_time_component_values'

!=== Executable statements:

seconds = Time%seconds
ticks = Time%ticks

end subroutine get_time_component_values


!```````````````````````````````````````````````````````````````````
subroutine module_time_init

!+ Initializes this module

!~ 28.12.2010 (last modified; TIM)

!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'module_time_init'

!=== Executable statements:

if (module_time_is_initialized) return  ! Silent return if already called

! Does not really do anything at the moment...

module_time_is_initialized = .TRUE.

end subroutine module_time_init


end module univ_integer_time_mod
