!```````````````````````````````````````````````````````````````````
module univ_namelist_utils_mod

!+

implicit none

!== Flexible procedure interfaces:

interface nml_num_value_check
   module procedure nml_num_value_check_i1_0d, nml_num_value_check_i2_0d,  &
        nml_num_value_check_i4_0d, nml_num_value_check_f4_0d,  &
        nml_num_value_check_f8_0d
   module procedure nml_num_value_check_i1_1d, nml_num_value_check_i2_1d,  &
        nml_num_value_check_i4_1d, nml_num_value_check_f4_1d,  &
        nml_num_value_check_f8_1d
end interface


contains


!```````````````````````````````````````````````````````````````````
subroutine gen_time_strrep_to_Time(gen_time_strrep, resultant_Time,  &
     syntax_error, mnsdy_Time, morbpd_Time, module_tps, units_str,  &
     mantissa_sign)

!+ Convert general time string representation into an object of type 'time_obj'

use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_MESSAGE_LEN
use univ_exception_mod, only: exception, FATAL
use univ_integer_time_mod, only: time_obj, operator(*), operator(/),  &
     operator(+), set_time

implicit none

!=== Argument declarations:

character(len=*), intent(IN) :: gen_time_strrep
type(time_obj), intent(IN) :: mnsdy_Time,  &   ! Mean diurnal period
                              morbpd_Time   ! Mean orbital (tropical) period
integer, intent(IN) :: module_tps   ! Ticks per second for 'time_obj' 
type(time_obj), intent(OUT) :: resultant_Time
logical, intent(OUT) :: syntax_error
character(len=MAX_PROC_NAME_LEN), optional, dimension(:), intent(OUT) ::  &
     units_str
integer, optional, intent(OUT) :: mantissa_sign

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'gen_time_strrep_to_Time'

integer, parameter :: max_n_vu_pairs = 3  ! = Number of known unit strings

integer, dimension(max_n_vu_pairs) :: ic_end, ic_start
character(len=MAX_MESSAGE_LEN) :: str_tmp
integer :: cnt, i, exponent, LHS_of_mantissa, RHS_of_mantissa, ic_dp, ic_e,  &
     ic_sp, ic_abslen, i_pair, i_RHS_beg, sign_tmp, module_tps_log10
type(time_obj) :: tmp_Time1, tmp_Time2, value_Time

!=== Executable statements:

! The general time string representation is composed of an integer or floating
!  point mantissa (with an optional exponent), a single space, and then a
!  string denoting the units of the preceding numerical value -- multiple,
!  additive value/unit pairs can be provided (separated from each other
!  by ', ').  Examples:
!
!  '23 body_days', '1.43e-1 body_days', '0.31 body_years', '33455. seconds',
!  '23 body_days, 33455. s', '10. body_years, 23.442 body_days',
!  '10. b_y, 23.442 b_d, 45.6666 seconds'

syntax_error = .FALSE.

resultant_Time = set_time(0, 0, module_tps)   ! Zero before summation
ic_abslen = len_trim(gen_time_strrep)
i = module_tps; module_tps_log10 = 0
do
   if (i < 10) exit
   module_tps_log10 = module_tps_log10+1; i = i/10
end do

! Count number of value/unit pairs that are present; record their start/end
!  indices:
ic_start(1) = 1; cnt = 0
do i=1,max_n_vu_pairs
   cnt = cnt+1
   ic_end(i) = index(gen_time_strrep(ic_start(i):ic_abslen), ', ')-1
   if (ic_end(i) == -1) exit   ! No further pairs
   if (i < max_n_vu_pairs) ic_start(i+1) = ic_end(i)+3
end do
if (cnt > 0) ic_end(cnt) = ic_abslen

if (present(units_str)) then
   if (size(units_str) < cnt) call exception(procedure_name, FATAL,  &
        'Optional array argument (units_str) has too few elements.')
end if
if (present(mantissa_sign)) mantissa_sign = 1  ! Initialize

sign_tmp = 0

! Process all value/unit pairs:
pair_loop: do i_pair=1,cnt

   str_tmp = gen_time_strrep(ic_start(i_pair):ic_end(i_pair))  ! Isolate pair
   
   ic_dp = index(str_tmp, '.')    ! Decimal point, if present
   i_RHS_beg = ic_dp+1   ! Where 'RHS_of_mantissa' substring begins

   LHS_of_mantissa = 0
   if (ic_dp > 1) then
      read (str_tmp(1:min(module_tps_log10, ic_dp-1)), *)  &
           LHS_of_mantissa   ! Must be a "real" numerical value
      sign_tmp = sign(1, LHS_of_mantissa)  ! Check for sign of mantissa
      if (sign_tmp < 0) LHS_of_mantissa = abs(LHS_of_mantissa)
   end if

   ic_sp = index(str_tmp, ' ')     ! Space separating value and units
   if (ic_sp == 0) then   ! Invalid gen_time_strrep
      syntax_error = .TRUE.
      return
   end if

   ! Determine 'RHS_of_mantissa' and any 'exponent'; fill end of
   !  'RHS_of_mantissa' with zeroes (up to magnitude of 'module_tps'):
   exponent = 0; RHS_of_mantissa = 0
   ic_e = index(str_tmp(1:ic_sp), 'e')               ! Exponent char, if present
   if (ic_e == 0) ic_e = index(str_tmp(1:ic_sp), 'E') !
   if (ic_e > 0) then    ! Must be a "real" numerical value
      if (i_RHS_beg < ic_e) then   ! Only if fractional part is present
         read (str_tmp(i_RHS_beg:min(i_RHS_beg+module_tps_log10-1, ic_e-1)),  &
              *) RHS_of_mantissa 
         RHS_of_mantissa = RHS_of_mantissa*  &
              (10**(max(0, (i_RHS_beg+module_tps_log10-1)-(ic_e-1))))
      end if
      read (str_tmp(ic_e+1:min(ic_e+module_tps_log10, ic_sp-1)), *) exponent
   else    ! Could be a "real" or "integer" numerical value
      if (i_RHS_beg < ic_sp) then   ! Only if fractional part is present
         read (str_tmp(i_RHS_beg:min(i_RHS_beg+module_tps_log10-1, ic_sp-1)),  &
              *) RHS_of_mantissa
         if (ic_dp > 0) then    ! Treat as "real" numerical value
            RHS_of_mantissa = RHS_of_mantissa*  &
                 (10**(max(0, (i_RHS_beg+module_tps_log10-1)-(ic_sp-1))))
         else   ! Treat as "integer" numerical value
            sign_tmp = sign(1, RHS_of_mantissa)  ! Check for sign of mantissa
            if (sign_tmp < 0) RHS_of_mantissa = abs(RHS_of_mantissa)
         end if
      end if
   end if

   ! Handle time units:
   select case(str_tmp(ic_sp+1:len_trim(str_tmp)))
   case ('body_days', 'b_d')    ! "Days" (diurnal periods)
      value_Time = RHS_of_mantissa*mnsdy_Time
      if (ic_dp > 0) then    ! Treat as "real" numerical value
         tmp_Time1 = LHS_of_mantissa*mnsdy_Time
         tmp_Time2 = value_Time/module_tps
         value_Time = tmp_Time1+tmp_Time2
      end if
   case ('body_years', 'b_y')   ! "Years" (tropical orbital periods)
      value_time = RHS_of_mantissa*morbpd_Time
      if (ic_dp > 0) then    ! Treat as "real" numerical value
         tmp_Time1 = LHS_of_mantissa*morbpd_Time
         tmp_Time2 = value_Time/module_tps
         value_Time = tmp_Time1+tmp_Time2       
      end if
   case ('seconds', 's')  ! SI seconds
      if (ic_dp > 0) then    ! Treat as "real" numerical value
         value_Time = set_time(LHS_of_mantissa, RHS_of_mantissa, module_tps)
      else   ! Treat as "integer" numerical value
         value_Time = set_time(RHS_of_mantissa, 0, module_tps)
      end if
   case default    ! Invalid gen_time_strrep
      syntax_error = .TRUE.
      return
   end select

   if (present(units_str)) units_str(i_pair) =  &
        str_tmp(ic_sp+1:len_trim(str_tmp))
   if (present(mantissa_sign)) mantissa_sign = mantissa_sign*sign_tmp

   if (exponent > 0) then
      do i=1,exponent
         tmp_Time1 = value_Time*10
         value_Time = tmp_Time1
      end do
   else if (exponent < 0) then
      do i=1,abs(exponent)
         tmp_Time1 = value_Time/10
         value_Time = tmp_Time1
      end do
   else
      continue   ! Merely multiplied by 1
   end if

   resultant_Time = resultant_Time+value_Time   ! Sum time value/unit pairs
end do pair_loop

end subroutine gen_time_strrep_to_Time


!```````````````````````````````````````````````````````````````````
subroutine Time_to_gen_time_strrep(Time, resultant_gen_time_strrep,  &
     rgt_strrep_len, mnsdy_Time, morbpd_Time, short_unit_style)

!+ Convert object of type 'time_obj' into a general time string representation

use univ_kind_defs_mod, only: i8
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_MESSAGE_LEN
use univ_exception_mod, only: exception, FATAL
use univ_integer_time_mod, only: time_obj, operator(*), operator(/),  &
     operator(-), get_time_component_values

implicit none

!=== Argument declarations:

type(time_obj), intent(IN) :: Time,  &
                              mnsdy_Time,  &   ! Mean diurnal period
                              morbpd_Time   ! Mean orbital (tropical) period
integer, intent(IN) :: rgt_strrep_len
logical, optional, intent(IN) :: short_unit_style
character(len=*), intent(OUT) :: resultant_gen_time_strrep

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'Time_to_gen_time_strrep'

character(len=MAX_MESSAGE_LEN) :: str_tmp
character(len=MAX_PROC_NAME_LEN) :: str_int
integer :: t_body_days, t_body_years, ic_end
integer(i8) :: t_seconds, t_ticks
type(time_obj) :: tmp_Time
logical :: use_short_unit_style

!=== Executable statements:

! The general time string representation is composed of an integer or floating
!  point mantissa (with an optional exponent), a single space, and then a
!  string denoting the units of the preceding numerical value -- multiple,
!  additive value/unit pairs can be provided (separated from each other
!  by ', ').  Examples:
!
!  '23 body_days', '1.43e-1 body_days', '0.31 body_years', '33455. seconds',
!  '23 body_days, 33455. s', '10. body_years, 23.442 body_days',
!  '10. b_y, 23.442 b_d, 45.6666 seconds'

use_short_unit_style = .FALSE.
if (present(short_unit_style)) use_short_unit_style = short_unit_style

if (use_short_unit_style) then    !
   str_tmp = '0 s'                !
else                              ! Initialize string
   str_tmp = '0 seconds'          !
end if                            !

t_body_years = Time/morbpd_Time; tmp_Time = Time-(t_body_years*morbpd_Time)
t_body_days = tmp_Time/mnsdy_Time; tmp_Time = tmp_Time-(t_body_days*mnsdy_Time)
call get_time_component_values(tmp_Time, t_seconds, t_ticks)

ic_end = 0

if (t_body_years > 0) then
   if (ic_end > 0) then
      str_tmp(ic_end+1:ic_end+2) = ', '
      ic_end = ic_end+2
   end if
   write (str_int, *) t_body_years
   if (use_short_unit_style) then
      str_tmp(ic_end+1:) = trim(adjustl(str_int))//' b_y'
   else
      str_tmp(ic_end+1:) = trim(adjustl(str_int))//' body_years'
   end if
   ic_end = len_trim(str_tmp)
end if

if (t_body_days > 0) then
   if (ic_end > 0) then
      str_tmp(ic_end+1:ic_end+2) = ', '
      ic_end = ic_end+2
   end if
   write (str_int, *) t_body_days
   if (use_short_unit_style) then
      str_tmp(ic_end+1:) = trim(adjustl(str_int))//' b_d'
   else
      str_tmp(ic_end+1:) = trim(adjustl(str_int))//' body_days'
   end if
   ic_end = len_trim(str_tmp)
end if

if (t_seconds > 0) then
   if (ic_end > 0) then
      str_tmp(ic_end+1:ic_end+2) = ', '
      ic_end = ic_end+2
   end if
   write (str_int, *) t_seconds
   str_tmp(ic_end+1:) = trim(adjustl(str_int))//'.'; ic_end = len_trim(str_tmp) 
   write (str_int, *) t_ticks
   if (use_short_unit_style) then
      str_tmp(ic_end+1:) = trim(adjustl(str_int))//' s'
   else
      str_tmp(ic_end+1:) = trim(adjustl(str_int))//' seconds'
   end if
   ic_end = len_trim(str_tmp)
end if

if (len_trim(str_tmp) > rgt_strrep_len) call exception(procedure_name, FATAL,  &
     'Character string overflow -- use a longer output string buffer.')

resultant_gen_time_strrep = trim(str_tmp)

end subroutine Time_to_gen_time_strrep


include 'univ_namelist_utils.f90.inc01'


end module univ_namelist_utils_mod
