!```````````````````````````````````````````````````````````````````
module univ_run_timing_mod

!+ Contains routines related to timing program execution
!
!    EXAMPLE OF USAGE:
!
!   use univ_run_timing_mod, only: run_timing_obj, initialize_elapsed_time,  &
!        record_wall_time, START, END, record_CPU_time, calc_elapsed_time,  &
!        display_elapsed_time
!
!   type(run_timing_obj) :: my_wall_timing
!   type(run_timing_obj), dimension(4) :: my_CPU_timings
!
!   call initialize_elapsed_time(my_CPU_timings)  ! Entire array at once
!
!   call record_wall_time(my_wall_timing, START)    ! Wall time (at start)
!
!   call record_CPU_time(my_CPU_timings(1), START)
!     ...do something...
!   call record_CPU_time(my_CPU_timings(1), END)
!   call calc_elapsed_time(my_CPU_timings(1), accumulate_elapsed_time=.TRUE.)
!
!   call record_CPU_time(my_CPU_timings(2), START)
!     ...do something...
!   call record_CPU_time(my_CPU_timings(2), END)
!   call calc_elapsed_time(my_CPU_timings(2), accumulate_elapsed_time=.TRUE.)
!
!   call record_wall_time(my_wall_timing, END)   ! Wall time (at finish)
!   call calc_elapsed_time(my_wall_timing, accumulate_elapsed_time=.FALSE.)
!
!   call display_elapsed_time(my_CPU_timings, 'CPU timings')
!   call display_elapsed_time(my_wall_timing, 'Elapsed wall time')

!~ 14.01.2011 (last modified; TIM)

use univ_kind_defs_mod, only: f8
use univ_integer_time_mod, only: time_obj
use univ_maxlen_string_mod, only: MAX_MESSAGE_LEN

implicit none

private

integer, public, parameter :: START = 1, END = 2

type run_timing_obj
   type(time_obj) :: Start_time, End_time  ! Relative start and end times
   type(time_obj) :: Elapsed_time
   character(len=MAX_MESSAGE_LEN) :: description
end type run_timing_obj

integer, parameter :: module_tps = 1000000000  ! Default ticks per second value

interface display_elapsed_time
   module procedure display_elapsed_time_0d, display_elapsed_time_1d
end interface

interface initialize_elapsed_time
   module procedure initialize_elapsed_time_0d
   module procedure initialize_elapsed_time_1d
end interface

public :: run_timing_obj, record_wall_time, record_CPU_time,  &
     calc_elapsed_time, initialize_elapsed_time, elapsed_time_s_eqv,  &
     display_elapsed_time


contains


!```````````````````````````````````````````````````````````````````
function wall_time_from_dnt()  result(Wall_time)

!+ Obtain current "wall time" from the Fortran 95 intrinsic function
!   `date_and_time`, transforming the UTC date/time to TT date/time, and then
!   finally storing it as an "integer time" object.

!~ 12.01.2011 (last modified; TIM)

use univ_kind_defs_mod, only: f8
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_integer_time_mod, only: set_time, time_obj
use univ_Earth_date_time_mod, only: UTC_date_from_dnt, UTC_to_TAI,  &
     G_date_time_to_spec_fmt, TAI_to_TT, canon_s_per_day_f

implicit none

!=== Argument declarations:

type(time_obj) :: Wall_time     ! Procedure result

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'wall_time_from_dnt'

integer :: ierr, year, month, day_of_month, hour, minute, TT_days, i_tmp,  &
     TT_ticks, TT_seconds
real(f8) :: f_seconds, UTC_JD1, UTC_JD2, TAI_JD1, TAI_JD2, TT_JD1, TT_JD2,  &
     f_tmp

!=== Executable statements:

call UTC_date_from_dnt(year, month, day_of_month, hour, minute, f_seconds)

call G_date_time_to_spec_fmt('UTC', year, month, day_of_month, hour, minute,  &
     f_seconds, UTC_JD1, UTC_JD2, ierr)

call UTC_to_TAI(UTC_JD1, UTC_JD2, TAI_JD1, TAI_JD2, ierr)

call TAI_to_TT(TAI_JD1, TAI_JD2, TT_JD1, TT_JD2, ierr)

i_tmp = int(TT_JD1)
f_tmp = TT_JD1-real(i_tmp, kind=f8)
TT_days = i_tmp

i_tmp = int(TT_JD2)
f_tmp = f_tmp+(TT_JD2-real(i_tmp, kind=f8))
TT_days = TT_days+i_tmp

f_tmp = f_tmp*canon_s_per_day_f
TT_seconds = int(f_tmp)

f_tmp = (f_tmp-real(TT_seconds, kind=f8))*real(module_tps, kind=f8)
TT_ticks = int(f_tmp)

Wall_time = set_time(TT_seconds, TT_ticks, module_tps, TT_days,  &
     body_rel_daylen_fp_rep = canon_s_per_day_f)

end function wall_time_from_dnt


!```````````````````````````````````````````````````````````````````
function CPU_time_from_ct()  result(CPU_run_time)

!+ Obtain current "CPU time" from the Fortran 95 intrinsic function
!   `cpu_time`, storing it as an "integer time" object.

!~ 12.01.2011 (last modified; TIM)

use univ_kind_defs_mod, only: f8
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_integer_time_mod, only: set_time, time_obj

implicit none

!=== Argument declarations:

type(time_obj) :: CPU_run_time     ! Procedure result

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'CPU_time_from_ct'

integer :: ticks, seconds
real(f8) :: f_tmp

real :: time_rep_s   ! No KIND for best intrinsic function compatibility

!=== Executable statements:

call cpu_time(time_rep_s)

seconds = int(time_rep_s)

f_tmp = (real(time_rep_s, kind=f8)-real(seconds, kind=f8))*  &
     real(module_tps, kind=f8)
ticks = int(f_tmp)

CPU_run_time = set_time(seconds, ticks, module_tps)

end function CPU_time_from_ct


!```````````````````````````````````````````````````````````````````
subroutine record_wall_time(rtiming_obj, mode)

!+ Public method that records a "wall time" to the appropriate element of a
!   'run_timing_obj'.

!~ 12.01.2011 (last modified; TIM)

!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

integer, intent(IN) :: mode  ! Either the 'START' or 'END' parameter (see above)

type(run_timing_obj), intent(INOUT) :: rtiming_obj

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'record_wall_time'

!=== Executable statements:

select case (mode)
case(START)
   rtiming_obj%Start_time = wall_time_from_dnt()
case(END)
   rtiming_obj%End_time = wall_time_from_dnt()
end select

end subroutine record_wall_time


!```````````````````````````````````````````````````````````````````
subroutine record_CPU_time(rtiming_obj, mode)

!+ Public method that records an instantaneous "CPU time" to the appropriate
!   element of a 'run_timing_obj'.

!~ 12.01.2011 (last modified; TIM)

!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

integer, intent(IN) :: mode  ! Either the 'START' or 'END' parameter (see above)

type(run_timing_obj), intent(INOUT) :: rtiming_obj

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'record_CPU_time'

!=== Executable statements:

select case (mode)
case(START)
   rtiming_obj%Start_time = CPU_time_from_ct()
case(END)
   rtiming_obj%End_time = CPU_time_from_ct()
end select

end subroutine record_CPU_time


!```````````````````````````````````````````````````````````````````
subroutine calc_elapsed_time(rtiming_obj, accumulate_elapsed_time)

!+ Public method that takes a 'run_timing_obj' as input, calculates the elapsed
!   time represented by the object's 'Start_time' and 'End_time' elements, and
!   records that value to the 'Elapsed_time' element of the object.

!~ 13.01.2011 (last modified; TIM)

!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_integer_time_mod, only: operator(-), operator(+)

implicit none

!=== Argument declarations:

logical, intent(IN) :: accumulate_elapsed_time

type(run_timing_obj), intent(INOUT) :: rtiming_obj

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'calc_elapsed_time'

!=== Executable statements:

if (accumulate_elapsed_time) then
   rtiming_obj%Elapsed_time = rtiming_obj%Elapsed_time+  &
       (rtiming_obj%End_time-rtiming_obj%Start_time)
else
   rtiming_obj%Elapsed_time = rtiming_obj%End_time-rtiming_obj%Start_time
end if

end subroutine calc_elapsed_time


!```````````````````````````````````````````````````````````````````
subroutine initialize_elapsed_time_0d(rtiming_obj)

!+ Public method (accessible as `initialize_elapsed_time`) that initializes
!   (sets to a time of zero) the 'Elapsed_time' element of a single
!   'run_timing_obj'.

!~ 13.01.2011 (last modified; TIM)

use univ_integer_time_mod, only: set_time, time_obj
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

type(run_timing_obj), intent(INOUT) :: rtiming_obj

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'initialize_elapsed_time'

!=== Executable statements:

rtiming_obj%Elapsed_time = set_time(0, 0, module_tps)

end subroutine initialize_elapsed_time_0d


!```````````````````````````````````````````````````````````````````
subroutine initialize_elapsed_time_1d(rtiming_obj_array)

!+ Public method (accessible as `initialize_elapsed_time`) that initializes
!   (sets to a time of zero) the 'Elapsed_time' element of an array of
!   'run_timing_obj' items.

!~ 13.01.2011 (last modified; TIM)

use univ_integer_time_mod, only: set_time, time_obj
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

type(run_timing_obj), dimension(:), intent(INOUT) :: rtiming_obj_array

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'initialize_elapsed_time'

integer :: i

!=== Executable statements:

do i=1,size(rtiming_obj_array, 1)
   rtiming_obj_array%Elapsed_time = set_time(0, 0, module_tps)
end do

end subroutine initialize_elapsed_time_1d


!```````````````````````````````````````````````````````````````````
function elapsed_time_s_eqv(rtiming_obj)  result(fp_seconds)

!+ Public method that returns a floating point equivalent [s] of the elapsed
!   time stored within a 'run_timing_obj'.

!~ 13.01.2011 (last modified; TIM)

use univ_kind_defs_mod, only: f8
use univ_integer_time_mod, only: time_eqv_fp_seconds
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

type(run_timing_obj), intent(INOUT) :: rtiming_obj

real(f8) :: fp_seconds       ! Procedure result

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'elapsed_time_s_eqv'

!=== Executable statements:

fp_seconds = time_eqv_fp_seconds(rtiming_obj%Elapsed_time)

end function elapsed_time_s_eqv


!```````````````````````````````````````````````````````````````````
subroutine display_elapsed_time_0d(rtiming_obj, header_str, Fortran_IO_unit)

!+ Public method (accessible as `display_elapsed_time`) that displays the
!   elapsed time within a single 'run_timing_obj' (to "stdout" by default, but
!   optionally to a given Fortran I/O unit).

!~ 25.05.2011 (last modified; TIM)

use univ_kind_defs_mod, only: f8
use univ_integer_time_mod, only: set_time, time_obj
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

character(len=*), intent(IN) :: header_str
integer, optional, intent(IN) :: Fortran_IO_unit ! Fortran I/O unit to write to

type(run_timing_obj), intent(INOUT) :: rtiming_obj

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'display_elapsed_time_0d'

integer :: IO_unit
real(f8) :: f_tmp

!=== Executable statements:

IO_unit = 6   ! Default is to write to "stdout"
if (present(Fortran_IO_unit)) IO_unit = Fortran_IO_unit

f_tmp = elapsed_time_s_eqv(rtiming_obj)
write (IO_unit, '(a,1x,a,a5,f12.3,a2)') trim(header_str),  &
     trim(rtiming_obj%description), ' --> ', f_tmp  , ' s'

end subroutine display_elapsed_time_0d


!```````````````````````````````````````````````````````````````````
subroutine display_elapsed_time_1d(rtiming_obj_array, header_str,  &
     Fortran_IO_unit)

!+ Public method (accessible as `display_elapsed_time`) that displays the
!   elapsed time within an array of 'run_timing_obj' items (to "stdout" by
!   default, but optionally to a given Fortran I/O unit).

!~ 13.01.2011 (last modified; TIM)

use univ_kind_defs_mod, only: f8
use univ_integer_time_mod, only: set_time, time_obj
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

character(len=*), intent(IN) :: header_str
integer, optional, intent(IN) :: Fortran_IO_unit ! Fortran I/O unit to write to

type(run_timing_obj), dimension(:), intent(INOUT) :: rtiming_obj_array

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'display_elapsed_time_1d'

integer :: IO_unit, i
real(f8) :: f_tmp

!=== Executable statements:

IO_unit = 6   ! Default is to write to "stdout"
if (present(Fortran_IO_unit)) IO_unit = Fortran_IO_unit

write (IO_unit, '(a)') trim(header_str)

do i=1,size(rtiming_obj_array, 1)
   f_tmp = elapsed_time_s_eqv(rtiming_obj_array(i))
   write (IO_unit, '(a,a5,f12.3,a2)') trim(rtiming_obj_array(i)%description),  &
        ' --> ', f_tmp  , ' s'
end do

end subroutine display_elapsed_time_1d


end module univ_run_timing_mod

