!```````````````````````````````````````````````````````````````````
module univ_string_mod

!+ Module that contains routines for operating on/with character strings

implicit none


contains


!```````````````````````````````````````````````````````````````````
function deblank(str_in)  result(str_out)

!+ Strips blanks (spaces only) from a string, and outputs a "deblanked" copy

!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

character(len=*), intent(IN) :: str_in

character(len=len_trim(str_in)) :: str_out     ! Procedure result

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'deblank'

integer :: str_in_len, i, n

!=== Executable statements:

str_out = ' '   ! Initialize output string
str_in_len = len_trim(str_in)

n = 0
do i=1,str_in_len
   if (str_in(i:i) /= ' ') then
      n = n+1
      str_out(n:n) = str_in(i:i)
   end if
end do

end function deblank


!```````````````````````````````````````````````````````````````````
subroutine char_array_to_string(input_char_array, output_string,  &
     output_string_maxlen)

!+ Converts a one-dimensional array of characters to a Fortran character string 

!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

character(len=1), dimension(:), intent(IN) :: input_char_array
integer, intent(IN) :: output_string_maxlen
character(len=*), intent(OUT) :: output_string

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'char_array_to_string'

integer :: i

!=== Executable statements:

! Initialize output string:
do i=1,output_string_maxlen
   output_string(i:i) = ' '
end do

! Copy array contents to string:
do i=1,min(size(input_char_array), output_string_maxlen)
   output_string(i:i) = input_char_array(i)
end do

end subroutine char_array_to_string


!```````````````````````````````````````````````````````````````````
function IO_fmt_for_string(str_len, add_parentheses)  result(fmt)

!+ Returns a Fortran I/O format string [e.g., '(a256)'] sufficient to
!   read/write a character string of length 'str_len' -- the surrounding
!   parentheses can optionally be suppressed.
!
!  String lengths < 10^10 can be correctly handled by this routine.

!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

integer, intent(IN) :: str_len
logical, optional, intent(IN) :: add_parentheses

character(len=12) :: fmt     ! Procedure result

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'IO_fmt_for_string'

logical :: want_to_add_parentheses
character(len=12) :: str_tmp1, str_tmp2

!=== Executable statements:

want_to_add_parentheses = .TRUE.
if (present(add_parentheses)) want_to_add_parentheses = add_parentheses

write (str_tmp1, '(i10)') str_len
str_tmp2 = adjustl(str_tmp1)

fmt = ' '   ! Initialize output string
if (want_to_add_parentheses) then
   write (fmt, '(a2,a,a1)') '(a', trim(str_tmp2), ')'
else
   write (fmt, '(a1,a)') 'a', trim(str_tmp2)
end if

end function IO_fmt_for_string


!```````````````````````````````````````````````````````````````````
pure function to_lowercase(str_in)  result(str_out)

!+ Given a string, returns a copy of that string that has all uppercase ASCII
!   characters replaced with lowercase ones.
!
!  String lengths < 10^10 can be correctly handled by this routine.

!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

character(len=*), intent(IN) :: str_in

character(len(str_in)) :: str_out     ! Procedure result

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'to_lowercase'

integer :: ic, i

character(len=26), parameter :: upc = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ',  &
                                lwc = 'abcdefghijklmnopqrstuvwxyz'

!=== Executable statements:

str_out = str_in
do i=1,len_trim(str_in)
   ic = index(upc, str_in(i:i))
   if (ic > 0) str_out(i:i) = lwc(ic:ic)
end do

end function to_lowercase


!```````````````````````````````````````````````````````````````````
pure function to_uppercase(str_in)  result(str_out)

!+ Given a string, returns a copy of that string that has all lowercase ASCII
!   characters replaced with uppercase ones.
!
!  String lengths < 10^10 can be correctly handled by this routine.

!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

character(len=*), intent(IN) :: str_in

character(len(str_in)) :: str_out     ! Procedure result

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'to_uppercase'

integer :: ic, i

character(len=26), parameter :: upc = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ',  &
                                lwc = 'abcdefghijklmnopqrstuvwxyz'

!=== Executable statements:

str_out = str_in
do i=1,len_trim(str_in)
   ic = index(lwc, str_in(i:i))
   if (ic > 0) str_out(i:i) = upc(ic:ic)
end do

end function to_uppercase


end module univ_string_mod
