!```````````````````````````````````````````````````````````````````
module univ_PS_proj_mod

!~ 11.11.2008 (last modified; TIM)

!+ Polar stereographic (PS) coordinate system routines

implicit none

private

public :: P3C_xy_PS

contains


!```````````````````````````````````````````````````````````````````
subroutine P3C_xy_PS(xpq,ypq,zpq,polelat,polelon,x,y,planetary_radius)

!~ 11.11.2008 (last modified; TIM)

!+ "Planetary 3D Cartesian" (P3C) coordinates to x/y PS coordinates
!
!  This subroutine computes coordinates (x,y) of point q projected onto the
!   polar stereographic plane whose pole point is located at planetographic
!   coordinates (polelat,polelon).  Input coordinates (xpq,ypq,zpq) of q are
!   defined in "Planetary 3D Cartesian" (P3C) space, where the origin is the
!   center of the planet, the z axis is the north pole, the x axis is the
!   equator and prime meridian, and the y axis is the equator and 90E.

use univ_kind_defs_mod, only: f4
use univ_convert_units_mod, only: deg_to_rad_f4

implicit none

!=== Argument declarations:

real(f4), intent(IN) :: xpq, ypq, zpq, polelat, polelon, planetary_radius

real(f4), intent(OUT) :: x, y

!== Local declarations:

real(f4) :: sinplat, cosplat, sinplon, cosplon, xq, yq, zq, xpp, ypp, zpp,  &
     t, prad2

!=== Executable statements:

! Evaluate sine and cosine of latitude and longitude of pole point p:

sinplat = sin(polelat*deg_to_rad_f4)
cosplat = cos(polelat*deg_to_rad_f4)
sinplon = sin(polelon*deg_to_rad_f4)
cosplon = cos(polelon*deg_to_rad_f4)

! Compute P3C coordinates of the pole point:

xpp = planetary_radius*cosplat*cosplon
ypp = planetary_radius*cosplat*sinplon
zpp = planetary_radius*sinplat

! Transform q point from P3C coordinates to 3D coordinates relative to the
!  polar stereographic plane with origin at the pole point, the z axis
!  pointing radially outward from the center of the planet, and the y axis
!  pointing northward along the local planetary meridian from the pole point:

xq = -sinplon*(xpq-xpp)+cosplon*(ypq-ypp)
yq = cosplat*(zpq-zpp)-sinplat*(cosplon*(xpq-xpp)+sinplon*(ypq-ypp))
zq = sinplat*(zpq-zpp)+cosplat*(cosplon*(xpq-xpp)+sinplon*(ypq-ypp))

! Parametric equation for line from antipodal point at (0,0,-2*prad) in 3D 
!  coordinates of polar stereographic plane to point q has the following
!  parameter (t) value on the polar stereographic plane (zq <= 0):

prad2 = planetary_radius*2._f4
t = prad2/(prad2+zq)

! This gives the following x and y coordinates for the projection of point q
! onto the polar stereographic plane:

x = xq*t
y = yq*t

end subroutine P3C_xy_PS


end module univ_PS_proj_mod
