!```````````````````````````````````````````````````````````````````
module univ_convert_units_mod

!+ Module containing unit conversion constants and routines

use univ_kind_defs_mod, only: f8, f4
use univ_math_constants_mod, only: pi_f8, pi_f4, ONE_f8, ONE_f4
use univ_phys_constants_mod, only: c_f8, c_f4, coulomb_const_f8,  &
     coulomb_const_f4

implicit none

!--- "Higher" precision definitions (see next section for "normal" precision):

!=== Temperature:
real(f8), parameter :: T00_f8 = 273.15_f8 ! [K] <-> [degC] value [K]

!=== Distance:
real(f8), parameter ::  &
     AU_to_m_f8 = 1.49597870700e11_f8,  & ! Distance [m -> AU] {1}
     m_to_AU_f8 = ONE_f8/AU_to_m_f8,  &   !   ^inverse^ [AU -> m]
     m_to_cm_f8 = 1.e2_f8,  &             ! Distance [m -> cm]
     cm_to_m_f8 = 1.e-2_f8,  &            !   ^inverse^ [cm -> m]
     m_to_um_f8 = 1.e6_f8,  &             ! Distance [m -> um]
     um_to_m_f8 = 1.e-6_f8,  &            !   ^inverse^ [um -> m]
     m_to_km_f8 = 1.e-3_f8,  &            ! Distance [m -> km]
     km_to_m_f8 = 1.e3_f8,  &             !   ^inverse^ [km -> m]
     um_to_cm_f8 = 1.e-4_f8,  &           ! Distance [um -> cm]
     cm_to_um_f8 = 1.e4_f8                !   ^inverse^ [cm -> um]

!=== Pressure:
real(f8), parameter ::  &
     Pa_to_dyne_o_cm2_f8 = 1.e1_f8,  &  ! Pressure [Pa -> dyne/cm^2]
     dyne_o_cm2_to_Pa_f8 = 1.e-1_f8,  & !   ^inverse^ [dyne/cm^2 -> Pa]
     Pa_to_bar_f8 = 1.e-5_f8,  &        ! Pressure [Pa -> bar]
     bar_to_Pa_f8 = 1.e5_f8,  &         !   ^inverse^ [bar -> Pa]
     Pa_to_mb_f8 = 1.e-2_f8,  &         ! Pressure [Pa -> mb]
     mb_to_Pa_f8 = 1.e2_f8,  &          !   ^inverse^ [mb -> Pa]
     Pa_to_hPa_f8 = 1.e-2_f8,  &        ! Pressure [Pa -> hPa]
     hPa_to_Pa_f8 = 1.e2_f8,  &         !   ^inverse^ [hPa -> Pa]
     mmHg_to_dyne_o_cm2_f8 = 1.3328e3_f8,  & ! Pressure [mmHg -> dyne/cm^2]
     dyne_o_cm2_to_mmHg_f8 = ONE_f8/mmHg_to_dyne_o_cm2_f8,  & ! [dyne/cm^2 -> mmHg]
     mmHg_to_Pa_f8 = mmHg_to_dyne_o_cm2_f8*dyne_o_cm2_to_Pa_f8,  & ! [mmHg -> Pa]
     Pa_to_mmHg_f8 = Pa_to_dyne_o_cm2_f8*dyne_o_cm2_to_mmHg_f8,  & ! [Pa -> mmHg]
     atm_to_Pa_f8 = 1.01325e5_f8,  &    ! [atm -> Pa]
     Pa_to_atm_f8 = ONE_f8/atm_to_Pa_f8,  &  ! [Pa -> atm]
     Pa_to_MPa_f8 = 1.e-6_f8,  &  ! [Pa -> MPa]
     MPa_to_Pa_f8 = 1.e6_f8  ! [MPa -> Pa]

!=== Charge:

  ! Do not use these MKS<->cgs conversions *alone* within formulae (an extra
  !  "Coulomb's constant" factor must be included)!:
real(f8), parameter ::  &
     C_to_esu_f8 = 10._f8*c_f8,  &       ! Charge [C -> esu] (esu = statcoulomb)
     esu_to_C_f8 = ONE_f8/C_to_esu_f8    !   ^inverse^ [esu -> C]

  ! *ONLY* use these factors to convert a form of Coulomb's Law (CL) to/from
  !  the MKS and cgs unit systems!: 
real(f8), parameter ::  &
     CL_cgs_to_MKS_f8 = coulomb_const_f8,  &    ! CL formula [MKS -> cgs]
     CL_MKS_to_cgs_f8 = ONE_f8/CL_cgs_to_MKS_f8 !   ^inverse^ [cgs -> MKS]

!=== Energy:
real(f8), parameter ::  &
     J_to_erg_f8 = 1.e7_f8,  &  ! Energy [J -> erg]
     erg_to_J_f8 = 1.e-7_f8,  & !   ^inverse^ [erg -> J]
     cal_to_J_f8 = 4.184_f8,  &         ! Energy [calorie -> J]
     J_to_cal_f8 = ONE_f8/cal_to_J_f8,  &   !   ^inverse^ [J -> calorie]
     J_to_MJ_f8 = 1.e-6_f8,  &    ! Energy [J -> MJ] 
     MJ_to_J_f8 = 1.e6_f8     ! ^inverse^ [MJ -> J]

!=== Mass:
real(f8), parameter ::  &
     kg_to_g_f8 = 1.e3_f8,  &  ! Mass [kg -> g]
     g_to_kg_f8 = 1.e-3_f8     !   ^inverse^ [g -> kg]

!=== Angular measure:
real(f8), parameter ::  &
     rad_to_deg_f8 = 180._f8/pi_f8,  &  ! Angular measure [rad -> deg]
     deg_to_rad_f8 = pi_f8/180._f8      !   ^inverse^ [deg -> rad]

!=== Area:
real(f8), parameter ::  &
     m2_to_cm2_f8 = m_to_cm_f8**2,  &  ! Area [m^2 -> cm^2]
     cm2_to_m2_f8 = cm_to_m_f8**2,  &  !   ^inverse^ [cm^2 -> m^2]
     m2_to_um2_f8 = m_to_um_f8**2,  &  ! Area [m^2 -> um^2]
     um2_to_m2_f8 = um_to_m_f8**2      !   ^inverse^ [um^2 -> m^2]

!=== Volume:
real(f8), parameter ::  &
     m3_to_cm3_f8 = m_to_cm_f8**3,  &  ! Volume [m^3 -> cm^3]
     cm3_to_m3_f8 = cm_to_m_f8**3      !   ^inverse^ [cm^3 -> m^3]

!=== Mass density:
real(f8), parameter ::  &
     kg_m3_to_g_cm3_f8 = kg_to_g_f8/m3_to_cm3_f8,  & ! Mass density [kg/m^3 -> g/cm^3]
     g_cm3_to_kg_m3_f8 = g_to_kg_f8/cm3_to_m3_f8     !   ^inverse^ [g/cm^3 -> kg/m^3]


!*** REFERENCES:
!
!{1} http://ssd.jpl.nasa.gov/?constants (circa 2014-10-08)
!    {The value for 1 au is from the IAU 2012 Resolution B1.}


!--- "Normal" precision definitions:
!
!    Note: explicitly calculate (with *_f4 constants) any values that were
!    calculated in the "higher" precision section -- for numerical consistency

real(f4), parameter :: &
     T00_f4 = real(T00_f8, kind=f4),  &
     AU_to_m_f4 = real(AU_to_m_f8, kind=f4),  & 
     m_to_AU_f4 = ONE_f4/AU_to_m_f4,  & 
     Pa_to_dyne_o_cm2_f4 = real(Pa_to_dyne_o_cm2_f8, kind=f4),  & 
     dyne_o_cm2_to_Pa_f4 = real(dyne_o_cm2_to_Pa_f8, kind=f4),  &
     Pa_to_bar_f4 = real(Pa_to_bar_f8, kind=f4),  &
     bar_to_Pa_f4 = real(bar_to_Pa_f8, kind=f4),  &
     Pa_to_mb_f4 = real(Pa_to_mb_f8, kind=f4),  &
     mb_to_Pa_f4 = real(mb_to_Pa_f8, kind=f4),  &
     Pa_to_hPa_f4 = real(Pa_to_hPa_f8, kind=f4),  &
     hPa_to_Pa_f4 = real(hPa_to_Pa_f8, kind=f4),  &
     mmHg_to_dyne_o_cm2_f4 = real(mmHg_to_dyne_o_cm2_f8, kind=f4),  &
     dyne_o_cm2_to_mmHg_f4 = ONE_f4/mmHg_to_dyne_o_cm2_f4,  &
     mmHg_to_Pa_f4 = real(mmHg_to_Pa_f8, kind=f4),  &
     Pa_to_mmHg_f4 = real(Pa_to_mmHg_f8, kind=f4),  &
     atm_to_Pa_f4 = real(atm_to_Pa_f8, kind=f4),  &
     Pa_to_atm_f4 = ONE_f4/atm_to_Pa_f4,  &
     Pa_to_MPa_f4 = real(Pa_to_MPa_f8, kind=f4),  &
     MPa_to_Pa_f4 = ONE_f4/Pa_to_MPa_f4,  &
     C_to_esu_f4 = 10._f4*c_f4,  &
     esu_to_C_f4 = ONE_f4/C_to_esu_f4,  &
     CL_cgs_to_MKS_f4 = coulomb_const_f4,  &
     CL_MKS_to_cgs_f4 = ONE_f4/CL_cgs_to_MKS_f4,  &
     J_to_erg_f4 = real(J_to_erg_f8, kind=f4),  &
     erg_to_J_f4 = real(erg_to_J_f8, kind=f4),  &
     cal_to_J_f4 = real(cal_to_J_f8, kind=f4),  &
     J_to_cal_f4 = real(J_to_cal_f8, kind=f4),  &
     J_to_MJ_f4 = real(J_to_MJ_f8, kind=f4),  & 
     MJ_to_J_f4 = ONE_f4/J_to_MJ_f4,  &
     kg_to_g_f4 = real(kg_to_g_f8, kind=f4),  &
     g_to_kg_f4 = real(g_to_kg_f8, kind=f4),  &
     m_to_cm_f4 = real(m_to_cm_f8, kind=f4),  &
     cm_to_m_f4 = real(cm_to_m_f8, kind=f4),  &
     m_to_um_f4 = real(m_to_um_f8, kind=f4),  &
     um_to_m_f4 = real(um_to_m_f8, kind=f4),  &
     m_to_km_f4 = real(m_to_km_f8, kind=f4),  &
     km_to_m_f4 = real(km_to_m_f8, kind=f4),  &
     um_to_cm_f4 = real(um_to_cm_f8, kind=f4),  &
     cm_to_um_f4 = real(cm_to_um_f8, kind=f4),  &
     rad_to_deg_f4 = 180._f4/pi_f4,  &
     deg_to_rad_f4 = pi_f4/180._f4,  & 
     m2_to_cm2_f4 = m_to_cm_f4**2,  &
     cm2_to_m2_f4 = cm_to_m_f4**2,  &
     m2_to_um2_f4 = m_to_um_f4**2,  &
     um2_to_m2_f4 = um_to_m_f4**2,  &
     m3_to_cm3_f4 = m_to_cm_f4**3,  &
     cm3_to_m3_f4 = cm_to_m_f4**3,  &
     kg_m3_to_g_cm3_f4 = kg_to_g_f4/m3_to_cm3_f4,  &
     g_cm3_to_kg_m3_f4 = g_to_kg_f4/cm3_to_m3_f4


contains


!```````````````````````````````````````````````````````````````````
subroutine distill_bytes(in_bytes, disp_value, disp_units)

!+ Determine whether given number of bytes is displayed best as B, KiB, MiB,
!   GiB, TiB, or PiB

use univ_kind_defs_mod, only: i8, f8
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

integer(i8), intent(IN) :: in_bytes

character(len=3), intent(OUT) :: disp_units
real(f8), intent(OUT) :: disp_value

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'distill_bytes'

integer(i8), parameter :: kibi = 2_i8**10_i8, mebi = 2_i8**20_i8,  &
     gibi = 2_i8**30_i8, tebi = 2_i8**40_i8, pebi = 2_i8**50_i8

!=== Executable statements:

if (in_bytes < kibi) then
   disp_units = 'B'
   disp_value = real(in_bytes, kind=f8)
else if (in_bytes < mebi) then
   disp_units = 'KiB'
   disp_value = real(in_bytes, kind=f8)/real(kibi, kind=f8)
else if (in_bytes < gibi) then
   disp_units = 'MiB'
   disp_value = real(in_bytes, kind=f8)/real(mebi, kind=f8)
else if (in_bytes < tebi) then
   disp_units = 'GiB'
   disp_value = real(in_bytes, kind=f8)/real(gibi, kind=f8)
else if (in_bytes < pebi) then
   disp_units = 'TiB'
   disp_value = real(in_bytes, kind=f8)/real(tebi, kind=f8)
else
   disp_units = 'PiB'
   disp_value = real(in_bytes, kind=f8)/real(pebi, kind=f8)
endif

end subroutine distill_bytes


end module univ_convert_units_mod
